//-------------------------------------------------------------------------------------------------
// AABB2.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_AABB2_
#define _include_AABB2_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.hpp"
#include "Vector2.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		AABB2
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class AABB2
{
public:
	Vector2		mins;
	Vector2		maxs;

	/// Construction / Destruction
	~AABB2(); // destructor
	AABB2(); // default constructor
	AABB2( const AABB2& AABB2 ); // copy constructor
	AABB2( const Vector2& point );
	AABB2( const Vector2& minimums, const Vector2& maximums );
	AABB2( const Vector2& minimums, const Vector2& maximums, const Vector2& padding );
	AABB2( float minX, float minY, float maxX, float maxY );

	/// Interface methods
	void SetFromMinsMaxs( const Vector2& minimums, const Vector2& maximums );
	void SetFromPoint( const Vector2& point );
	void SetFromTwoPoints( const Vector2& point1, const Vector2& point2 );
	void SetFromMinsMaxsPadding( const Vector2& minimums, const Vector2& maximums, const Vector2& padding );
	void SetFromPointPadding( const Vector2& point, const Vector2& padding );
	void SetFromTwoPointsPadding( const Vector2& point1, const Vector2& point2, const Vector2& padding );
	void SetFromMinXYMaxXY( float minX, float minY, float maxX, float maxY );
	void SetFromMinXYMaxXYPadXY( float minX, float minY, float maxX, float maxY, float padX, float padY );
	void AddPadding( const Vector2& padding );
	void AddPadding( float padX, float padY );
	const AABB2 Lerp( const AABB2& lerpTo, const float lerpTo_fraction ) const;
	const float CalcDiagonalLengthSquared() const;
	const float CalcDiagonalLength() const;
	const Vector2 CalcDimensions() const;
	const Vector2 CalcCenter() const;
	const bool IsOverlapping( const AABB2& compare );
	const AABB2 CalcOverlap( const AABB2& compare );
	const AABB2& StretchBoundsToIncludePoint( const Vector2& point );
	const AABB2& StretchBoundsToIncludeBox( const AABB2& boxBounds );
	const bool AABB2::IsPointInsideBounds( const Vector2& point );

	/// Operators
	const AABB2& operator = ( const AABB2& rhs );
	const AABB2& operator += ( const Vector2& translation );
	const AABB2& operator -= ( const Vector2& translation );
	const AABB2& operator *= ( const float scale );
	const AABB2& operator /= ( const float divisor );

	const AABB2 operator - () const;
	const AABB2 operator * ( const float scale ) const;

	const bool operator == ( const AABB2& rhs ) const;
	const bool operator != ( const AABB2& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const AABB2 operator * ( const float scale, const AABB2& rhs );
const bool EpsilonEqual( const AABB2& lhs, const AABB2& rhs, const float epsilon );


//-------------------------------------------------------------------------------------------------
// ~AABB2() - destructor
//-------------------------------------------------------------------------------------------------
inline AABB2::~AABB2()
{
}


//-------------------------------------------------------------------------------------------------
// AABB2() - default constructor
//-------------------------------------------------------------------------------------------------
inline AABB2::AABB2()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// AABB2( const AABB2& AABB2 ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline AABB2::AABB2( const AABB2& AABB2 )
:
mins( AABB2.mins ),
maxs( AABB2.maxs )
{
}


//-------------------------------------------------------------------------------------------------
// AABB2( Vector2 )
//-------------------------------------------------------------------------------------------------
inline AABB2::AABB2( const Vector2& point )
:
mins( point ),
maxs( point )
{
}


//-------------------------------------------------------------------------------------------------
// AABB2( Vector2, Vector2 )
//-------------------------------------------------------------------------------------------------
inline AABB2::AABB2( const Vector2& minimums, const Vector2& maximums )
:
mins( minimums ),
maxs( maximums )
{
}


//-------------------------------------------------------------------------------------------------
// AABB2( Vector2, Vector2, Vector2 )
//-------------------------------------------------------------------------------------------------
inline AABB2::AABB2( const Vector2& minimums, const Vector2& maximums, const Vector2& padding )
:
mins( minimums - padding ),
maxs( maximums + padding )
{
}


//-------------------------------------------------------------------------------------------------
// AABB2( float, float, float, float )
//-------------------------------------------------------------------------------------------------
inline AABB2::AABB2( float minX, float minY, float maxX, float maxY )
:
mins( minX, minY ),
maxs( maxX, maxY )
{
}


//-------------------------------------------------------------------------------------------------
// SetFromMinsMaxs
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromMinsMaxs( const Vector2& minimums, const Vector2& maximums )
{
	mins = minimums;
	maxs = maximums;
}


//-------------------------------------------------------------------------------------------------
// SetFromPoint
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromPoint( const Vector2& point )
{
	mins = point;
	maxs = point;
}


//-------------------------------------------------------------------------------------------------
// SetFromTwoPoints
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromTwoPoints( const Vector2& point1, const Vector2& point2 )
{
	if( point1.x > point2.x )
	{
		mins.x = point2.x;
		maxs.x = point1.x;
	}
	else
	{
		mins.x = point1.x;
		maxs.x = point2.x;
	}

	if( point1.y > point2.y )
	{
		mins.y = point2.y;
		maxs.y = point1.y;
	}
	else
	{
		mins.y = point1.y;
		maxs.y = point2.y;
	}
}


//-------------------------------------------------------------------------------------------------
// SetFromMinsMaxsPadding
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromMinsMaxsPadding( const Vector2& minimums, const Vector2& maximums, const Vector2& padding )
{
	mins = minimums - padding;
	maxs = maximums + padding;
}


//-------------------------------------------------------------------------------------------------
// SetFromPointPadding
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromPointPadding( const Vector2& point, const Vector2& padding )
{
	mins = point - padding;
	maxs = point + padding;
}


//-------------------------------------------------------------------------------------------------
// SetFromTwoPointsPadding
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromTwoPointsPadding( const Vector2& point1, const Vector2& point2, const Vector2& padding )
{
	if( point1.x > point2.x )
	{
		mins.x = point2.x - padding.x;
		maxs.x = point1.x + padding.x;
	}
	else
	{
		mins.x = point1.x - padding.x;
		maxs.x = point2.x + padding.x;
	}

	if( point1.y > point2.y )
	{
		mins.y = point2.y - padding.y;
		maxs.y = point1.y + padding.y;
	}
	else
	{
		mins.y = point1.y - padding.y;
		maxs.y = point2.y + padding.y;
	}
}


//-------------------------------------------------------------------------------------------------
// SetFromMinXYMaxXY
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromMinXYMaxXY( float minX, float minY, float maxX, float maxY )
{
	mins.SetXY( minX, minY );
	maxs.SetXY( maxX, maxY );
}


//-------------------------------------------------------------------------------------------------
// SetFromMinXYMaxXYPadXY
//-------------------------------------------------------------------------------------------------
inline void AABB2::SetFromMinXYMaxXYPadXY( float minX, float minY, float maxX, float maxY, float padX, float padY )
{
	mins.SetXY( minX - padX, minY - padY );
	maxs.SetXY( maxX + padX, maxY + padY );
}


//-------------------------------------------------------------------------------------------------
// AddPadding
//-------------------------------------------------------------------------------------------------
inline void AABB2::AddPadding( const Vector2& padding )
{
	mins -= padding;
	maxs += padding;
}


//-------------------------------------------------------------------------------------------------
// AddPadding
//-------------------------------------------------------------------------------------------------
inline void AABB2::AddPadding( float padX, float padY )
{
	mins.x -= padX;
	mins.y -= padY;
	maxs.x += padX;
	maxs.y += padY;
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::operator = ( const AABB2& rhs )
{
	if( this == &rhs )
		return *this;

	mins = rhs.mins;
	maxs = rhs.maxs;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::operator += ( const Vector2& translation )
{
	mins += translation;
	maxs += translation;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::operator -= ( const Vector2& translation )
{
	mins -= translation;
	maxs -= translation;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::operator *= ( const float scale )
{
	mins *= scale;
	maxs *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::operator /= ( const float divisor )
{
	if( divisor )
	{
		mins /= divisor;
		maxs /= divisor;
	}
	else
	{
		mins = Vector2::ZERO;
		maxs = Vector2::ZERO;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const AABB2 AABB2::operator - () const
{
	return AABB2( -mins, -maxs );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const AABB2 AABB2::operator * ( const float scale ) const
{
	return( AABB2( mins * scale, maxs * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool AABB2::operator == ( const AABB2& rhs ) const
{
	return( (mins == rhs.mins) && (maxs == rhs.maxs) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool AABB2::operator != ( const AABB2& rhs ) const
{
	return( (mins != rhs.mins) || (maxs != rhs.maxs) );
}


//-------------------------------------------------------------------------------------------------
// CalcDimensions
//-------------------------------------------------------------------------------------------------
inline const Vector2 AABB2::CalcDimensions() const
{
	return Vector2( maxs.x - mins.x, maxs.y - mins.y );
}


//-------------------------------------------------------------------------------------------------
// CalcCenter
//-------------------------------------------------------------------------------------------------
inline const Vector2 AABB2::CalcCenter() const
{
	return Vector2( (maxs.x + mins.x) * 0.5f, (maxs.y + mins.y) * 0.5f );
}


//-------------------------------------------------------------------------------------------------
// IsOverlapping
//-------------------------------------------------------------------------------------------------
inline const bool AABB2::IsOverlapping( const AABB2& compare )
{
	return!(( mins.x > compare.maxs.x ) ||
			( maxs.x < compare.mins.x ) ||
			( mins.y > compare.maxs.y ) ||
			( maxs.y < compare.mins.y ));
}


//-------------------------------------------------------------------------------------------------
// CalcOverlap
// 
// Note that if the two boxes do not actually overlap, the resulting AABB
//	may have negative sizes (mins > maxs, etc).  Call IsOverlapping() first
//	if it is unknown whether the boxes even overlap at all (and, in most
//	cases, do not bother calling this function if IsOverlapping() fails).
//-------------------------------------------------------------------------------------------------
inline const AABB2 AABB2::CalcOverlap( const AABB2& compare )
{
	/// NOTE: this is theoretically identical to the below commented-out
	//	code; however, either one could well be faster (due to the creation
	//	of extra temp variables, either because of the '?' operator here or
	//	the "overlapResult" variable below.  Curious.
	return AABB2(	(mins.x < compare.mins.x ? compare.mins.x : mins.x),
					(mins.y < compare.mins.y ? compare.mins.y : mins.y),
					(maxs.x > compare.maxs.x ? compare.maxs.x : maxs.x),
					(maxs.y > compare.maxs.y ? compare.maxs.y : maxs.y)		);
/*
	AABB2 overlapResult;

	/// Determine the maximum of the mins (x)
	if( mins.x < compare.mins.x )
	{
		overlapResult.mins.x = compare.mins.x;
	}
	else
	{
		overlapResult.mins.x = mins.x;
	}

	/// Determine the minimum of the maxs (x)
	if( maxs.x < compare.maxs.x )
	{
		overlapResult.maxs.x = maxs.x;
	}
	else
	{
		overlapResult.maxs.x = compare.maxs.x;
	}

	/// Determine the maximum of the mins (y)
	if( mins.y < compare.mins.y )
	{
		overlapResult.mins.y = compare.mins.y;
	}
	else
	{
		overlapResult.mins.y = mins.y;
	}

	/// Determine the minimum of the maxs (y)
	if( maxs.y < compare.maxs.y )
	{
		overlapResult.maxs.y = maxs.y;
	}
	else
	{
		overlapResult.maxs.y = compare.maxs.y;
	}

	return overlapResult;
*/
}


//-------------------------------------------------------------------------------------------------
// CalcDiagonalLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float AABB2::CalcDiagonalLengthSquared() const
{
	Vector2 size = CalcDimensions();
	return (size.x * size.x) + (size.y * size.y);
}


//-------------------------------------------------------------------------------------------------
// CalcDiagonalLength
//-------------------------------------------------------------------------------------------------
inline const float AABB2::CalcDiagonalLength() const
{
	Vector2 size = CalcDimensions();
	return (float) sqrt( (size.x * size.x) + (size.y * size.y) );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const AABB2 AABB2::Lerp( const AABB2& lerpTo, const float lerpTo_fraction ) const
{
	return AABB2( mins + lerpTo_fraction * (lerpTo.mins - mins),
					maxs + lerpTo_fraction * (lerpTo.maxs - maxs));
}


//-------------------------------------------------------------------------------------------------
// StretchBoundsToIncludePoint
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::StretchBoundsToIncludePoint( const Vector2& point )
{
	if( point.x < mins.x )
		mins.x = point.x;

	if( point.y < mins.y )
		mins.y = point.y;

	if( point.x > maxs.x )
		maxs.x = point.x;

	if( point.y > maxs.y )
		maxs.y = point.y;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// StretchBoundsToIncludeBox
//-------------------------------------------------------------------------------------------------
inline const AABB2& AABB2::StretchBoundsToIncludeBox( const AABB2& boxBounds )
{
	if( boxBounds.mins.x < mins.x )
		mins.x = boxBounds.mins.x;

	if( boxBounds.maxs.x > maxs.x )
		maxs.x = boxBounds.maxs.x;

	if( boxBounds.mins.y < mins.y )
		mins.y = boxBounds.mins.y;

	if( boxBounds.maxs.y > maxs.y )
		maxs.y = boxBounds.maxs.y;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// IsPointInsideBounds
//-------------------------------------------------------------------------------------------------
inline const bool AABB2::IsPointInsideBounds( const Vector2& point )
{
	return!(( point.x > maxs.x ) ||
			( point.x < mins.x ) ||
			( point.y > maxs.y ) ||
			( point.y < mins.y ));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, AABB2)
//-------------------------------------------------------------------------------------------------
inline const AABB2 operator * ( const float scale, const AABB2& rhs )
{
	return( AABB2( scale * rhs.mins, scale * rhs.maxs ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const AABB2& lhs, const AABB2& rhs, const float epsilon )
{
	return(	( fabs( lhs.mins.x - rhs.mins.x ) < epsilon )&&
			( fabs( lhs.mins.y - rhs.mins.y ) < epsilon ) &&
			( fabs( lhs.maxs.x - rhs.maxs.x ) < epsilon ) &&
			( fabs( lhs.maxs.y - rhs.maxs.y ) < epsilon ) );
}


#endif // _include_AABB2_
