//-------------------------------------------------------------------------------------------------
// Rgba.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_Rgba_
#define _include_Rgba_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.hpp"
#include "engineMath.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		Rgba
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class Rgba
{
public:
	static const Rgba WHITE;
	static const Rgba BLACK;
	static const Rgba RED;
	static const Rgba GREEN;
	static const Rgba BLUE;
	static const Rgba YELLOW;
	static const Rgba PURPLE;
	static const Rgba CYAN;
	static const Rgba ORANGE;
	static const Rgba GREY;
	static const Rgba DARKRED;
	static const Rgba DARKGREEN;
	static const Rgba DARKBLUE;
	static const Rgba DARKYELLOW;
	static const Rgba DARKPURPLE;
	static const Rgba DARKCYAN;
	static const Rgba DARKORANGE;
	static const Rgba DARKGREY;

	byte_t r; // these are an exception to the "no-public-members" rule of thumb
	byte_t g; // ...for the sake of clarity and brevity in use throughout the code
	byte_t b;
	byte_t a;

	/// Construction / Destruction
	~Rgba(); // destructor
	Rgba(); // default constructor
	Rgba( const Rgba& rgba ); // copy constructor
//	Rgba( float red, float green, float blue, float alpha = 1.0f ); - CAUSES COMPILER AMBIGUITY PROBLEMS
	Rgba( byte_t red, byte_t green, byte_t blue, byte_t alpha = (byte_t)0xff );

	/// Interface methods
	float CalcLengthSquared() const;
	float CalcLength() const;
	float CalcLengthRGB() const;
	float Normalize();
	void SetRGBAFloats( float red, float green, float blue, float alpha = 1.0f );
	void SetRGBABytes( byte_t red, byte_t green, byte_t blue, byte_t alpha = (byte_t)0xff );
	void GetRGBAFloats( float& red, float& green, float& blue, float& alpha ) const;
	void GetRGBABytes( byte_t& red, byte_t& green, byte_t& blue, byte_t& alpha ) const;
	void GetRGBFloats( float& red, float& green, float& blue ) const;
	void GetRGBBytes( byte_t& red, byte_t& green, byte_t& blue ) const;
	void SetFromDwordRGBA( dword_t rgba );
	void SetFromDwordBGRA( dword_t bgra );
	void SetFromDwordARGB( dword_t argb );
	void SetFromDwordABGR( dword_t abgr );
	dword_t GetDwordRGBA() const;
	dword_t GetDwordBGRA() const;
	dword_t GetDwordARGB() const;
	dword_t GetDwordABGR() const;
	const Rgba& ClampToRange( const Rgba& min, const Rgba& max );
	float CalcMaxFloatComponent() const;
	byte_t CalcMaxByteComponent() const;
	const Rgba Lerp( const Rgba& rhs, float rhs_fraction );
	static float ByteToFloat( byte_t b );
	static byte_t FloatToByte( float f );

	/// Operators
	const Rgba& operator = ( const Rgba& rhs );
	const Rgba& operator += ( const Rgba& rhs );
	const Rgba& operator -= ( const Rgba& rhs );
	const Rgba& operator *= ( float scale );
	const Rgba& operator /= ( float divisor );

	const Rgba operator - ( const Rgba& rhs ) const;
	const Rgba operator + ( const Rgba& rhs ) const;
	const Rgba operator * ( float scale ) const;
	float operator | ( const Rgba& rhs ) const; // dot product

	byte_t operator [] ( const int index ) const;
	byte_t& operator [] ( const int index );

	bool operator == ( const Rgba& rhs ) const;
	bool operator != ( const Rgba& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const Rgba operator * ( float scale, const Rgba& rhs );
bool FloatEpsilonEqual( const Rgba& lhs, const Rgba& rhs, float epsilon );
bool ByteEpsilonEqual( const Rgba& lhs, const Rgba& rhs, byte_t epsilon );


//-------------------------------------------------------------------------------------------------
// Rgba::~Rgba()
//-------------------------------------------------------------------------------------------------
inline Rgba::~Rgba()
{
}


//-------------------------------------------------------------------------------------------------
// Rgba::Rgba() - default constructor
//-------------------------------------------------------------------------------------------------
inline Rgba::Rgba()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// Rgba::Rgba( const Rgba& rgba ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline Rgba::Rgba( const Rgba& rgba )
:
r( rgba.r ),
g( rgba.g ),
b( rgba.b ),
a( rgba.a )
{
}


//-------------------------------------------------------------------------------------------------
// Rgba( float, float, float, float )
//-------------------------------------------------------------------------------------------------
/* THIS CAUSES TOO MUCH COMPILER AMBIGUITY
inline Rgba::Rgba( float red, float green, float blue, float alpha )
:
r( FloatToByte( red ) ),
g( FloatToByte( green ) ),
b( FloatToByte( blue ) ),
a( FloatToByte( alpha ) )
{
}
*/

//-------------------------------------------------------------------------------------------------
// Rgba( byte_t, byte_t, byte_t, byte_t )
//-------------------------------------------------------------------------------------------------
inline Rgba::Rgba( byte_t red, byte_t green, byte_t blue, byte_t alpha )
:
r( red ),
g( green ),
b( blue ),
a( alpha )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const Rgba& Rgba::operator = ( const Rgba& rhs )
{
	if( this == &rhs )
		return *this;

	r = rhs.r;
	g = rhs.g;
	b = rhs.b;
	a = rhs.a;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const Rgba& Rgba::operator += ( const Rgba& rhs )
{
	r = (byte_t) MinInt( 255, r + rhs.r );
	g = (byte_t) MinInt( 255, g + rhs.g );
	b = (byte_t) MinInt( 255, b + rhs.b );
	a = (byte_t) MinInt( 255, a + rhs.a );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const Rgba& Rgba::operator -= ( const Rgba& rhs )
{
	r = (byte_t) MaxInt( 0, r - rhs.r );
	g = (byte_t) MaxInt( 0, g - rhs.g );
	b = (byte_t) MaxInt( 0, b - rhs.b );
	a = (byte_t) MaxInt( 0, a - rhs.a );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const Rgba& Rgba::operator *= ( float scale )
{
	int scaledR = (int)( scale * (float)r );
	int scaledG = (int)( scale * (float)g );
	int scaledB = (int)( scale * (float)b );
	int scaledA = (int)( scale * (float)a );

	r = (byte_t) ClampInt( scaledR, 0, 255 );
	g = (byte_t) ClampInt( scaledG, 0, 255 );
	b = (byte_t) ClampInt( scaledB, 0, 255 );
	a = (byte_t) ClampInt( scaledA, 0, 255 );

	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const Rgba& Rgba::operator /= ( float divisor )
{
	if( divisor )
	{
		const float scale = 1.0f / divisor;

		int scaledR = (int)( scale * (float)r );
		int scaledG = (int)( scale * (float)g );
		int scaledB = (int)( scale * (float)b );
		int scaledA = (int)( scale * (float)a );

		r = (byte_t) ClampInt( scaledR, 0, 255 );
		g = (byte_t) ClampInt( scaledG, 0, 255 );
		b = (byte_t) ClampInt( scaledB, 0, 255 );
		a = (byte_t) ClampInt( scaledA, 0, 255 );
	}
	else
	{
		r = (byte_t) 0;
		g = (byte_t) 0;
		b = (byte_t) 0;
		a = (byte_t) 0;
	}

	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const Rgba Rgba::operator - ( const Rgba& rhs ) const
{
	return Rgba(
		(byte_t) MaxInt( 0, r - rhs.r ),
		(byte_t) MaxInt( 0, g - rhs.g ),
		(byte_t) MaxInt( 0, b - rhs.b ),
		(byte_t) MaxInt( 0, a - rhs.a ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const Rgba Rgba::operator + ( const Rgba& rhs ) const
{
	return Rgba(
		(byte_t) MinInt( 255, r + rhs.r ),
		(byte_t) MinInt( 255, g + rhs.g ),
		(byte_t) MinInt( 255, b + rhs.b ),
		(byte_t) MinInt( 255, a + rhs.a ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const Rgba Rgba::operator * ( float scale ) const
{
	int scaledR = (int)( scale * (float)r );
	int scaledG = (int)( scale * (float)g );
	int scaledB = (int)( scale * (float)b );
	int scaledA = (int)( scale * (float)a );

	return Rgba(
		(byte_t) ClampInt( scaledR, 0, 255 ),
		(byte_t) ClampInt( scaledG, 0, 255 ),
		(byte_t) ClampInt( scaledB, 0, 255 ),
		(byte_t) ClampInt( scaledA, 0, 255 ) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline byte_t Rgba::operator [] ( const int index ) const
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 4 ), "Rgba[%d] was an out of range index!", index );
	return ( &r )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline byte_t& Rgba::operator [] ( const int index )
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 4 ), "Rgba[%d] was an out of range index!", index );
	return ( &r )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline bool Rgba::operator == ( const Rgba& rhs ) const
{
	return( (r == rhs.r) && (g == rhs.g) && (b == rhs.b) && (a == rhs.a) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline bool Rgba::operator != ( const Rgba& rhs ) const
{
	return( (r != rhs.r) || (g != rhs.g) || (b != rhs.b) || (a != rhs.a) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline float Rgba::CalcLengthSquared() const
{
	float fR = ByteToFloat( r );
	float fG = ByteToFloat( g );
	float fB = ByteToFloat( b );
	float fA = ByteToFloat( a );

	return( fR*fR + fG*fG + fB*fB + fA*fA );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline float Rgba::CalcLength() const
{
	return( (float) sqrt( CalcLengthSquared() ) );
}


//-------------------------------------------------------------------------------------------------
// Normalize
//-------------------------------------------------------------------------------------------------
inline float Rgba::Normalize()
{
	float length = CalcLength();
	if( length )
	{
		*this /= length;
	}

	return( length );
}


//-------------------------------------------------------------------------------------------------
// SetRGBAFloats
//-------------------------------------------------------------------------------------------------
inline void Rgba::SetRGBAFloats( float red, float green, float blue, float alpha )
{
	r = FloatToByte( red );
	g = FloatToByte( green );
	b = FloatToByte( blue );
	a = FloatToByte( alpha );
}


//-------------------------------------------------------------------------------------------------
// SetRGBABytes
//-------------------------------------------------------------------------------------------------
inline void Rgba::SetRGBABytes( byte_t red, byte_t green, byte_t blue, byte_t alpha )
{
	r = red;
	g = green;
	b = blue;
	a = alpha;
}


//-------------------------------------------------------------------------------------------------
// GetRGBABytes
//-------------------------------------------------------------------------------------------------
inline void Rgba::GetRGBABytes( byte_t& red, byte_t& green, byte_t& blue, byte_t& alpha ) const
{
	red		= r;
	green	= g;
	blue	= b;
	alpha	= a;
}


//-------------------------------------------------------------------------------------------------
// GetRGBBytes
//-------------------------------------------------------------------------------------------------
inline void Rgba::GetRGBBytes( byte_t& red, byte_t& green, byte_t& blue ) const
{
	red		= r;
	green	= g;
	blue	= b;
}


//-------------------------------------------------------------------------------------------------
// GetRGBAFloats
//-------------------------------------------------------------------------------------------------
inline void Rgba::GetRGBAFloats( float& red, float& green, float& blue, float& alpha ) const
{
	red		= ByteToFloat( r );
	green	= ByteToFloat( g );
	blue	= ByteToFloat( b );
	alpha	= ByteToFloat( a );
}


//-------------------------------------------------------------------------------------------------
// GetRGBFloats
//-------------------------------------------------------------------------------------------------
inline void Rgba::GetRGBFloats( float& red, float& green, float& blue ) const
{
	red		= ByteToFloat( r );
	green	= ByteToFloat( g );
	blue	= ByteToFloat( b );
}


//-------------------------------------------------------------------------------------------------
// SetFromDwordRGBA
//-------------------------------------------------------------------------------------------------
inline void Rgba::SetFromDwordRGBA( dword_t rgba )
{
	r = (byte_t)( (rgba & 0xff000000) >> 24 );
	g = (byte_t)( (rgba & 0x00ff0000) >> 16 );
	b = (byte_t)( (rgba & 0x0000ff00) >> 8 );
	a = (byte_t)( (rgba & 0x000000ff) >> 0 );
}


//-------------------------------------------------------------------------------------------------
// SetFromDwordBGRA
//-------------------------------------------------------------------------------------------------
inline void Rgba::SetFromDwordBGRA( dword_t bgra )
{
	b = (byte_t)( (bgra & 0xff000000) >> 24 );
	g = (byte_t)( (bgra & 0x00ff0000) >> 16 );
	r = (byte_t)( (bgra & 0x0000ff00) >> 8 );
	a = (byte_t)( (bgra & 0x000000ff) >> 0 );
}


//-------------------------------------------------------------------------------------------------
// SetFromDwordARGB
//-------------------------------------------------------------------------------------------------
inline void Rgba::SetFromDwordARGB( dword_t argb )
{
	a = (byte_t)( (argb & 0xff000000) >> 24 );
	r = (byte_t)( (argb & 0x00ff0000) >> 16 );
	g = (byte_t)( (argb & 0x0000ff00) >> 8 );
	b = (byte_t)( (argb & 0x000000ff) >> 0 );
}


//-------------------------------------------------------------------------------------------------
// SetFromDwordABGR
//-------------------------------------------------------------------------------------------------
inline void Rgba::SetFromDwordABGR( dword_t abgr )
{
	a = (byte_t)( (abgr & 0xff000000) >> 24 );
	b = (byte_t)( (abgr & 0x00ff0000) >> 16 );
	g = (byte_t)( (abgr & 0x0000ff00) >> 8 );
	r = (byte_t)( (abgr & 0x000000ff) >> 0 );
}


//-------------------------------------------------------------------------------------------------
// GetDwordRGBA
//-------------------------------------------------------------------------------------------------
inline dword_t Rgba::GetDwordRGBA() const
{
	return(	(r << 24) &
			(g << 16) &
			(b << 8) &
			(a << 0) );
}


//-------------------------------------------------------------------------------------------------
// GetDwordBGRA
//-------------------------------------------------------------------------------------------------
inline dword_t Rgba::GetDwordBGRA() const
{
	return(	(b << 24) &
			(g << 16) &
			(r << 8) &
			(a << 0) );
}


//-------------------------------------------------------------------------------------------------
// GetDwordARGB
//-------------------------------------------------------------------------------------------------
inline dword_t Rgba::GetDwordARGB() const
{
	return(	(a << 24) &
			(r << 16) &
			(g << 8) &
			(b << 0) );
}


//-------------------------------------------------------------------------------------------------
// GetDwordABGR
//-------------------------------------------------------------------------------------------------
inline dword_t Rgba::GetDwordABGR() const
{
	return(	(a << 24) &
			(b << 16) &
			(g << 8) &
			(r << 0) );
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline const Rgba& Rgba::ClampToRange( const Rgba& min, const Rgba& max )
{
	/// Clamp R component
	if( r > max.r )
		r = max.r;
	else if( r < min.r )
		r = min.r;

	/// Clamp G component
	if( g > max.g )
		g = max.g;
	else if( g < min.g )
		g = min.g;

	/// Clamp B component
	if( b > max.b )
		b = max.b;
	else if( b < min.b )
		b = min.b;

	/// Clamp A component
	if( a > max.a )
		a = max.a;
	else if( a < min.a )
		a = min.a;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// CalcMaxFloatComponent
//-------------------------------------------------------------------------------------------------
inline float Rgba::CalcMaxFloatComponent() const
{
	return ByteToFloat( CalcMaxByteComponent() );
}


//-------------------------------------------------------------------------------------------------
// CalcMaxByteComponent
//-------------------------------------------------------------------------------------------------
inline byte_t Rgba::CalcMaxByteComponent() const
{
	/// Assume r is the largest
	byte_t max = r;

	/// Check if g is bigger
	if( g > max )
		max = g;

	/// Check if b is bigger
	if( b > max )
		max = b;

	/// Check if a is bigger
	if( a > max )
		max = a;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const Rgba Rgba::Lerp( const Rgba& rhs, float rhs_fraction )
{
	float red	= (float)r + ( rhs_fraction * (float)( rhs.r - r ) );
	float green	= (float)g + ( rhs_fraction * (float)( rhs.g - g ) );
	float blue	= (float)b + ( rhs_fraction * (float)( rhs.b - b ) );
	float alpha	= (float)a + ( rhs_fraction * (float)( rhs.a - a ) );

	return Rgba(
		(byte_t) red,
		(byte_t) green,
		(byte_t) blue,
		(byte_t) alpha );
}


//-------------------------------------------------------------------------------------------------
// ByteToFloat
//-------------------------------------------------------------------------------------------------
STATIC inline float Rgba::ByteToFloat( byte_t b )
{
	return( float( b ) * (1.0f / 255.0f) );
}


//-------------------------------------------------------------------------------------------------
// FloatToByte
//-------------------------------------------------------------------------------------------------
STATIC inline byte_t Rgba::FloatToByte( float f )
{
	return( (byte_t) (f * 255.0f) );
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, Rgba)
//-------------------------------------------------------------------------------------------------
inline const Rgba operator * ( float scale, const Rgba& rhs )
{
	int scaledR = (int)( scale * (float)rhs.r );
	int scaledG = (int)( scale * (float)rhs.g );
	int scaledB = (int)( scale * (float)rhs.b );
	int scaledA = (int)( scale * (float)rhs.a );

	return Rgba(
		(byte_t) ClampInt( scaledR, 0, 255 ),
		(byte_t) ClampInt( scaledG, 0, 255 ),
		(byte_t) ClampInt( scaledB, 0, 255 ),
		(byte_t) ClampInt( scaledA, 0, 255 ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const Rgba& lhs, const Rgba& rhs, float epsilon )
{
	return( fabs( lhs.r - rhs.r ) < epsilon &&
			fabs( lhs.g - rhs.g ) < epsilon &&
			fabs( lhs.b - rhs.b ) < epsilon &&
			fabs( lhs.a - rhs.a ) < epsilon );
}


#endif // _include_Rgba_
