//-------------------------------------------------------------------------------------------------
// Texture.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "Texture.hpp"
#include "engineAssert.hpp"


//-------------------------------------------------------------------------------------------------
// Class constants and statics
//-------------------------------------------------------------------------------------------------
const dword_t Texture::TEXNUM_UNREGISTERED = 0xffffffff;
STATIC bool Texture::_isOpenGLReady = false;



//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
Texture::Texture( const std::string& filePath )
:
_width( 0 ),
_height( 0 ),
_bytesPerPixel( 0 ),
_internalFormat( 0 ),
_glTexNum( TEXNUM_UNREGISTERED )
{
	Image image( filePath );
	_InitFromImage( image );
	_CreateOpenGLTexture( image );
}


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
Texture::Texture( const Image& image )
:
_width( 0 ),
_height( 0 ),
_bytesPerPixel( 0 ),
_internalFormat( 0 ),
_glTexNum( TEXNUM_UNREGISTERED )
{
	_InitFromImage( image );
	_CreateOpenGLTexture( image );
}


//-------------------------------------------------------------------------------------------------
// Destructor
//-------------------------------------------------------------------------------------------------
Texture::~Texture()
{
}


//-------------------------------------------------------------------------------------------------
// InitFromImage
//-------------------------------------------------------------------------------------------------
void Texture::_InitFromImage( const Image& image )
{
	_width = image._width;
	_height = image._height;
	_bytesPerPixel = image._bytesPerPixel;
	_internalFormat = image._internalFormat;
}


//-------------------------------------------------------------------------------------------------
// CreateOpenGLTexture
//-------------------------------------------------------------------------------------------------
void Texture::_CreateOpenGLTexture( const Image& image )
{
	/// Make sure opengl has been started
	VALIDATE_MSG1( _isOpenGLReady, "ERROR: tried to create texture from image \"%s\" before TheRenderer was initialized; make sure Texture objects do not get instantiated until PostInit()", image.GetFilePath() );

	/// Make sure the texture dimensions are a power of two, and within the range [8,2048]
	VALIDATE_MSG2( IsAPowerOfTwo( image._width ), "ERROR: texture \"%s\" had a non-power-of-two width of %d; image width\n must be 8, 16, 32, 64, 128, 256, 512, 1024, or 2048", image.GetFilePath(), image._width );
	VALIDATE_MSG2( IsAPowerOfTwo( image._height ), "ERROR: texture \"%s\" had a non-power-of-two height of %d; image height\n must be 8, 16, 32, 64, 128, 256, 512, 1024, or 2048", image.GetFilePath(), image._height );
	VALIDATE_MSG2( (image._width >= 8) && (image._width <= 2048), "ERROR: texture \"%s\" had an illegal width of %d; must be 8, 16, 32, 64, 128, 256, 512, 1024, or 2048", image.GetFilePath(), image._width );
	VALIDATE_MSG2( (image._height >= 8) && (image._height <= 2048), "ERROR: texture \"%s\" had an illegal height of %d; must be 8, 16, 32, 64, 128, 256, 512, 1024, or 2048", image.GetFilePath(), image._height );

	/// Create an opengl texture name (number) for this image
	_glTexNum = 0;
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 ); // pixel info is byte-aligned
	glGenTextures( 1, &_glTexNum ); // create a texnum for this new image
	glBindTexture( GL_TEXTURE_2D, _glTexNum ); // select this as the current 2d texture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );

	/// Upload image data to OpenGL
	glTexImage2D( 
		GL_TEXTURE_2D,			// creating a 2d texture
		0,						// use the root picmip level for this texture (0 = best)
		_internalFormat,		// format opengl should use internally for this image
		_width,					// width of the source data, in pixels
		_height,				// height of the source data, in pixels
		0,						// do not add a border
		image._bufferFormat,	// the layout of the pixels in the buffer
		GL_UNSIGNED_BYTE,		// pixel color components are unsigned bytes
		image._pixelData );		// buffer containing pixel data bytes
}



