//-------------------------------------------------------------------------------------------------
// TheInputSystem.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheInputSystem.hpp"
#include "TheVideoSystem.hpp"
#include "TheConsole.hpp"
#include "TheGameInterfaceAbstract.hpp"
#include "engineAssert.hpp"


//-------------------------------------------------------------------------------------------------
// Subsystem name macro (for assertions, etc)
//-------------------------------------------------------------------------------------------------
#define SUBSYSTEM_NAME "TheInputSystem"


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC bool				TheInputSystem::_isInitialized		= false;


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheInputSystem::TheInputSystem()
:
_mouseDeltaX( 0 ),
_mouseDeltaY( 0 ),
_mousePosX( 0 ),
_mousePosY( 0 ),
_lastKnownVisibleMousePosition( IntVector2::ZERO ),
_useWindowsMouseCursor( true ),
_doesAppHaveFocus( true )
{
	/// Initialize complex members
	memset( _isKeyDown, 0, sizeof( _isKeyDown ) );
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheInputSystem::~TheInputSystem()
{
	/// Make sure the subsystem was shut down before being destroyed
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was destroyed while running; Shutdown() should be called first!" );
}


//-------------------------------------------------------------------------------------------------
// Startup
//-------------------------------------------------------------------------------------------------
void TheInputSystem::Startup()
{
	theConsole.Printf( "::: Initialization : Input System (starting)\n" );

	/// Make sure the subsystem doesn't get initialized twice
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was initialized twice; either a second "SUBSYSTEM_NAME" was started or Startup() was called on an already-running system." );

	/// Set initialization complete
	_isInitialized = true;
	theConsole.Printf( "::: Initialization : Input System (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheInputSystem::Shutdown()
{
	theConsole.Printf( "::: Shutdown : Input System (starting)\n" );

	/// Restore the cursor if it was in lock & hide mode
	theInputSystem.UseWindowsMouseCursor( true );

	/// De-initialize the system
	_isInitialized = false;
	theConsole.Printf( "::: Shutdown : Input System (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// IsKeyDown
//-------------------------------------------------------------------------------------------------
const bool TheInputSystem::IsKeyDown( byte_t vKeyCode ) const
{
	ASSERT_MSG0( _isInitialized, SUBSYSTEM_NAME"::IsKeyDown() was called while "SUBSYSTEM_NAME" was uninitialized; call Startup() first." );
	ASSERT_MSG1( vKeyCode >= 0 && vKeyCode <= 0xff, SUBSYSTEM_NAME"::IsKeyDown( %d ) : illegal <vKeyCode>.", vKeyCode );

	return _isKeyDown[ vKeyCode ];
}


//-------------------------------------------------------------------------------------------------
// ProcessInputEvent
//-------------------------------------------------------------------------------------------------
const bool TheInputSystem::ProcessInputEvent( const UINT messageID, const WPARAM wParam, const LPARAM lParam )
{
	inputEventType_e eventType = INPUT_EVENT_NONE;
	bool wasHandled = false;
	byte_t key = byte_t( wParam );

	/// Determine which type of input event it was
	switch( messageID )
	{
		case WM_KEYDOWN:
			if( HIWORD( lParam ) & KF_REPEAT )
				eventType = INPUT_EVENT_KEYREPEAT;
			else
				eventType = INPUT_EVENT_KEYDOWN;
			break;

		case WM_KEYUP:
			eventType = INPUT_EVENT_KEYUP;
			break;

		case WM_LBUTTONDBLCLK:
			eventType = INPUT_EVENT_KEYREPEAT;
			key = VK_LBUTTON;
			break;

		case WM_RBUTTONDBLCLK:
			eventType = INPUT_EVENT_KEYREPEAT;
			key = VK_RBUTTON;
			break;

		case WM_MBUTTONDBLCLK:
			eventType = INPUT_EVENT_KEYREPEAT;
			key = VK_MBUTTON;
			break;

		case 0x020A: // WM_MOUSEWHEEL
			short highWord = (short)( wParam >> 16 );
			if( highWord > 0 )
			{
				eventType = INPUT_EVENT_MOUSEWHEEL_UP;
				key = (byte_t)( highWord / 120 ); // WHEEL_DELTA
			}
			else if( highWord < 0 )
			{
				eventType = INPUT_EVENT_MOUSEWHEEL_DOWN;
				key = (byte_t)( (-highWord) / 120 ); // WHEEL_DELTA
			}
			break;
	}

	/// Update the key state table
	if( (eventType == INPUT_EVENT_KEYDOWN) || (eventType == INPUT_EVENT_KEYREPEAT) )
	{
		_isKeyDown[ key ] = true;
	}
	else if( eventType == INPUT_EVENT_KEYUP )
	{
		_isKeyDown[ key ] = false;
	}

	/// Tell the game
	if( eventType != INPUT_EVENT_NONE )
	{
		wasHandled = TheGameInterfaceAbstract::Singleton()->ProcessInputEvent( eventType, key );
	}

	return wasHandled;
}


//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void TheInputSystem::Update()
{
	/// Check if the game has gone out of focus
	if( !_doesAppHaveFocus )
	{
		/// Report no mouse movement to the game if it doesn't have focus
		_mouseDeltaX = 0;
		_mouseDeltaY = 0;
		return;
	}

	/// Get the windows mouse cursor position
	POINT mousePos;
	::GetCursorPos( &mousePos );

	/// Get movement deltas as compared with last frame
	_mouseDeltaX = mousePos.x - _mousePosX;
	_mouseDeltaY = mousePos.y - _mousePosY;

	if( _useWindowsMouseCursor )
	{
		/// Update internal mouse cursor position
		_mousePosX = mousePos.x;
		_mousePosY = mousePos.y;
		_lastKnownVisibleMousePosition.SetXY( _mousePosX, _mousePosY );
	}
	else
	{
		/// Clamp (invisible) mouse back to center of screen to make room for more deltas
		int xScreenMiddle = theVideoSystem.GetCurrentVidModeWidth() / 2;
		int yScreenMiddle = theVideoSystem.GetCurrentVidModeHeight() / 2;
		SetMousePosScreenAbsolute( xScreenMiddle, yScreenMiddle );
	}
}


//-------------------------------------------------------------------------------------------------
// UseWindowsMouseCursor
//-------------------------------------------------------------------------------------------------
void TheInputSystem::UseWindowsMouseCursor( const bool enable )
{
	/// Check if the state is already correct
	if( _useWindowsMouseCursor == enable )
		return; // no change

	/// Check if enabling or disabling the windows mouse cursor
	if( enable )
	{
		_useWindowsMouseCursor = true;
		SetMousePosScreenAbsolute( _lastKnownVisibleMousePosition.x, _lastKnownVisibleMousePosition.y );
		::ShowCursor( TRUE );
	}
	else
	{
		_useWindowsMouseCursor = false;
		_lastKnownVisibleMousePosition.SetXY( _mousePosX, _mousePosY );
		::ShowCursor( FALSE );
	}
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltas
// 
// Sets <mouseDeltaX> and <mouseDeltaY> to the amount of mouse movement for this frame
//-------------------------------------------------------------------------------------------------
void TheInputSystem::GetMouseDeltas( int& mouseDeltaX, int& mouseDeltaY )
{
	mouseDeltaX = _mouseDeltaX;
	mouseDeltaY = _mouseDeltaY;
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltas
// 
// Returns the amount of mouse movement for this frame
//-------------------------------------------------------------------------------------------------
const IntVector2 TheInputSystem::GetMouseDeltas()
{
	return IntVector2( _mouseDeltaX, _mouseDeltaY );
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltaX
//-------------------------------------------------------------------------------------------------
int TheInputSystem::GetMouseDeltaX()
{
	return _mouseDeltaX;
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltaY
//-------------------------------------------------------------------------------------------------
int TheInputSystem::GetMouseDeltaY()
{
	return _mouseDeltaY;
}




//-------------------------------------------------------------------------------------------------
// GetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::GetMousePosScreenAbsolute( int& screenMousePosX, int& screenMousePosY ) const
{
	screenMousePosX = _lastKnownVisibleMousePosition.x;
	screenMousePosY = _lastKnownVisibleMousePosition.y;
}


//-------------------------------------------------------------------------------------------------
// GetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::GetMousePosClientAbsolute( int& clientMousePosX, int& clientMousePosY ) const
{
	theVideoSystem.ConvertScreenCoordsToClientCoords( _lastKnownVisibleMousePosition.x, _lastKnownVisibleMousePosition.y, clientMousePosX, clientMousePosY );
}


//-------------------------------------------------------------------------------------------------
// GetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
const IntVector2 TheInputSystem::GetMousePosScreenAbsolute() const
{
//	return IntVector2( _mousePosX, _mousePosY );
	return _lastKnownVisibleMousePosition;
}


//-------------------------------------------------------------------------------------------------
// GetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
const IntVector2 TheInputSystem::GetMousePosClientAbsolute() const
{
	IntVector2 retVal;
	theVideoSystem.ConvertScreenCoordsToClientCoords( _lastKnownVisibleMousePosition.x, _lastKnownVisibleMousePosition.y, retVal.x, retVal.y );
	return retVal;
}


//-------------------------------------------------------------------------------------------------
// GetMousePosScreenAbsoluteFloat
//-------------------------------------------------------------------------------------------------
const Vector2 TheInputSystem::GetMousePosScreenAbsoluteFloat() const
{
	return Vector2( (float)_lastKnownVisibleMousePosition.x, (float)_lastKnownVisibleMousePosition.y );
}


//-------------------------------------------------------------------------------------------------
// GetMousePosClientAbsoluteFloat
//-------------------------------------------------------------------------------------------------
const Vector2 TheInputSystem::GetMousePosClientAbsoluteFloat() const
{
	int x, y;
	theVideoSystem.ConvertScreenCoordsToClientCoords( _lastKnownVisibleMousePosition.x, _lastKnownVisibleMousePosition.y, x, y );
	return Vector2( (float)x, (float)y );
}

	
//-------------------------------------------------------------------------------------------------
// SetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosScreenAbsolute( const int screenMousePosX, const int screenMousePosY )
{
	_mousePosX = screenMousePosX;
	_mousePosY = screenMousePosY;
	::SetCursorPos( _mousePosX, _mousePosY );
}


//-------------------------------------------------------------------------------------------------
// SetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosClientAbsolute( const int clientMousePosX, const int clientMousePosY )
{
	theVideoSystem.ConvertClientCoordsToScreenCoords( clientMousePosX, clientMousePosY, _mousePosX, _mousePosY );
	::SetCursorPos( _mousePosX, _mousePosY );
}


//-------------------------------------------------------------------------------------------------
// SetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosScreenAbsolute( const IntVector2& screenMousePos )
{
	SetMousePosScreenAbsolute( screenMousePos.x, screenMousePos.y );
}


//-------------------------------------------------------------------------------------------------
// SetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosClientAbsolute( const IntVector2& clientMousePos )
{
	SetMousePosClientAbsolute( clientMousePos.x, clientMousePos.y );
}


//-------------------------------------------------------------------------------------------------
// SetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosScreenAbsolute( const Vector2& screenMousePos )
{
	SetMousePosScreenAbsolute( (int)screenMousePos.x, (int)screenMousePos.y );
}


//-------------------------------------------------------------------------------------------------
// SetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosClientAbsolute( const Vector2& clientMousePos )
{
	SetMousePosClientAbsolute( (int)clientMousePos.x, (int)clientMousePos.y );
}




//-------------------------------------------------------------------------------------------------
// OnAppGainFocus
//-------------------------------------------------------------------------------------------------
void TheInputSystem::OnAppGainFocus()
{
	_doesAppHaveFocus = true;

	/// Force the cursor back to where it was when we lost focus
	::SetCursorPos( _mousePosX, _mousePosY );
}


//-------------------------------------------------------------------------------------------------
// OnAppLoseFocus
//-------------------------------------------------------------------------------------------------
void TheInputSystem::OnAppLoseFocus()
{
	_doesAppHaveFocus = false;
}


