//-------------------------------------------------------------------------------------------------
// TheMessageHandler.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheMessageHandler.hpp"
#include "TheInputSystem.hpp"
#include "TheVideoSystem.hpp"
#include "TheConsole.hpp"
#include "TheApplication.hpp"
#include "engineAssert.hpp"


//-------------------------------------------------------------------------------------------------
// Subsystem name macro (for assertions, etc)
//-------------------------------------------------------------------------------------------------
#define SUBSYSTEM_NAME "TheMessageHandler"


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC bool					TheMessageHandler::_isInitialized = false;


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheMessageHandler::TheMessageHandler()
{
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheMessageHandler::~TheMessageHandler()
{
	/// Make sure the subsystem was shut down before being destroyed
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was destroyed while running; Shutdown() should be called first!" );
}


//-------------------------------------------------------------------------------------------------
// Startup
//-------------------------------------------------------------------------------------------------
void TheMessageHandler::Startup()
{
	theConsole.Printf( "::: Initialization : Message Handler (starting)\n" );

	/// Make sure the subsystem doesn't get initialized twice
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was initialized twice; either a second "SUBSYSTEM_NAME" was started or Startup() was called on an already-running system." );

	/// Set initialization complete
	_isInitialized = true;
	theConsole.Printf( "::: Initialization : Message Handler (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheMessageHandler::Shutdown()
{
	theConsole.Printf( "::: Shutdown : Message Handler (starting)\n" );

	/// De-initialize the system
	_isInitialized = false;
	theConsole.Printf( "::: Shutdown : Message Handler (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// HandleWin32Message
// 
// WinProc() - the windows message handling procedure - calls this function before doing
//	anything else to allow the application an opportunity to handle each Win32 WM_XXXX-style
//	message as it comes in.  Return value is true if the message was handled (or trapped)
//	internally by the application; returns false if not (in which case the message
//	will be passed on to Windows for further processing via DefWindowProc()).
//-------------------------------------------------------------------------------------------------
const bool TheMessageHandler::HandleWin32Message(
	const HWND hWnd, const UINT messageID, const WPARAM wParam, const LPARAM lParam )
{
	ASSERT_MSG0( _isInitialized, SUBSYSTEM_NAME"::HandleWin32Message() was called while "SUBSYSTEM_NAME" was uninitialized; call Startup() first." );

	/// FIXME: not yet implemented!
	UNUSED_ARG hWnd;
	UNUSED_ARG wParam;
	UNUSED_ARG lParam;

	bool wasHandled = false;

	switch( messageID )
	{
		case WM_CLOSE:
			theApplication.RequestShutdown();
//			PostQuitMessage( 0 );
			wasHandled = true;
			break;
		case WM_KEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
		case 0x020A: // WM_MOUSEWHEEL
			wasHandled = theInputSystem.ProcessInputEvent( messageID, wParam, lParam );
			break;
		case WM_LBUTTONDOWN:
			wasHandled = theInputSystem.ProcessInputEvent( WM_KEYDOWN, VK_LBUTTON, 0 );
			break;
		case WM_LBUTTONUP:
			wasHandled = theInputSystem.ProcessInputEvent( WM_KEYUP, VK_LBUTTON, 0 );
			break;
		case WM_RBUTTONDOWN:
			wasHandled = theInputSystem.ProcessInputEvent( WM_KEYDOWN, VK_RBUTTON, 0 );
			break;
		case WM_RBUTTONUP:
			wasHandled = theInputSystem.ProcessInputEvent( WM_KEYUP, VK_RBUTTON, 0 );
			break;
		case WM_MBUTTONDOWN:
			wasHandled = theInputSystem.ProcessInputEvent( WM_KEYDOWN, VK_MBUTTON, 0 );
			break;
		case WM_MBUTTONUP:
			wasHandled = theInputSystem.ProcessInputEvent( WM_KEYUP, VK_MBUTTON, 0 );
			break;
		case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDBLCLK:
			wasHandled = theInputSystem.ProcessInputEvent( messageID, wParam, lParam );
			break;
		case WM_MOVE:
		{
			int newClientX = (int)(short) LOWORD( lParam );
			int newClientY = (int)(short) HIWORD( lParam );
			theVideoSystem.NotifyOfWindowMove( newClientX, newClientY );
			break;
		}
		case WM_ACTIVATEAPP:
			if( wParam == TRUE )
			{
				/// App just gained focus
				theApplication.OnAppGainFocus();
			}
			else
			{
				/// App just lost focus
				theApplication.OnAppLoseFocus();
			}
			break;
		default:
			break;
	}

	return wasHandled;
}



