//-------------------------------------------------------------------------------------------------
// TheRenderer.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_TheRenderer_
#define _include_TheRenderer_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.hpp"
#include "Rgba.hpp"
#include "Vector2.hpp"
#include "IntVector2.hpp"
#include "AABB2.hpp"
#include "LineSegment2.hpp"
#include "Image.hpp"
#include "Texture.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		TheRenderer
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class TheRenderer
{
private:
	enum
	{
		DEFAULT_FONT_HEIGHT			= 16,
		DEFAULT_FONT_PADDING_Y		= -2,
	};

	enum
	{
		CONSOLE_DISPLAY_NONE		= 0,
		CONSOLE_DISPLAY_PARTIAL		= 1,
		CONSOLE_DISPLAY_FULL		= 2,
		CONSOLE_DISPLAY_NUMBER_OF_MODES,
	};

	/// Static members
	static bool _isInitialized;

	/// Private members
	bool _isOpenGLInitialized;
	float _zBufferDepthMin;
	float _zBufferDepthMax;
	int _defaultFontListBase;	// opengl display list "name" (ID number) where ascii font starts
	int _consoleDisplayMode;
	bool _displayFPS;
	int _numConsolePartialRows;
	int _currentTexNum;

	/// Private methods
	void _InitOpenGLStates();
	void _InitFonts();
	void _RenderConsole() const;
	void _RenderFPSIndicator() const;

public:
	/// Construction, initialization, and singleton methods
	TheRenderer(); // constructor
	~TheRenderer(); // destructor

	void Startup();
	void Shutdown();

	/// Public methods
	bool IsOpenGLInitialized() const { return _isOpenGLInitialized; }

	void ClearScreenf( float redFloat, float greenFloat, float blueFloat, float alphaFloat ) const;
	void ClearScreenb( byte_t redByte, byte_t greenByte, byte_t blueByte, byte_t alphaByte ) const;
	void ClearScreen( const Rgba& rgba ) const;
	void SetViewport( const IntVector2& topLeftScreenXY, const IntVector2& size ) const;
	void SetScissor( const IntVector2& topLeftScreenXY, const IntVector2& size ) const;
	void SetViewportAndScissor( const IntVector2& topLeftScreenXY, const IntVector2& size ) const;
	void ResetViewport() const;
	void ResetScissor() const;
	void ResetViewportAndScissor() const;
	void RenderScene() const;
	void XYPrintf( float x, float y, const char* format, ... ) const;
	void XYPrintf( float x, float y, const Rgba& rgba, const char* format, ... ) const;
	void XYPrintff( float x, float y, float redFloat, float greenFloat, float blueFloat, float alphaFloat, const char* format, ... ) const;
	void XYPrintfb( byte_t x, byte_t y, byte_t redByte, byte_t greenByte, byte_t blueByte, byte_t alphaByte, const char* format, ... ) const;
	void XYPrintf( const Vector2& pos, const char* format, ... ) const;
	void XYPrintf( const Vector2& pos, const Rgba& rgba, const char* format, ... ) const;
	void XYPrintff( const Vector2& pos, float redFloat, float greenFloat, float blueFloat, float alphaFloat, const char* format, ... ) const;
	void XYPrintfb( const Vector2& pos, byte_t redByte, byte_t greenByte, byte_t blueByte, byte_t alphaByte, const char* format, ... ) const;
	void XYPrintChars( const Vector2& pos, const Rgba& rgba, const char* text, int textLength ) const;
	void ShowConsoleToggle();
	void ShowConsole( bool show );
	void ShowFpsToggle();

	void SetCoordinateSystemToScreenCoordinates2d( bool putOriginAtBottomLeft = false ) const;
	void SetCoordinateSystem2d( float left, float top, float right, float bottom ) const;
	void SetCoordinateSystem2d( const Vector2& topLeft, const Vector2& bottomRight ) const;
	void TranslateCoordinateSystem2d( const Vector2& translation );
	void TranslateCoordinateSystem2d( const Vector2& newReferencePoint, const Vector2& oldReferencePoint );
	void ScaleCoordinateSystem2d( const Vector2& scaleFactors, const Vector2& centerPoint );
	void ScaleCoordinateSystem2d( float scaleFactor, const Vector2& centerPoint );
	void ScaleCoordinateSystem2d( const Vector2& scaleFactors );
	void ScaleCoordinateSystem2d( float scaleFactor );
	void RotateCoordinateSystem2d( float radians, const Vector2& centerPoint );
	void RotateCoordinateSystem2d( float radians );
	void PushCoordinateSystem();
	void PopCoordinateSystem();

	void SetColor( const Rgba& rgba ) const;
	void SetColorf( float redFloat, float greenFloat, float blueFloat, float alphaFloat = 1.0f ) const;
	void SetColorb( byte_t redByte, byte_t greenByte, byte_t blueByte, byte_t alphaByte = (byte_t)0xff ) const;

	void DrawTexturedRect2d( int openglTexNum, const Vector2& mins, const Vector2& maxs, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRect2d( const Texture& texture, const Vector2& mins, const Vector2& maxs, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRect2d( int openglTexNum, const Vector2& mins, const Vector2& maxs, const Vector2& texCoordForMins, const Vector2& texCoordForMaxs, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRect2d( const Texture& texture, const Vector2& mins, const Vector2& maxs, const Vector2& texCoordForMins, const Vector2& texCoordForMaxs, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRectCentered2d( int openglTexNum, const Vector2& centerPoint, const Vector2& rectSize, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRectCentered2d( const Texture& texture, const Vector2& centerPoint, const Vector2& rectSize, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRectCentered2d( int openglTexNum, const Vector2& centerPoint, const Vector2& rectSize, const Vector2& texCoordForMins, const Vector2& texCoordForMaxs, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );
	void DrawTexturedRectCentered2d( const Texture& texture, const Vector2& centerPoint, const Vector2& rectSize, const Vector2& texCoordForMins, const Vector2& texCoordForMaxs, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true, bool invertTexCoordYs = true );

	void SetCurrentTexture( int openglTexNum, bool forceBind = false );
	void SetCurrentTexture( const Texture& texture, bool forceBind = false );
	void DrawVector2d( const Vector2& start, const Vector2& end, const Rgba& startColor = Rgba::WHITE, const Rgba& endColor = Rgba::WHITE );
	void DrawVector2d( const LineSegment2& lineSegment, const Rgba& startColor = Rgba::WHITE, const Rgba& endColor = Rgba::WHITE );
	void DrawCircle2d( const Vector2& centerPoint, float radius, const Rgba& rgba = Rgba::WHITE, int numSides = -1 );
	void DrawAABB2d( const AABB2& bounds, const Rgba& color = Rgba::WHITE, bool isFilled = false );
	void DrawAABB2d( const Vector2& mins, const Vector2& maxs, const Rgba& color = Rgba::WHITE, bool isFilled = false );
};



//-------------------------------------------------------------------------------------------------
// Global singleton instance
//-------------------------------------------------------------------------------------------------
extern TheRenderer theRenderer;


#endif // _include_TheRenderer_
