//-------------------------------------------------------------------------------------------------
// TheSoundSystem.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheSoundSystem.hpp"
#include "TheConsole.hpp"
#include "engineAssert.hpp"


//-------------------------------------------------------------------------------------------------
// Subsystem name macro (for assertions, etc)
//-------------------------------------------------------------------------------------------------
#define SUBSYSTEM_NAME "TheSoundSystem"


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC bool				TheSoundSystem::_isInitialized		= false;


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheSoundSystem::TheSoundSystem()
{
}

//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheSoundSystem::~TheSoundSystem()
{
	/// Make sure the subsystem was shut down before being destroyed
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was destroyed while running; Shutdown() should be called first!" );
}

//-------------------------------------------------------------------------------------------------
// Startup
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::Startup()
{
	theConsole.Printf( "::: Initialization : Sound (starting)\n" );

	/// Make sure the subsystem doesn't get initialized twice
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was initialized twice; either a second "SUBSYSTEM_NAME" was started or Startup() was called on an already-running system." );

	// don't allow hardware channels at all, since they don't show up on the spectrum analysis
	FSOUND_SetMinHardwareChannels( 0 );
	AbortOnFmodError();

	FSOUND_SetMaxHardwareChannels( 0 ); 
	AbortOnFmodError();

	FSOUND_Init( 44100, 1023, FSOUND_INIT_USEDEFAULTMIDISYNTH );
	AbortOnFmodError();

	/// Set initialization complete
	_isInitialized = true;
	theConsole.Printf( "::: Initialization : Sound (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::Shutdown()
{
	theConsole.Printf( "::: Shutdown : Sound (starting)\n" );

	// Free any allocated samples
	UnregisterAllSamples();

	// Stop fmod
    FSOUND_Close();

	/// De-initialize the system
	_isInitialized = false;
	theConsole.Printf( "::: Shutdown : Sound (finished)\n" );
}

//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::Update()
{
	// find any expired sound instances and forget about them
	for( SampleContainer::iterator iter = m_samples.begin(); iter != m_samples.end(); ++iter )
	{
		SampleInfo& si = iter->second;

		// find instances that match and stop them
		for( ;; )
		{
			SampleInfo::PlaybackInstanceContainer::iterator firstIter = si.playbackInstances.begin();
			if( firstIter == si.playbackInstances.end() )
			{
				break;
			}

			SampleInfo::PlaybackInstance& instance = firstIter->second;
			if( FSOUND_IsPlaying( instance.fmodChannel ) )
			{
				// stop at first instance that is playing, as all instances after it should
				// so be playing since the instances are sorted by the time they should
				// be finished playing
				break;
			}
		
			// done playing, remove it
			si.playbackInstances.erase( firstIter );
			--si.currentPlaybackCount;
		}
	}
}

//-------------------------------------------------------------------------------------------------
// AbortOnFmodError
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::AbortOnFmodError()
{
	const int fmodError = FSOUND_GetError();
	if( FMOD_ERR_NONE == fmodError )
	{
		return;
	}

	const char* const errorString = FMOD_ErrorString( fmodError );
	VALIDATE_MSG0( FMOD_ERR_NONE == fmodError, errorString );
}


//-------------------------------------------------------------------------------------------------
// RegisterSample
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::RegisterSample( const int soundNumber, const std::string& name )
{
	SampleContainer::iterator i = m_samples.find( soundNumber );
	if( i != m_samples.end() )
	{
		UnregisterSample( soundNumber );
	}

	SampleInfo& s = m_samples[ soundNumber ];
	
	s.sample = FSOUND_Sample_Load( FSOUND_UNMANAGED, name.c_str(), FSOUND_2D, 0, 0 );
	VALIDATE_MSG1( m_samples[ soundNumber ].sample != NULL, "Failed to load sound %s", name.c_str() );
	s.currentPlaybackCount = 0;
	s.maxPlaybackCount = 255;
}

//-------------------------------------------------------------------------------------------------
// UnregisterSample
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::UnregisterSample( const int soundNumber )
{
	SampleContainer::iterator i = m_samples.find( soundNumber );
	if( i != m_samples.end() )
	{
		FSOUND_Sample_Free( i->second.sample );
		m_samples.erase( i );
	}
	else
	{
		theConsole.Printf( CHAN_SOUND, "WARNING: tried to unregister non-existent sound #%i\n", soundNumber );
	}
}

//-------------------------------------------------------------------------------------------------
// UnregisterAllSamples
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::UnregisterAllSamples()
{
	while( !m_samples.empty() )
	{
		SampleContainer::iterator i = m_samples.begin();
		FSOUND_Sample_Free( i->second.sample );
		m_samples.erase( i );
	}
}

//-------------------------------------------------------------------------------------------------
// SetConcurrentPlaybackLimit
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::SetConcurrentPlaybackLimit( const dword_t soundNumber, const byte_t maxPlaybackCount )
{
	ASSERT_MSG0( maxPlaybackCount > 0, SUBSYSTEM_NAME"::SetConcurrentPlaybackLimit() : <maxPlaybackCount> must be a positive nonzero number." );
	SampleContainer::iterator i = m_samples.find( soundNumber );
	if( i != m_samples.end() )
	{
		i->second.maxPlaybackCount = maxPlaybackCount;
	}
	else
	{
		theConsole.Printf( CHAN_SOUND, "WARNING: tried to set playback limit for non-existent sound #%i\n", soundNumber );
	}
}

//-------------------------------------------------------------------------------------------------
// PlaySample
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::PlaySample(
	const dword_t soundNumber, 
	const dword_t instanceId,
	const PlaybackLoopMode loopMode,
	const float volume, 
	const float speed, 
	const float pan )
{
	SampleContainer::iterator sampleIter = m_samples.find( soundNumber );
	if( sampleIter == m_samples.end() )
	{
		ASSERT_MSG1( sampleIter != m_samples.end(), SUBSYSTEM_NAME"::PlaySample() : tried to play non-existent sound #%i\n", soundNumber );
		theConsole.Printf( CHAN_SOUND, "WARNING: tried to play non-existent sound #%i\n", soundNumber );
		return;
	}
	ASSERT_MSG1( sampleIter->second.sample, SUBSYSTEM_NAME"::PlaySample() : sample data was NULL for sound #%i", soundNumber );
	ASSERT_MSG2( volume >= 0.0f, SUBSYSTEM_NAME"::PlaySample() : volume (%f) for sound #%i was out of range [0,1].", volume, soundNumber );
	ASSERT_MSG2( volume <= 1.0f, SUBSYSTEM_NAME"::PlaySample() : volume (%f) for sound #%i was out of range [0,1].", volume, soundNumber );
	ASSERT_MSG1( speed != 0.0f, SUBSYSTEM_NAME"::PlaySample() : playback speed for sound #%i was zero.", soundNumber );
	ASSERT_MSG2( pan >= -1.0f, SUBSYSTEM_NAME"::PlaySample() : pan (%f) for sound #%i was out of range [-1,1].", pan, soundNumber );
	ASSERT_MSG2(	pan <= 1.0f, SUBSYSTEM_NAME"::PlaySample() : pan (%f) for sound #%i was out of range [-1,1].", pan, soundNumber );

	SampleInfo& si = sampleIter->second;
	FSOUND_SAMPLE* const fmodSample = si.sample;

	// if there are too many concurrent playbacks of this sound, restart the oldest one
	int fmodChannel = -1;
	if( si.currentPlaybackCount == si.maxPlaybackCount )
	{
		ASSERT_MSG1( !si.playbackInstances.empty(), SUBSYSTEM_NAME"::PlaySample() : playbackInstances list for sound #%i was unexpectedly empty.", soundNumber );
		SampleInfo::PlaybackInstanceContainer::iterator firstInstanceIter = si.playbackInstances.begin();
		const SampleInfo::PlaybackInstance& pi = firstInstanceIter->second;
		FSOUND_SetCurrentPosition( pi.fmodChannel, 0 ); // restart
		fmodChannel = pi.fmodChannel;
		
		// forget about this instance since it has ended right now
		si.playbackInstances.erase( firstInstanceIter );
		
		// don't change currentPlaybackCount since a sound is being replaced
	}
	else
	{
		// play the sample normally
		fmodChannel = FSOUND_PlaySound( FSOUND_FREE, fmodSample );
		if( fmodChannel == -1 )
		{
			theConsole.Printf( CHAN_SOUND, "WARNING: tried to play sound #%i, but no free sound channels\n", soundNumber );
			return;
		}

		// successfully played, increment counter
		++si.currentPlaybackCount;
	}

	// let the sample remember when this instance of it will end
	const dword_t expirationTime = CalcExpirationTime( si, loopMode, speed );
	SampleInfo::PlaybackInstance pi;
	pi.fmodChannel	= fmodChannel;
	pi.id			= instanceId;
	pi.loopMode		= loopMode;
	si.playbackInstances.insert( std::make_pair( expirationTime, pi ) );

	// modify the channel's parameters to match the requested parameters
	FSOUND_SetLoopMode( fmodChannel, loopMode );

	FSOUND_SetVolume( fmodChannel, (int)(volume * 255) ); 

	int freq = 0;
	FSOUND_Sample_GetDefaults( fmodSample, &freq, NULL, NULL, NULL );
	FSOUND_SetFrequency( fmodChannel, (int)( freq * speed ) );

	// see documentation for FSOUND_SetPan RE: stereo sample panning
	// "FSOUND_STEREOPAN. This is meant for stereo samples, but will work on mono 
	// samples as well. It makes both left and right FULL volume instead of 50/50
	// as middle panning does. See remarks section for more information on this."
	int newPan;
	const dword_t sampleMode = FSOUND_Sample_GetMode( fmodSample );
	if( (pan == 0.0f) && ( FSOUND_STEREO & sampleMode ) )
	{
		newPan = FSOUND_STEREOPAN;
	}
	else
	{
		newPan = (int)(pan * 128) + 128;
		if( newPan > 255 ) { newPan = 255; }
	}
	FSOUND_SetPan( fmodChannel, newPan );
}

//-------------------------------------------------------------------------------------------------
// StopSample
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::StopSample( const dword_t soundNumber, const dword_t instanceId )
{
	// find the sample
	SampleContainer::iterator sampleIter = m_samples.find( soundNumber );
	if( sampleIter == m_samples.end() )
	{
		ASSERT_MSG1( sampleIter != m_samples.end(), SUBSYSTEM_NAME"::StopSample() : tried to stop non-existent sound #%i", soundNumber );
		theConsole.Printf( CHAN_SOUND, "WARNING: tried to stop non-existent sound #%i\n", soundNumber );
		return;
	}
	SampleInfo& si = sampleIter->second;

	// find instances that match and stop them
	SampleInfo::PlaybackInstanceContainer::iterator instanceIter = si.playbackInstances.begin();
	SampleInfo::PlaybackInstanceContainer::iterator endInstanceIter = si.playbackInstances.end();
	SampleInfo::PlaybackInstanceContainer::iterator nextInstanceIter = instanceIter;
	while( instanceIter != endInstanceIter )
	{
		++nextInstanceIter;

		SampleInfo::PlaybackInstance& instance = instanceIter->second;
		if( instance.id == instanceId )
		{
			FSOUND_StopSound( instance.fmodChannel );
			si.playbackInstances.erase( instanceIter );
			--si.currentPlaybackCount;
		}
		
		instanceIter = nextInstanceIter;
	}
}

//-------------------------------------------------------------------------------------------------
// StopSample
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::StopSample( const dword_t soundNumber )
{
	// find the sample
	SampleContainer::iterator sampleIter = m_samples.find( soundNumber );
	if( sampleIter == m_samples.end() )
	{
		ASSERT_MSG1( sampleIter != m_samples.end(), SUBSYSTEM_NAME"::StopSample() : tried to stop non-existent sound #%i", soundNumber );
		theConsole.Printf( CHAN_SOUND, "WARNING: tried to stop non-existent sound #%i\n", soundNumber );
		return;
	}
	SampleInfo& si = sampleIter->second;

	// stop all instances
	while( !si.playbackInstances.empty() )
	{
		SampleInfo::PlaybackInstanceContainer::iterator iter = si.playbackInstances.begin();
		SampleInfo::PlaybackInstance& instance = iter->second;
		FSOUND_StopSound( instance.fmodChannel );
		si.playbackInstances.erase( iter );
		--si.currentPlaybackCount;
	}

}

//-------------------------------------------------------------------------------------------------
// StopPlayingAllSamples
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::StopPlayingAllSamples()
{
	for( SampleContainer::iterator iter = m_samples.begin(); iter != m_samples.end(); ++iter )
	{
		StopSample( iter->first );
	}
}

//-------------------------------------------------------------------------------------------------
// Pause
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::Pause( const dword_t soundNumber, const dword_t instanceId )
{
	soundNumber, instanceId; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::Pause() not implemented!" )
}


//-------------------------------------------------------------------------------------------------
// Unpause
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::Unpause( const dword_t soundNumber, const dword_t instanceId )
{
	soundNumber, instanceId; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::Unpause() not implemented!" )
}


//-------------------------------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::ChangeVolume( 
	const dword_t soundNumber, 
	const dword_t instanceId, 
	const float targetVolume, 
	const float fadeTimeSeconds )
{
	soundNumber, instanceId, targetVolume, fadeTimeSeconds; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::ChangeVolume() not implemented!" )
}

//-------------------------------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::ChangePan( 
	const dword_t soundNumber, 
	const dword_t instanceId, 
	const float targetPan, 
	const float fadeTimeSeconds )
{
	soundNumber, instanceId, targetPan, fadeTimeSeconds; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::ChangePan() not implemented!" )
}

//-------------------------------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::ChangeSpeed(
	const dword_t soundNumber, 
	const dword_t instanceId, 
	const float targetSpeed, 
	const float fadeTimeSeconds )
{
	soundNumber, instanceId, targetSpeed, fadeTimeSeconds; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::ChangeSpeed() not implemented!" )
}


//-------------------------------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::QuerySampleStatus( 
	const dword_t soundNumber,
	const dword_t instanceId,
	bool& isFinished,
	float& currentLeftVolume,
	float& currentRightVolume,
	float& currentSpeed,
	float& currentPan,
	float& percentPlayed )
{
	soundNumber, instanceId, isFinished, currentLeftVolume, currentRightVolume, currentSpeed, currentPan, percentPlayed; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::QuerySampleStatus() not implemented!" )
}


//-------------------------------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::QueryGlobalStats(
	int& numPlayingSamples,
	int& numLoadedSamples,
	float& cpuUsePercent,
	unsigned int& currentBytesAllocated,
	unsigned int& maxBytesAllocated )
{
	numPlayingSamples, numLoadedSamples, cpuUsePercent, currentBytesAllocated, maxBytesAllocated; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::QueryGlobalStats() not implemented!" )
}


//-------------------------------------------------------------------------------------------------
// EnableSpectrumData
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::EnableSpectrumData()
{
	FSOUND_DSPUNIT* fft = FSOUND_DSP_GetFFTUnit();
	FSOUND_DSP_SetActive( fft, TRUE );
}


//-------------------------------------------------------------------------------------------------
// DisableSpectrumData
//-------------------------------------------------------------------------------------------------
void TheSoundSystem::DisableSpectrumData()
{
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::DisableSpectrumData() not implemented!" )
}


//-------------------------------------------------------------------------------------------------
// GetSpectrumData512
//-------------------------------------------------------------------------------------------------
const float* TheSoundSystem::GetSpectrumData512()
{ 
	return FSOUND_DSP_GetSpectrum();
}

//-------------------------------------------------------------------------------------------------
// CalcExpirationTime
//-------------------------------------------------------------------------------------------------
dword_t TheSoundSystem::CalcExpirationTime( const SampleInfo& si, const PlaybackLoopMode loop, const float speed )
{
	speed; // placate the compiler for now
#pragma message( __FILE__ "(1) : WARNING: TheSoundSystem::CalcExpirationTime() - unused parameter \"speed\"" )

	if( loop != LOOP_DISABLE )
	{
		return 0xffffffff; // "never" expires
	}
	else
	{
		const unsigned int numSamples = FSOUND_Sample_GetLength( si.sample );
		int samplesPerSec;
		FSOUND_Sample_GetDefaults( si.sample, &samplesPerSec, 0, 0, 0 );
		const dword_t timeToPlay = (dword_t)( 1000.0 * ((double)numSamples / (double)samplesPerSec) );
		return timeToPlay;			
	}
}
