//-------------------------------------------------------------------------------------------------
// Vector4.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_Vector4_
#define _include_Vector4_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.hpp"
#include "Vector3.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		Vector4
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class Vector4
{
public:
	float x; // these are an exception to the "no-public-members" rule of thumb
	float y; // ...for the sake of clarity and brevity in use throughout the code
	float z;
	float t;

	static const Vector4 ZERO;
	static const Vector4 ONE;

	/// Construction / Destruction
	~Vector4(); // destructor
	Vector4(); // default constructor
	Vector4( const Vector4& vec ); // copy constructor
	Vector4( const Vector3& vec ); // construct from 3d vector
	Vector4( const Vector2& vec ); // construct from 2d vector
	Vector4( const float x, const float y, const float z, const float t );

	/// Interface methods
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const float CalcLengthXYZ() const;
	const float Normalize();
	const float SetLength( float newLength );

	void SetXYZT( const float new_x, const float new_y, const float new_z, const float new_t );
	void GetXYZT( float& get_x, float& get_y, float& get_z, float& get_t ) const;
	const Vector4& Clamp( const Vector4& min, const Vector4& max );
	const float CalcMaxComponent() const;
	const float CalcMinComponent() const;
	const Vector4 Lerp( const Vector4& rhs, const float rhs_fraction ) const;

	/// Operators
	const Vector4& operator = ( const Vector4& rhs );
	const Vector4& operator += ( const Vector4& rhs );
	const Vector4& operator -= ( const Vector4& rhs );
	const Vector4& operator *= ( const float scale );
	const Vector4& operator /= ( const float divisor );

	const Vector4 operator - () const;
	const Vector4 operator - ( const Vector4& rhs ) const;
	const Vector4 operator + ( const Vector4& rhs ) const;
	const Vector4 operator * ( const float scale ) const;
	const float operator | ( const Vector4& rhs ) const; // dot product

	const float operator [] ( const int index ) const;
	float& operator [] ( const int index );

	const bool operator == ( const Vector4& rhs ) const;
	const bool operator != ( const Vector4& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const Vector4 operator * ( const float scale, const Vector4& rhs );
const bool EpsilonEqual( const Vector4& lhs, const Vector4& rhs, const float epsilon );


//-------------------------------------------------------------------------------------------------
// Vector4::~Vector4()
//-------------------------------------------------------------------------------------------------
inline Vector4::~Vector4()
{
}


//-------------------------------------------------------------------------------------------------
// Vector4::Vector4() - default constructor
//-------------------------------------------------------------------------------------------------
inline Vector4::Vector4()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// Vector4::Vector4( const Vector4& vec ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline Vector4::Vector4( const Vector4& vec )
:
x( vec.x ),
y( vec.y ),
z( vec.z ),
t( vec.t )
{
}


//-------------------------------------------------------------------------------------------------
// Vector4::Vector4( const Vector3& vec ) - construct from 3d vector
//-------------------------------------------------------------------------------------------------
inline Vector4::Vector4( const Vector3& vec )
:
x( vec.x ),
y( vec.y ),
z( vec.z ),
t( 0.0f )
{
}


//-------------------------------------------------------------------------------------------------
// Vector4::Vector4( const Vector2& vec ) - construct from 2d vector
//-------------------------------------------------------------------------------------------------
inline Vector4::Vector4( const Vector2& vec )
:
x( vec.x ),
y( vec.y ),
z( 0.0f ),
t( 0.0f )
{
}


//-------------------------------------------------------------------------------------------------
// Vector4( float, float, float, float )
//-------------------------------------------------------------------------------------------------
inline Vector4::Vector4( const float x, const float y, const float z, const float t )
:
x( x ),
y( y ),
z( z ),
t( t )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const Vector4& Vector4::operator = ( const Vector4& rhs )
{
	if( this == &rhs )
		return *this;

	x = rhs.x;
	y = rhs.y;
	z = rhs.z;
	t = rhs.t;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const Vector4& Vector4::operator += ( const Vector4& rhs )
{
	x += rhs.x;
	y += rhs.y;
	z += rhs.z;
	t += rhs.t;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const Vector4& Vector4::operator -= ( const Vector4& rhs )
{
	x -= rhs.x;
	y -= rhs.y;
	z -= rhs.z;
	t -= rhs.t;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const Vector4& Vector4::operator *= ( const float scale )
{
	x *= scale;
	y *= scale;
	z *= scale;
	t *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const Vector4& Vector4::operator /= ( const float divisor )
{
	if( divisor )
	{
		float invDivisor = 1.0f / divisor;
		x *= invDivisor;
		y *= invDivisor;
		z *= invDivisor;
		t *= invDivisor;
	}
	else
	{
		x = 0.0f;
		y = 0.0f;
		z = 0.0f;
		t = 0.0f;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const Vector4 Vector4::operator - () const
{
	return Vector4( -x, -y, -z, -t );
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const Vector4 Vector4::operator - ( const Vector4& rhs ) const
{
	return( Vector4( x - rhs.x, y - rhs.y, z - rhs.z, y - rhs.t ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const Vector4 Vector4::operator + ( const Vector4& rhs ) const
{
	return( Vector4( x + rhs.x, y + rhs.y, z + rhs.z, t + rhs.t ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const Vector4 Vector4::operator * ( const float scale ) const
{
	return( Vector4( x * scale, y * scale, z * scale, t * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator | (dot product)
//-------------------------------------------------------------------------------------------------
inline const float Vector4::operator | ( const Vector4& rhs ) const
{
	return( (x * rhs.x) + (y * rhs.y) + (z * rhs.z) + (t * rhs.t) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline const float Vector4::operator [] ( const int index ) const
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 4 ), "Vector4[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline float& Vector4::operator [] ( const int index )
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 4 ), "Vector4[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool Vector4::operator == ( const Vector4& rhs ) const
{
	return( (x == rhs.x) && (y == rhs.y) && (z == rhs.z) && (t == rhs.t) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool Vector4::operator != ( const Vector4& rhs ) const
{
	return( (x != rhs.x) || (y != rhs.y) || (z != rhs.z) || (t != rhs.t) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float Vector4::CalcLengthSquared() const
{
	return( x*x + y*y + z*z + t*t );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float Vector4::CalcLength() const
{
	return( (float) sqrt( x*x + y*y + z*z + t*t ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthXY
//-------------------------------------------------------------------------------------------------
inline const float Vector4::CalcLengthXYZ() const
{
	return( (float) sqrt( x*x + y*y + z*z ) );
}


//-------------------------------------------------------------------------------------------------
// Normalize
//-------------------------------------------------------------------------------------------------
inline const float Vector4::Normalize()
{
	float length = CalcLength();
	if( length )
	{
		float invLength = 1.0f / length;
		x *= invLength;
		y *= invLength;
		z *= invLength;
		t *= invLength;
	}

	return( length );
}


//-------------------------------------------------------------------------------------------------
// SetLength
//-------------------------------------------------------------------------------------------------
inline const float Vector4::SetLength( float newLength )
{
	float oldLength = Normalize();
	(*this) *= newLength;
	return oldLength;
}


//-------------------------------------------------------------------------------------------------
// SetXYZT
//-------------------------------------------------------------------------------------------------
inline void Vector4::SetXYZT( const float new_x, const float new_y, const float new_z, const float new_t )
{
	x = new_x;
	y = new_y;
	z = new_z;
	t = new_t;
}


//-------------------------------------------------------------------------------------------------
// GetXYZT
//-------------------------------------------------------------------------------------------------
inline void Vector4::GetXYZT( float& get_x, float& get_y, float& get_z, float& get_t ) const
{
	get_x = x;
	get_y = y;
	get_z = z;
	get_t = t;
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline const Vector4& Vector4::Clamp( const Vector4& min, const Vector4& max )
{
	/// Clamp X component
	if( x > max.x )
		x = max.x;
	else if( x < min.x )
		x = min.x;

	/// Clamp Y component
	if( y > max.y )
		y = max.y;
	else if( y < min.y )
		y = min.y;

	/// Clamp Z component
	if( z > max.z )
		z = max.z;
	else if( z < min.z )
		z = min.z;

	/// Clamp t component
	if( t > max.t )
		t = max.t;
	else if( t < min.t )
		t = min.t;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// CalcMaxComponent
//-------------------------------------------------------------------------------------------------
inline const float Vector4::CalcMaxComponent() const
{
	/// Assume x is the largest
	float max = x;

	/// Check if y is bigger
	if( y > max )
		max = y;

	/// Check if z is bigger
	if( z > max )
		max = z;

	/// Check if t is bigger
	if( t > max )
		max = t;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// CalcMinComponent
//-------------------------------------------------------------------------------------------------
inline const float Vector4::CalcMinComponent() const
{
	/// Assume x is the smallest
	float min = x;

	/// Check if y is smaller
	if( y < min )
		min = y;

	/// Check if z is smaller
	if( z < min )
		min = z;

	/// Check if t is smaller
	if( t < min )
		min = t;

	return( min );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const Vector4 Vector4::Lerp( const Vector4& rhs, const float rhs_fraction ) const
{
	return Vector4( x + rhs_fraction * (rhs.x - x),
					y + rhs_fraction * (rhs.y - y),
					z + rhs_fraction * (rhs.z - z),
					t + rhs_fraction * (rhs.t - t));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, Vector4)
//-------------------------------------------------------------------------------------------------
inline const Vector4 operator * ( const float scale, const Vector4& rhs )
{
	return( Vector4( scale * rhs.x, scale * rhs.y, scale * rhs.z, scale * rhs.t ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const Vector4& lhs, const Vector4& rhs, const float epsilon )
{
	return( fabs( lhs.x - rhs.x ) < epsilon &&
			fabs( lhs.y - rhs.y ) < epsilon &&
			fabs( lhs.z - rhs.z ) < epsilon &&
			fabs( lhs.t - rhs.t ) < epsilon );
}


#endif // _include_Vector4_
