//-------------------------------------------------------------------------------------------------
// collision.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
// 
// Common suite of collision detection functions.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "collision.hpp"



//-------------------------------------------------------------------------------------------------
// CalcLineCircleIntersects
// 
// Finds the x,y coordinates of the intersection(s) of the line
// 
//		<lineEquationA>x + <lineEquationB>y + <lineEquationC> = 0
// 
// with a circle of radius <circleRadius> centered at <circleCenter>.
// 
// Returns the number of real intersection solutions, which may be 0, 1, or 2.
// If return value is 0: no solution, <intersectA> and <intersectB> are invalid
// If return value is 1: one solution, only <intersectA> is valid
// If return value is 2: two solutions, <intersectA> and <intersectB> are valid
// 
// Essentially, this finds the intersection between the following two equations:
// 
// (x - h)^2 + (y - k)^2 = R^2     // R = circle radius; (h,k) is circle center
// 
//		and
// 
// Ax + By + C = 0		// Standard-form line equation
//-------------------------------------------------------------------------------------------------
int CalcLineCircleIntersects( const Vector2& circleCenter, double circleRadius, double lineEquationA, double lineEquationB, double lineEquationC, Vector2& intersectA, Vector2& intersectB )
{
	/// Determine if x or y is likely to produce the best solution(s)
	if( fabs( lineEquationA ) < fabs( lineEquationB ) )
	{
		/// Calculate quadratic equation coefficients, solving for x
		double lineEquationBSquared = (lineEquationB * lineEquationB);
		double twoLineEquationBCircleY = (2.0 * lineEquationB * circleCenter.y );
		double A = lineEquationBSquared + (lineEquationA * lineEquationA);
		double B = (2.0 * lineEquationA * lineEquationC) + (lineEquationA * twoLineEquationBCircleY) - (2.0 * lineEquationBSquared * circleCenter.x);
		double C =
			+ (lineEquationBSquared * circleCenter.x * circleCenter.x)
			+ (lineEquationBSquared * circleCenter.y * circleCenter.y)
			- (lineEquationBSquared * circleRadius * circleRadius)
			+ (lineEquationC * twoLineEquationBCircleY)
			+ (lineEquationC * lineEquationC);

		int numberOfSolutions = QuadraticEquation( A, B, C, intersectA.x, intersectB.x );
		if( numberOfSolutions == 0 )
		{
			return 0;
		}
		else if( numberOfSolutions == 1 )
		{
			intersectA.y = (float)( (-lineEquationC - (lineEquationA * intersectA.x)) / lineEquationB );
			return 1;
		}
		else
		{
			double oneOverLineEquationB = 1.0 / lineEquationB;
			intersectA.y = (float)( (-lineEquationC - (lineEquationA * intersectA.x)) * oneOverLineEquationB );
			intersectB.y = (float)( (-lineEquationC - (lineEquationA * intersectB.x)) * oneOverLineEquationB );
			return 2;
		}
	}
	else
	{
		/// Calculate quadratic equation coefficients, solving for y
		double lineEquationASquared = (lineEquationA * lineEquationA);
		double twoLineEquationACircleX = (2.0 * lineEquationA * circleCenter.x );
		double A = lineEquationASquared + (lineEquationB * lineEquationB);
		double B = (2.0 * lineEquationB * lineEquationC) + (lineEquationB * twoLineEquationACircleX) - (2.0 * lineEquationASquared * circleCenter.y);
		double C =
			+ (lineEquationASquared * circleCenter.y * circleCenter.y)
			+ (lineEquationASquared * circleCenter.x * circleCenter.x)
			- (lineEquationASquared * circleRadius * circleRadius)
			+ (lineEquationC * twoLineEquationACircleX)
			+ (lineEquationC * lineEquationC);

		int numberOfSolutions = QuadraticEquation( A, B, C, intersectA.y, intersectB.y );
		if( numberOfSolutions == 0 )
		{
			return 0;
		}
		else if( numberOfSolutions == 1 )
		{
			intersectA.x = (float)( (-lineEquationC - (lineEquationB * intersectA.y)) / lineEquationA );
			return 1;
		}
		else
		{
			double oneOverLineEquationA = 1.0 / lineEquationA;
			intersectA.x = (float)( (-lineEquationC - (lineEquationB * intersectA.y)) * oneOverLineEquationA );
			intersectB.x = (float)( (-lineEquationC - (lineEquationB * intersectB.y)) * oneOverLineEquationA );
			return 2;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// CalcLineCircleIntersects
// 
// Finds the x,y coordinates of the intersection(s) of the line
// 
//		<lineEquationA>x + <lineEquationB>y + <lineEquationC> = 0
// 
// with a circle of radius <circleRadius> centered at the origin.
// 
// Returns the number of real intersection solutions, which may be 0, 1, or 2.
// If return value is 0: no solution, <intersectA> and <intersectB> are invalid
// If return value is 1: one solution, only <intersectA> is valid
// If return value is 2: two solutions, <intersectA> and <intersectB> are valid
// 
// Essentially, this finds the intersection between the following two equations:
// 
// (x)^2 + (y)^2 = R^2     // R = circle radius, circle is centered at the origin
// 
//		and
// 
// Ax + By + C = 0		// Standard-form line equation
//-------------------------------------------------------------------------------------------------
int CalcLineCircleIntersects( double circleRadius, double lineEquationA, double lineEquationB, double lineEquationC, Vector2& intersectA, Vector2& intersectB )
{
	/// Determine if x or y is likely to produce the best solution(s)
	if( fabs( lineEquationA ) < fabs( lineEquationB ) )
	{
		/// Calculate quadratic equation coefficients, solving for x
		double lineEquationBSquared = (lineEquationB * lineEquationB);
		double A = lineEquationBSquared + (lineEquationA * lineEquationA);
		double B = (2.0 * lineEquationA * lineEquationC);
		double C = (lineEquationC * lineEquationC) - (lineEquationBSquared * circleRadius * circleRadius);

		int numberOfSolutions = QuadraticEquation( A, B, C, intersectA.x, intersectB.x );
		if( numberOfSolutions == 0 )
		{
			return 0;
		}
		else if( numberOfSolutions == 1 )
		{
			intersectA.y = (float)( (-lineEquationC - (lineEquationA * intersectA.x)) / lineEquationB );
			return 1;
		}
		else
		{
			double oneOverLineEquationB = 1.0 / lineEquationB;
			intersectA.y = (float)( (-lineEquationC - (lineEquationA * intersectA.x)) * oneOverLineEquationB );
			intersectB.y = (float)( (-lineEquationC - (lineEquationA * intersectB.x)) * oneOverLineEquationB );
			return 2;
		}
	}
	else
	{
		/// Calculate quadratic equation coefficients, solving for y
		double lineEquationASquared = (lineEquationA * lineEquationA);
		double A = lineEquationASquared + (lineEquationB * lineEquationB);
		double B = (2.0 * lineEquationB * lineEquationC);
		double C = (lineEquationC * lineEquationC) - (lineEquationASquared * circleRadius * circleRadius);

		int numberOfSolutions = QuadraticEquation( A, B, C, intersectA.y, intersectB.y );
		if( numberOfSolutions == 0 )
		{
			return 0;
		}
		else if( numberOfSolutions == 1 )
		{
			intersectA.x = (float)( (-lineEquationC - (lineEquationB * intersectA.y)) / lineEquationA );
			return 1;
		}
		else
		{
			double oneOverLineEquationA = 1.0 / lineEquationA;
			intersectA.x = (float)( (-lineEquationC - (lineEquationB * intersectA.y)) * oneOverLineEquationA );
			intersectB.x = (float)( (-lineEquationC - (lineEquationB * intersectB.y)) * oneOverLineEquationA );
			return 2;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// TraceMovingDiscsCollide2d
// 
// Returns a value between 0.0f and 1.0f which represents (fractionally) how far through
//	the simultaneous movement of both discs we got before a collision occurred.  A return
//	value of 0.0f implies that the discs were overlapping to begin with; a value of 1.0f
//	implies that no collision occurred.
// 
// FIXME: a faster solution might be:
//	. Give all motion to disc1; give all radius to disc2
//	S = disc1 move start
//	E = disc1 move end
//	D = disc1 move displacement
//	R = combined radii of discs
//	P (unknown) = point of intersection
//	t (unknown) = collision fraction, where 0.0 = collision at S, 1.0 = collision at E
// 
//	. and solve the quadratic:
//	A = D dot D
//	B = 2( S dot D )
//	C = ( S dot S ) - (R*R)
//	t = quadratic( A, B, C )
//-------------------------------------------------------------------------------------------------
float TraceMovingDiscsCollide2d( const Vector2& disc1pos, const Vector2& disc1move, float disc1radius,
								 const Vector2& disc2pos, const Vector2& disc2move, float disc2radius )
{
	/// Transform position and velocity into disc2's frame of reference
	const Vector2 disc1posRelative = disc1pos - disc2pos;
	const Vector2 disc1moveRelative = disc1move - disc2move;
	const LineSegment2 disc1sweep( disc1posRelative, disc1moveRelative );

	/// Transform disc1 to a singularity; disc2 gets the combined radius
	const float combinedRadius = disc1radius + disc2radius;

	/// Solve line-verses-circle quadratic
	Vector2 intersectA, intersectB;
	int numSolutions = CalcLineCircleIntersects( combinedRadius, disc1sweep, intersectA, intersectB );
	if( numSolutions == 0 )
	{
		return 1.0f; // no valid collision points
	}
	else if( numSolutions == 1 )
	{
		/// Check if point occurs before the line segment start
		const float dotA = (intersectA - disc1posRelative) | disc1moveRelative;
		if( dotA < 0.0f )
		{
			return 1.0f; // no valid collision points
		}

		/// Check if point occurs after line segment end
		const float moveRelativeSquared = disc1moveRelative.CalcLengthSquared();
		if( dotA > moveRelativeSquared )
		{
			return 1.0f; // no valid collision points
		}

		/// Determine exactly where on the line segment the point occurred
		return( dotA / moveRelativeSquared );
	}
	else // numSolutions == 2
	{
		float traceFraction = 1.0f;

		/// Get dot products for points A and B
		const float dotA = (intersectA - disc1posRelative) | disc1moveRelative;
		const float dotB = (intersectB - disc1posRelative) | disc1moveRelative;
		const float moveRelativeSquared = disc1moveRelative.CalcLengthSquared();
		bool isAValid = true;
		bool isBValid = true;

		/// Check if A lies within the start and end of the line segment
		if( dotA < 0.0f || dotA > moveRelativeSquared )
		{
			isAValid = false;
		}

		/// Check if B lies within the start and end of the line segment
		if( dotB < 0.0f || dotB > moveRelativeSquared )
		{
			isBValid = false;
		}

		if( !( isAValid || isBValid ) )
		{
			return 1.0f; // no valid collision points
		}

		const float oneOverMoveRelativeSquared = 1.0f / moveRelativeSquared;
		if( isAValid )
		{
			traceFraction = dotA * oneOverMoveRelativeSquared;
		}

		if( isBValid )
		{
			traceFraction = MinFloat( traceFraction, dotB * oneOverMoveRelativeSquared );
		}

		return traceFraction;
	}
}


//-------------------------------------------------------------------------------------------------
// TraceMovingDiscAndLineCollide2d
// 
// Returns a value between 0.0f and 1.0f which represents (fractionally) how far through
//	the simultaneous movement of the disc and line we got before a collision occurred.
//	A return value of 0.0f implies that the objects were overlapping to begin with;
//	a value of 1.0f implies that no collision occurred.
// 
// NOTE: This is a LINE check, not a line segment check.  <LineSeg> merely defines the line.
//-------------------------------------------------------------------------------------------------
float TraceMovingDiscAndLineCollide2d(
	const Vector2& discPos, const Vector2& discMove, float discRadius,
	const LineSegment2& line, const Vector2& lineMove, float lineRadius )
{
	const float lineSegLenSquared = line.CalcLengthSquared();
	if( lineSegLenSquared == 0.0f )
	{
		return TraceMovingDiscsCollide2d( discPos, discMove, discRadius, line.start, lineMove, lineRadius );
	}

	const float combinedRadii = discRadius + lineRadius;
	const float lineSegLen = (float)sqrt( lineSegLenSquared );
	const Vector2 unitLineSegDisplacement = line.displacement / lineSegLen;
	const Vector2 unitLineSegNormal( -unitLineSegDisplacement.y, unitLineSegDisplacement.x );
	const float discMovePerpendicular = discMove | unitLineSegNormal;
	if( discMovePerpendicular == 0.0f )
	{
		float distSquared = DistanceSquaredFromLineSegmentToPoint2d( line, discPos, NULL );
		if( distSquared < (combinedRadii * combinedRadii) )
		{
			return 0.0f; // objects began interpenetrating
		}
		else
		{
			return 1.0f; // objects never collided (not interpenetrating and no perpendicular motion)
		}
	}

	const float distToLinePerpendicular = (line.start - discPos) | unitLineSegNormal;
	const float distToLineRadiusPerpendicular = distToLinePerpendicular - combinedRadii;
	const float collideFraction = distToLineRadiusPerpendicular / discMovePerpendicular;

	if( collideFraction < 0.0f )
	{
		return 1.0f; // collision would have occurred before even getting to discPos
	}
	else if( collideFraction > 1.0f )
	{
		return 1.0f; // no collision occurred
	}
	else
	{
		return collideFraction;
	}
}



//-------------------------------------------------------------------------------------------------
// TraceMovingDiscAndLineSegmentCollide2d
// 
// Returns a value between 0.0f and 1.0f which represents (fractionally) how far through
//	the simultaneous movement of the disc and line we got before a collision occurred.
//	A return value of 0.0f implies that the objects were overlapping to begin with;
//	a value > 1.0f implies that no collision occurred.
//-------------------------------------------------------------------------------------------------
/*
float TraceMovingDiscAndLineSegmentCollide2d( const LineSegment2& discSweep, float radius, const LineSegment2& lineSeg )
{
}
*/



//-------------------------------------------------------------------------------------------------
// TraceMovingPointAndLineSegmentCollide2d
// 
// Line equation for <pointSweep> is	Ax + By - C = 0
// Line equation for <lineSeg> is		Fx + Gy - H = 0
//-------------------------------------------------------------------------------------------------
bool TraceMovingPointAndLineSegmentCollide2d( const LineSegment2& pointSweep, const LineSegment2& lineSeg, float& traceFraction )
{
	traceFraction = FLT_MAX;
	if( !DoLineSegmentsIntersect( pointSweep, lineSeg ) )
	{
		return false;
	}

	const Vector2& pointStart = pointSweep.start;
	const Vector2& lineStart = lineSeg.start;

	/// Calculate line equation for point sweep (Ax + By - C = 0)
	float A = -pointSweep.displacement.y;
	float B = +pointSweep.displacement.x;
	float C = (A * pointStart.x) + (B * pointStart.y);

	/// Calculate line equation for line segment (Fx + Gy - H = 0)
	float F = -lineSeg.displacement.y;
	float G = +lineSeg.displacement.x;
	float H = (F * lineStart.x) + (G * lineStart.y);

	/// Calculate intersection of two line equations
	float invDenominator = 1.0f / (A*G - B*F);
	float intersectX = (C*G - B*H) * invDenominator;
	float intersectY = (A*H - C*F) * invDenominator;
	Vector2 intersectPoint( intersectX, intersectY );

	/// Calculate the parametric value of this point when projected onto the <pointSweep> line segment
	traceFraction = pointSweep.CalcParametricForPoint( intersectPoint );
	return( (traceFraction >= 0.0f) && (traceFraction <= 1.0f) );
}



