//-------------------------------------------------------------------------------------------------
// collision.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
// 
// Common suite of collision detection functions.
//-------------------------------------------------------------------------------------------------
#ifndef _include_collision_
#define _include_collision_


#include "Vector2.hpp"
#include "LineSegment2.hpp"


//-------------------------------------------------------------------------------------------------
// Prototypes for functions defined in collision.cpp
//-------------------------------------------------------------------------------------------------
int CalcLineCircleIntersects( const Vector2& circleCenter, double circleRadius, double lineEquationA, double lineEquationB, double lineEquationC, Vector2& intersectA, Vector2& intersectB );
int CalcLineCircleIntersects( double circleRadius, double lineEquationA, double lineEquationB, double lineEquationC, Vector2& intersectA, Vector2& intersectB );
float TraceMovingDiscsCollide2d( const Vector2& disc1pos, const Vector2& disc1move, float disc1radius,
								 const Vector2& disc2pos, const Vector2& disc2move, float disc2radius );
float TraceMovingDiscAndLineCollide2d(	const Vector2& discPos, const Vector2& discMove, float discRadius,
										const LineSegment2& line, const Vector2& lineMove, float lineRadius );
//float TraceMovingDiscAndLineSegmentCollide2d( const LineSegment2& discSweep, float radius, const LineSegment2& lineSeg );
bool TraceMovingPointAndLineSegmentCollide2d( const LineSegment2& pointSweep, const LineSegment2& lineSeg, float& traceFraction );


//-------------------------------------------------------------------------------------------------
// DistanceSquaredBetweenPoints2d
//-------------------------------------------------------------------------------------------------
inline float DistanceSquaredBetweenPoints2d( const Vector2& point1, const Vector2& point2 )
{
	const float x = point2.x - point1.x;
	const float y = point2.y - point1.y;
	return x*x + y*y;
}


//-------------------------------------------------------------------------------------------------
// DistanceBetweenPoints2d
//-------------------------------------------------------------------------------------------------
inline float DistanceBetweenPoints2d( const Vector2& point1, const Vector2& point2 )
{
	const float x = point2.x - point1.x;
	const float y = point2.y - point1.y;
	return (float) sqrt( x*x + y*y );
}


//-------------------------------------------------------------------------------------------------
// IsDistanceBetweenPointsGreater
//-------------------------------------------------------------------------------------------------
inline bool IsDistanceBetweenPointsGreater( const Vector2& point1, const Vector2& point2, float distance )
{
	float distanceSquared = DistanceSquaredBetweenPoints2d( point1, point2 );
	return distanceSquared > (distance * distance);
}


//-------------------------------------------------------------------------------------------------
// IsDistanceBetweenPointsLess
//-------------------------------------------------------------------------------------------------
inline bool IsDistanceBetweenPointsLess( const Vector2& point1, const Vector2& point2, float distance )
{
	float distanceSquared = DistanceSquaredBetweenPoints2d( point1, point2 );
	return distanceSquared < (distance * distance);
}


//-------------------------------------------------------------------------------------------------
// DoDiscsOverlap2d
//-------------------------------------------------------------------------------------------------
inline bool DoDiscsOverlap2d( const Vector2& discCenter1, const Vector2& discCenter2, float discRadius1, float discRadius2 )
{
	const float radii = discRadius1 + discRadius2;
	return( DistanceSquaredBetweenPoints2d( discCenter1, discCenter2 ) <= (radii * radii) );
}


//-------------------------------------------------------------------------------------------------
// ClosestPointOnLineSegmentToPoint2d
//-------------------------------------------------------------------------------------------------
inline const Vector2 ClosestPointOnLineSegmentToPoint2d( const LineSegment2& lineSegment, const Vector2& point )
{
	// notation: line segment starts at A and ends at B, and the remote point is C
	const Vector2 AC = point - lineSegment.start;
	const Vector2 CB = lineSegment.displacement - AC;

	/// Calculate the dot product of AC and AB
	const float dotA = AC | lineSegment.displacement;
	
	/// Check if the dot angle is obtuse
	if( dotA <= 0.0f )
	{
		/// A is the closest point on the line segment to C
		return lineSegment.start;
	}

	/// Calculate the dot product of BC and BA (equal to CB dot AB)
	const float dotB = CB | lineSegment.displacement;

	/// Check if the dot angle is obtuse
	if( dotB <= 0.0f )
	{
		/// B is the closest point on the line segment to C
		return lineSegment.start + lineSegment.displacement;
	}

	/// The closest point is between A and B; interpolate the dots to find out where
	const float fraction = dotA / (dotA + dotB);
	return lineSegment.start + (fraction * lineSegment.displacement);
}


//-------------------------------------------------------------------------------------------------
// DistanceSquaredFromLineSegmentToPoint2d
//-------------------------------------------------------------------------------------------------
inline float DistanceSquaredFromLineSegmentToPoint2d( const LineSegment2& lineSegment, const Vector2& point, float* pLineFraction )
{
	// notation: line segment starts at A and ends at B, and the remote point is C
	const Vector2 AC = point - lineSegment.start;
	const Vector2 CB = lineSegment.displacement - AC;

	/// Calculate the dot product of AC and AB
	const float dotA = AC | lineSegment.displacement;
	
	/// Check if the dot angle is obtuse
	if( dotA <= 0.0f )
	{
		/// A is the closest point on the line segment to C
		if( pLineFraction )
		{
			*pLineFraction = 0.0f;
		}
		return DistanceSquaredBetweenPoints2d( point, lineSegment.start );
	}

	/// Calculate the dot product of BC and BA (equal to CB dot AB)
	const float dotB = CB | lineSegment.displacement;

	/// Check if the dot angle is obtuse
	if( dotB <= 0.0f )
	{
		/// B is the closest point on the line segment to C
		if( pLineFraction )
		{
			*pLineFraction = 1.0f;
		}
		return DistanceSquaredBetweenPoints2d( point, lineSegment.start + lineSegment.displacement );
	}

	/// The closest point is between A and B; interpolate the dots to find out where
	const float fraction = dotA / (dotA + dotB);
	const Vector2 closestPoint = lineSegment.start + (fraction * lineSegment.displacement);

	if( pLineFraction )
	{
		*pLineFraction = fraction;
	}
	return DistanceSquaredBetweenPoints2d( point, closestPoint );
}


//-------------------------------------------------------------------------------------------------
// DistanceFromLineSegmentToPoint2d
//-------------------------------------------------------------------------------------------------
inline float DistanceFromLineSegmentToPoint2d( const LineSegment2& lineSegment, const Vector2& point, float* pLineFraction )
{
	return (float) sqrt( DistanceSquaredFromLineSegmentToPoint2d( lineSegment, point, pLineFraction ) );
}


//-------------------------------------------------------------------------------------------------
// DoDiscAndLineSegmentOverlap2d
//-------------------------------------------------------------------------------------------------
inline bool DoDiscAndLineSegmentOverlap2d( const Vector2& discCenter, float discRadius, const LineSegment2& lineSegment )
{
	return( DistanceSquaredFromLineSegmentToPoint2d( lineSegment, discCenter, NULL ) <= (discRadius * discRadius) );
}





//-------------------------------------------------------------------------------------------------
// DoLineSegmentsIntersect
//-------------------------------------------------------------------------------------------------
inline bool DoLineSegmentsIntersect( const LineSegment2& lineSeg1, const LineSegment2& lineSeg2 )
{
	// notation: segment1 goes from A to B; segment2 goes from C to D

	const Vector2 CA = lineSeg1.start - lineSeg2.start;
	const Vector2 CB = CA + lineSeg1.displacement;
	const Vector2 AD = lineSeg2.displacement - CA;

	/// Calculate the dot product of CA and line 2's normal (-y,x)
	const float dotA = (-lineSeg2.displacement.y * CA.x) + (lineSeg2.displacement.x * CA.y);

	/// Calculate the dot product of CB and line 2's normal
	const float dotB = (-lineSeg2.displacement.y * CB.x) + (lineSeg2.displacement.x * CB.y);

	/// Check if both A and B are on the front side of line 2
	if( dotA > 0.0f && dotB > 0.0f )
		return false;

	/// Check if both A and B are on the back side of line 2
	if( dotA < 0.0f && dotB < 0.0f )
		return false;

	/// Calculate the dot product of AC and line 1's normal
	const float dotC = (lineSeg1.displacement.y * CA.x) + (-lineSeg1.displacement.x * CA.y);

	/// Calculate the dot product of AD and line 1's normal
	const float dotD = (-lineSeg1.displacement.y * AD.x) + (lineSeg1.displacement.x * AD.y);

	/// Check if both of segment 1's endpoints are on the front side of line 2
	if( dotC > 0.0f && dotD > 0.0f )
		return false;

	/// Check if both of segment 1's endpoints are on the back side of line 2
	if( dotC < 0.0f && dotD < 0.0f )
		return false;

	return true;
}


//-------------------------------------------------------------------------------------------------
// DistanceSquaredBetweenLineSegments2d
// 
// FIXME: these function calls create redundant temp vectors; rewrite all of this
//	as a single function right here to minimize temp variable creation overhead.
//-------------------------------------------------------------------------------------------------
inline float DistanceSquaredBetweenLineSegments2d( const LineSegment2& lineSeg1, const LineSegment2& lineSeg2 )
{
	// notation: segment1 goes from A to B; segment2 goes from C to D

	/// Check if line segments intersect
	if( DoLineSegmentsIntersect( lineSeg1, lineSeg2 ) )
		return 0.0f;

	/// Get the distance squared from A to CD
	const float distSquared_A_CD = DistanceSquaredFromLineSegmentToPoint2d( lineSeg2, lineSeg1.start, NULL );

	/// Get the distance squared from B to CD
	const float distSquared_B_CD = DistanceSquaredFromLineSegmentToPoint2d( lineSeg2, lineSeg1.start + lineSeg1.displacement, NULL );

	/// Get the distance squared from C to AB
	const float distSquared_C_AB = DistanceSquaredFromLineSegmentToPoint2d( lineSeg1, lineSeg2.start, NULL );

	/// Get the distance squared from D to AB
	const float distSquared_D_AB = DistanceSquaredFromLineSegmentToPoint2d( lineSeg1, lineSeg2.start + lineSeg2.displacement, NULL );

	/// Return the smallest of the distances
	return MinFloat( distSquared_A_CD, distSquared_B_CD, distSquared_C_AB, distSquared_D_AB );
}


//-------------------------------------------------------------------------------------------------
// DistanceBetweenLineSegments2d
//-------------------------------------------------------------------------------------------------
inline float DistanceBetweenLineSegments2d( const LineSegment2& lineSeg1, const LineSegment2& lineSeg2 )
{
	return (float) sqrt( DistanceSquaredBetweenLineSegments2d( lineSeg1, lineSeg2 ) );
}


//-------------------------------------------------------------------------------------------------
// DoMovingDiscAndLineSegmentCollide2d
//-------------------------------------------------------------------------------------------------
inline bool DoMovingDiscAndLineSegmentCollide2d( const Vector2& discCenter, float radius, const Vector2& discMove, const LineSegment2& lineSeg )
{
	/// Construct a line segment representing the disc's motion
	const LineSegment2 discCenterMove( discCenter, discMove );

	/// Construct axis-aligned bounding boxes that contain the disc's motion and the line segment
	AABB2 discBounds = discCenterMove.CalcExtents();
	const AABB2 lineSegBounds = lineSeg.CalcExtents();

	/// Expand the disc's motion-bounds to encompass its radius as well
	discBounds.mins.x -= radius;
	discBounds.maxs.x += radius;
	discBounds.mins.y -= radius;
	discBounds.maxs.y += radius;

	/// Do a trivial rejection check; if AABBs don't overlap, there's no chance of intersection
	if( !discBounds.IsOverlapping( lineSegBounds ) )
		return false;

	/// Overlap occurs if line segments are closer than the disc's radius
	const float distSquared = DistanceSquaredBetweenLineSegments2d( discCenterMove, lineSeg );
	return distSquared < (radius * radius);
}


//-------------------------------------------------------------------------------------------------
// DoMovingDiscsCollide2d
//-------------------------------------------------------------------------------------------------
inline bool DoMovingDiscsCollide2d( const Vector2& disc1pos, const Vector2& disc1move, float disc1radius, const Vector2& disc2pos, const Vector2& disc2move, float disc2radius )
{
	LineSegment2 motion1( disc1pos, disc1move - disc2move );
	const float distSquared = DistanceSquaredFromLineSegmentToPoint2d( motion1, disc2pos, NULL );
	const float radiiSummed = disc1radius + disc2radius;
	const float radiiSquared = radiiSummed * radiiSummed;
	return distSquared < radiiSquared;
}


//-------------------------------------------------------------------------------------------------
// DoMovingDiscsCollide2d
// 
// This version differs from the above only in that it assumes disc2 is not moving
//-------------------------------------------------------------------------------------------------
inline bool DoMovingDiscsCollide2d( const Vector2& disc1pos, const Vector2& disc1move, float disc1radius, const Vector2& disc2pos, float disc2radius )
{
	LineSegment2 motion1( disc1pos, disc1move );
	const float distSquared = DistanceSquaredFromLineSegmentToPoint2d( motion1, disc2pos, NULL );
	const float radiiSummed = disc1radius + disc2radius;
	const float radiiSquared = radiiSummed * radiiSummed;
	return distSquared < radiiSquared;
}


//-------------------------------------------------------------------------------------------------
// DoAABBAndLineSegmentCollide2d
//-------------------------------------------------------------------------------------------------
/*
inline bool DoAABBAndLineSegmentCollide2d( const AABB2& aabb, const LineSegment2& lineSeg )
{
	/// Check majority case trivial reject
	AABB2 lineSegBounds( lineSeg );
	if( !lineSegBounds.IsOverlapping( aabb ) )
		return false;

//	Vector2 end = lineSeg.start + lineSeg.displacement;
	float overlapMinX = ClampFloat( lineSegBounds.mins.x, aabb.mins.x, aabb.maxs.x );
	float overlapMaxX = ClampFloat( lineSegBounds.maxs.x, aabb.mins.x, aabb.maxs.x );
	float overlapMinY = ClampFloat( lineSegBounds.mins.y, aabb.mins.y, aabb.maxs.y );
	float overlapMaxY = ClampFloat( lineSegBounds.maxs.y, aabb.mins.y, aabb.maxs.y );

	asdf
}
*/


//-------------------------------------------------------------------------------------------------
// QuadraticEquation
// 
// Returns the number of real solutions (either 0, 1, or 2).
// 
// If 0 solutions exist, both <solutionA> and <solutionB> are invalid.
// If 1 solution exists, it is stored in <solutionA>; <solutionB> is invalid.
// If 2 solutions exist, they are stored in <solutionA> and <solutionB>.
// 
// Solves the equation  Ax^2 + Bx + C = 0  for x, where the solution(s) are:
// 
//	-B + sqrt( B^2 - 4AC )
//	----------------------
//           2A
// 
// and
// 
//	-B - sqrt( B^2 - 4AC )
//	----------------------
//           2A
// 
// NOTE: if the term (B^2 - 4AC) is:
//		positive	there are two real solutions
//		zero		there is one real solution
//		negative	there are no real solutions
//-------------------------------------------------------------------------------------------------
inline int QuadraticEquation( double A, double B, double C, double& solutionA, double& solutionB )
{
	double radical = (B*B) - (4.0 * A * C);
	if( radical < 0.0 )
	{
		return 0;
	}
	else if( radical == 0.0 )
	{
		solutionA = (-B) / (2.0 * A);
		return 1;
	}
	else
	{
		double sqrtRadical = sqrt( radical );
		solutionA = (-B + sqrtRadical ) / (2.0 * A);
		solutionB = (-B - sqrtRadical ) / (2.0 * A);
		return 2;
	}
}


//-------------------------------------------------------------------------------------------------
// QuadraticEquation
// 
// Same as above, but with float solutions instead of doubles.
//-------------------------------------------------------------------------------------------------
inline int QuadraticEquation( double A, double B, double C, float& solutionA, float& solutionB )
{
	double radical = (B*B) - (4.0 * A * C);
	if( radical < 0.0 )
	{
		return 0;
	}
	else if( radical == 0.0 )
	{
		solutionA = (float)( (-B) / (2.0 * A) );
		return 1;
	}
	else
	{
		double sqrtRadical = sqrt( radical );
		solutionA = (float)( (-B + sqrtRadical ) / (2.0 * A) );
		solutionB = (float)( (-B - sqrtRadical ) / (2.0 * A) );
		return 2;
	}
}


//-------------------------------------------------------------------------------------------------
// CalcLineCircleIntersects
// 
// Finds the x,y coordinates of the intersection(s) of the line upon which
//	<lineSegment> lies with a circle of radius <circleRadius> centered at <circleCenter>.
// 
// Returns the number of real intersection solutions, which may be 0, 1, or 2.
// If return value is 0: no solution, <intersectA> and <intersectB> are invalid
// If return value is 1: one solution, only <intersectA> is valid
// If return value is 2: two solutions, <intersectA> and <intersectB> are valid
//-------------------------------------------------------------------------------------------------
inline int CalcLineCircleIntersects( const Vector2& circleCenter, double circleRadius, const LineSegment2& lineSeg, Vector2& intersectA, Vector2& intersectB )
{
	Vector2 relocatedLineSegStart = lineSeg.start - circleCenter;
	double lineEquationA = -lineSeg.displacement.y;
	double lineEquationB = lineSeg.displacement.x;
	double lineEquationC = -( (lineEquationA * relocatedLineSegStart.x) + (lineEquationB * relocatedLineSegStart.y) );

	int numSolutions = CalcLineCircleIntersects( circleRadius, lineEquationA, lineEquationB, lineEquationC, intersectA, intersectB );
	intersectA += circleCenter;
	intersectB += circleCenter;
	return numSolutions;
}


//-------------------------------------------------------------------------------------------------
// CalcLineCircleIntersects
// 
// Finds the x,y coordinates of the intersection(s) of the line upon which
//	<lineSegment> lies with a circle of radius <circleRadius> centered at the origin.
// 
// Returns the number of real intersection solutions, which may be 0, 1, or 2.
// If return value is 0: no solution, <intersectA> and <intersectB> are invalid
// If return value is 1: one solution, only <intersectA> is valid
// If return value is 2: two solutions, <intersectA> and <intersectB> are valid
//-------------------------------------------------------------------------------------------------
inline int CalcLineCircleIntersects( double circleRadius, const LineSegment2& lineSeg, Vector2& intersectA, Vector2& intersectB )
{
	double lineEquationA = -lineSeg.displacement.y;
	double lineEquationB = lineSeg.displacement.x;
	double lineEquationC = -( (lineEquationA * lineSeg.start.x) + (lineEquationB * lineSeg.start.y) );

	int numSolutions = CalcLineCircleIntersects( circleRadius, lineEquationA, lineEquationB, lineEquationC, intersectA, intersectB );
	return numSolutions;
}





#endif // _include_collision_
