
#include "pch.hpp"
#include "Ball.hpp"
#include "Map.hpp"
#include "TheGame.hpp"


//-------------------------------------------------------------------------------------------------
// class constants
//-------------------------------------------------------------------------------------------------
const float DEFAULT_BALL_RADIUS = 12.0f;
const float GRAVITY_STRENGTH = 1000.0f;
const Vector2 GRAVITY_ACCELERATION_VECTOR( 0.0f, GRAVITY_STRENGTH );
//const double BALL_DRAG_COEFFICIENT_PER_SECOND = 0.9; // a value of 1.0 disables this effect (air drag)
//const double BALL_DRAG_ROTATIONAL_COEFFICIENT_PER_SECOND = 0.5f; // a value of 1.0 disables this effect (rotational air drag)
const Vector2 DEFAULT_BALL_START_POS( 512.0f, 560.0f );
const Vector2 DEFAULT_BALL_START_VEL( 500.0f, -40.0f );
const float DAMPENING_MULTIPLIER_DUE_TO_INELASTIC_COLLISION = 0.8f;
const float DAMPENING_MULTIPLIER_DUE_TO_FRICTION = 0.97f;
const float DAMPENING_MULTIPLIER_DUE_TO_INTERPENETRATION = 0.90f;
const float SCALE_BACK_COLLIDE_FRACTION_BALL_PLACEMENT = 0.99f;
const float INTERPENETRATE_OVERKILL_MULTIPLIER = 1.01f;
//const float MIN_PHYSICS_TIME_EPSILON = 0.001f;
//const int MAX_PHYSICS_RUNS_PER_FRAME = 8;
const int MAX_PHYSICS_INTERPENETRATION_RESOLVES_PER_FRAME = 8;
const float MIN_PHYSICS_TIME_DELTA = 0.005f;
const float MAX_PHYSICS_TIME_DELTA = 0.02f;
const float FULL_BALL_ENERGY = 1000.0f; // the speed at which ball impacts are played at full volume
const float SPIN_FACTOR_FROM_COLLISION = 2.5f;
//const float SPIN_FACTOR_FROM_COLLISION_WITH_VERTEX = 2.5f;
//const float SPIN_FACTOR_FROM_COLLISION_WITH_LINEDEF = 2.5f;
const float ROTATION_FRICTION_EFFICIENCY = 0.35f;


bool g_TEMP_physicsDebugMode = false;
bool g_TEMP_enableFreezingPhysics = false;
bool g_TEMP_isPhysicsFrozen = false;
bool g_TEMP_isCurrentlyInterpenetrating = false;
Vector2 g_TEMP_lineNormal = Vector2::ZERO;
Vector2 g_TEMP_lineDisp = Vector2::ZERO;
Vector2 g_TEMP_ballOldVel = Vector2::ZERO;
Vector2 g_TEMP_ballNewVel = Vector2::ZERO;
MapVertex* g_TEMP_collisionVertex = NULL;
MapLinedef* g_TEMP_collisionLinedef = NULL;
bool g_TEMP_showCollisionPlanes = false;


Ball::Ball( const std::string& tgaBall, const std::string& tgaHighlight )
:	m_animHighlight( tgaHighlight ),
	m_animBall( tgaBall ),
	m_pos( DEFAULT_BALL_START_POS ),
	m_vel( DEFAULT_BALL_START_VEL ),
	m_angleDegrees( 0 ),
	m_rotationalVelocity( 0.0f ),
	m_radius( DEFAULT_BALL_RADIUS )
{
}


Ball::~Ball()
{
}

void Ball::Draw()
{
	glEnable( GL_BLEND );

	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	m_animBall.Draw( m_pos, Vector2::ONE, m_angleDegrees );

	glBlendFunc( GL_SRC_ALPHA, GL_ONE );
	m_animHighlight.Draw( m_pos );

	if( g_TEMP_physicsDebugMode )
	{
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		if( g_TEMP_collisionVertex )
		{
			glLineWidth( 5.0f );
			if( g_TEMP_showCollisionPlanes )
			{
				theRenderer.DrawVector2d( m_pos, m_pos + g_TEMP_ballOldVel, Rgba::GREEN, Rgba::GREEN );
				theRenderer.DrawVector2d( m_pos, m_pos + g_TEMP_ballNewVel, Rgba::RED, Rgba::RED );
				theRenderer.DrawVector2d( g_TEMP_collisionVertex->m_pos - g_TEMP_lineDisp, g_TEMP_collisionVertex->m_pos + g_TEMP_lineDisp, Rgba::DARKYELLOW, Rgba::DARKYELLOW );
				theRenderer.DrawVector2d( g_TEMP_collisionVertex->m_pos, g_TEMP_collisionVertex->m_pos + g_TEMP_lineNormal, Rgba::DARKPURPLE, Rgba::DARKPURPLE );
			}
			theRenderer.DrawCircle2d( g_TEMP_collisionVertex->m_pos, 5.0f, Rgba::PURPLE, 16 );
		}
		else if( g_TEMP_collisionLinedef )
		{
			glLineWidth( 3.0f );
			if( g_TEMP_showCollisionPlanes )
			{
				theRenderer.DrawVector2d( m_pos, m_pos + g_TEMP_ballOldVel, Rgba::GREEN, Rgba::GREEN );
				theRenderer.DrawVector2d( m_pos, m_pos + g_TEMP_ballNewVel, Rgba::RED, Rgba::RED );
				theRenderer.DrawVector2d( m_pos - g_TEMP_lineDisp, m_pos + g_TEMP_lineDisp, Rgba::DARKYELLOW, Rgba::DARKYELLOW );
				theRenderer.DrawVector2d( m_pos, m_pos + g_TEMP_lineNormal, Rgba::DARKPURPLE, Rgba::DARKPURPLE );
			}
			theRenderer.DrawVector2d( g_TEMP_collisionLinedef->m_start->m_pos, g_TEMP_collisionLinedef->m_end->m_pos, Rgba::DARKBLUE, Rgba::DARKPURPLE );
		}

		if( g_TEMP_physicsDebugMode && g_TEMP_isCurrentlyInterpenetrating )
		{
			g_TEMP_isCurrentlyInterpenetrating = false;
			glLineWidth( 8.0f );
			theRenderer.DrawCircle2d( m_pos, m_radius * 2, Rgba::WHITE, 32 );
		}

		glLineWidth( 1.0f );
	}
}


//-------------------------------------------------------------------------------------------------
// ResolveVertexInterpenetration
//-------------------------------------------------------------------------------------------------
void Ball::ResolveVertexInterpenetration( const Vector2& vertexPos )
{
	theConsole.Printf( "Resolving vertex interpenetration on frame #%d\n", theApplication.GetFrameNumber() );

	Vector2 vertexToBall = m_pos - vertexPos;
	vertexToBall.Normalize();

	Vector2 vertexToBallAdjusted = vertexToBall * m_radius * INTERPENETRATE_OVERKILL_MULTIPLIER;
	m_pos = vertexPos + vertexToBallAdjusted;
}


//-------------------------------------------------------------------------------------------------
// ResolveVertexInterpenetration
//-------------------------------------------------------------------------------------------------
void Ball::ResolveVertexInterpenetration( MapVertex* vertex )
{
	ResolveVertexInterpenetration( vertex->m_pos );
}


//-------------------------------------------------------------------------------------------------
// ResolveLinedefInterpenetration
//-------------------------------------------------------------------------------------------------
void Ball::ResolveLinedefInterpenetration( MapLinedef* linedef )
{
	theConsole.Printf( "Resolving linedef interpenetration on frame #%d\n", theApplication.GetFrameNumber() );

	/// Find closest point on linedef to ball center
	LineSegment2 lineSeg( linedef->m_start->m_pos, linedef->m_end->m_pos - linedef->m_start->m_pos );
	Vector2 closestPoint = ClosestPointOnLineSegmentToPoint2d( lineSeg, m_pos );
	ResolveVertexInterpenetration( closestPoint );
	linedef->m_highlightFrameNumber = theApplication.GetFrameNumber();
}


//-------------------------------------------------------------------------------------------------
// ResolveVertexInterpenetrations
//-------------------------------------------------------------------------------------------------
bool Ball::ResolveVertexInterpenetrations()
{
	bool wasInterpenetrated = false;

	std::list<MapVertex*>::iterator begin = theGame->GetCurrentMap().mapArea.m_vertexList.begin();
	std::list<MapVertex*>::iterator end = theGame->GetCurrentMap().mapArea.m_vertexList.end();
	std::list<MapVertex*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* vertex = (*iter);
		if( DoDiscsOverlap2d( m_pos, vertex->m_pos, m_radius, 0.0f ) )
		{
			ResolveVertexInterpenetration( vertex );
			wasInterpenetrated = true;
		}
	}

	return wasInterpenetrated;
}


//-------------------------------------------------------------------------------------------------
// ResolveLinedefInterpenetrations
//-------------------------------------------------------------------------------------------------
bool Ball::ResolveLinedefInterpenetrations()
{
	bool wasInterpenetrated = false;

	std::list<MapLinedef*>::iterator begin = theGame->GetCurrentMap().mapArea.m_linedefList.begin();
	std::list<MapLinedef*>::iterator end = theGame->GetCurrentMap().mapArea.m_linedefList.end();
	std::list<MapLinedef*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapLinedef* linedef = (*iter);
		LineSegment2 lineSeg( linedef->m_start->m_pos, linedef->m_end->m_pos - linedef->m_start->m_pos );
		if( DoDiscAndLineSegmentOverlap2d( m_pos, m_radius, lineSeg ) )
		{
			ResolveLinedefInterpenetration( linedef );
			wasInterpenetrated = true;
		}
	}

	return wasInterpenetrated;
}


//-------------------------------------------------------------------------------------------------
// ResolveInterpenetrations
//-------------------------------------------------------------------------------------------------
bool Ball::ResolveInterpenetrations()
{
	bool wasInterpenetrated = ResolveVertexInterpenetrations() || ResolveLinedefInterpenetrations();
	bool arePenetrationsResolved = !wasInterpenetrated;

	int resolutionCountThisFrame = 0;
	while( !arePenetrationsResolved )
	{
		m_vel *= DAMPENING_MULTIPLIER_DUE_TO_INTERPENETRATION;
		arePenetrationsResolved = !( ResolveVertexInterpenetrations() || ResolveLinedefInterpenetrations() );
		resolutionCountThisFrame ++;
		if( resolutionCountThisFrame > MAX_PHYSICS_INTERPENETRATION_RESOLVES_PER_FRAME )
		{
			break;
		}
	}

	if( resolutionCountThisFrame > 0 )
	{
		theConsole.Printf( "WARNING: required %d passes to resolve interpenetrations this frame!\n", resolutionCountThisFrame );
	}

	return wasInterpenetrated;
}

/*
//-------------------------------------------------------------------------------------------------
// RunPhysicsTime
//-------------------------------------------------------------------------------------------------
void Ball::RunPhysicsTime()
{
	/// Determine if the ball is currently interpenetrating any linedefs or vertexes
	bool wasInterpenetrating = ResolveInterpenetrations();
	if( wasInterpenetrating )
	{
		g_TEMP_isPhysicsFrozen = true;
	}

	/// Calculate the ball's theoretical movement vector and end position
	Vector2 movement = vel * physicsTimeRemainingToResolveThisFrame;
	Vector2 newPos = pos + movement;

	/// Check if the ball collided (or would have collided) with any items
	std::list< Item* > collideItems;
	theGame->GetCurrentMap().CheckForCollisionAgainstItems( pos, radius, movement, collideItems );
	if( !collideItems.empty() )
	{
		for( std::list< Item* >::iterator itemIter = collideItems.begin(); itemIter != collideItems.end(); ++itemIter )
		{
			if( (*itemIter)->IsCollectable() )
			{
				theGame->GetCurrentMap().RemoveItem( *itemIter );
			}
		}
	}

	/// Check if the ball collided with any vertex
	float vertexTimeFraction;
	float linedefTimeFraction;
	MapVertex* collideVertex = theGame->GetCurrentMap().CheckForCollisionAgainstVertexes( pos, radius, movement, vertexTimeFraction );
	MapLinedef* collideLinedef = theGame->GetCurrentMap().CheckForCollisionAgainstLinedefs( pos, radius, movement, linedefTimeFraction );

	if( collideVertex && vertexTimeFraction <= linedefTimeFraction )
	{
		float ballEnergy = vel.CalcLengthSquared();
		float ballEnergyFraction = ballEnergy / (FULL_BALL_ENERGY * FULL_BALL_ENERGY);
		ballEnergyFraction = ClampFloat( ballEnergyFraction, 0.0f, 1.0f );
		ResolveCollisionWithVertex( collideVertex, movement, vertexTimeFraction, ballEnergyFraction );
		float timeFractionRemaining = 1.0f - vertexTimeFraction;
		timeFractionRemaining = ClampFloat( timeFractionRemaining, 0.0f, 1.0f );
		physicsTimeRemainingToResolveThisFrame *= timeFractionRemaining;
		theGame->BallBounce( ballEnergyFraction );
	}
	else if( collideLinedef )
	{
		float ballEnergy = vel.CalcLengthSquared();
		float ballEnergyFraction = ballEnergy / (FULL_BALL_ENERGY * FULL_BALL_ENERGY);
		ballEnergyFraction = ClampFloat( ballEnergyFraction, 0.0f, 1.0f );
		ResolveCollisionWithLinedef( collideLinedef, movement, linedefTimeFraction, ballEnergyFraction );
		float timeFractionRemaining = 1.0f - linedefTimeFraction;
		timeFractionRemaining = ClampFloat( timeFractionRemaining, 0.0f, 1.0f );
		physicsTimeRemainingToResolveThisFrame *= timeFractionRemaining;
		theGame->BallBounce( ballEnergyFraction );
	}
	else
	{
		pos = newPos;
		physicsTimeRemainingToResolveThisFrame = 0.0f;
	}
}


//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void Ball::Update()
{
	if( g_TEMP_physicsDebugMode && g_TEMP_enableFreezingPhysics && g_TEMP_isPhysicsFrozen )
	{
		return;
	}

	/// Determine how much time has elapsed since the last frame
	float timeSlice = (float)theApplication.GetFrameSeconds();
	physicsTimeRemainingToResolveThisFrame = timeSlice;

	/// Apply gravity
	Vector2 gravitationalAcceleration = GRAVITY_ACCELERATION_VECTOR * timeSlice;
	vel += gravitationalAcceleration;

	/// Apply ambient drag factor
	double inverseDragCoefficient = 1.0 - BALL_DRAG_COEFFICIENT_PER_SECOND;
	double scaledInverseDragCoefficient = inverseDragCoefficient * timeSlice;
	double scaledDragCoefficient = 1.0 - scaledInverseDragCoefficient;
	vel *= (float)scaledDragCoefficient;

	double inverseRotationalDragCoefficient = 1.0 - BALL_DRAG_ROTATIONAL_COEFFICIENT_PER_SECOND;
	double scaledInverseRotationalDragCoefficient = inverseRotationalDragCoefficient * timeSlice;
	double scaledRotationalDragCoefficient = 1.0 - scaledInverseRotationalDragCoefficient;
	rotationalVelocity *= (float)scaledRotationalDragCoefficient;

	/// Update ball rotation angle based on angular velocity
	angleDegrees += (rotationalVelocity * timeSlice);

	theConsole.Printf( "------------------------- physics start\n" );

	/// Simulate physics -- through multiple collisions if necessary -- until all time is spent
	int physicsRunCount = 0;
	while( physicsTimeRemainingToResolveThisFrame > MIN_PHYSICS_TIME_EPSILON )
	{
		if( physicsRunCount > 0 )
		{
			theConsole.Printf( "Running partial physics frame, frame #%d, run #%d, timeLeft = %g seconds\n",
				theApplication.GetFrameNumber(), physicsRunCount + 1, physicsTimeRemainingToResolveThisFrame );
		}
		RunPhysicsTime();
		physicsRunCount ++;

		if( physicsRunCount > MAX_PHYSICS_RUNS_PER_FRAME )
		{
			theConsole.Printf( "WARNING: Used all %d physics runs this frame, still had %.2f%% time remaining to resolve!\n",
				MAX_PHYSICS_RUNS_PER_FRAME, 100.0f * physicsTimeRemainingToResolveThisFrame / timeSlice );

			break;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// ResolveCollisionWithVertex
//-------------------------------------------------------------------------------------------------
void Ball::ResolveCollisionWithVertex( MapVertex* vertex, const Vector2& originalMovement, float collideFraction, float ballEnergyFraction )
{
	theConsole.Printf( "Collision on frame %d with VERTEX %x (at %g,%g):\n", theApplication.GetFrameNumber(), (unsigned int)vertex, vertex->m_pos.x, vertex->m_pos.y );
	theConsole.Printf( "  Ball movement was %g,%g ; collideFraction = %g\n", originalMovement.x, originalMovement.y, collideFraction );

	Vector2 oldPosTEMP = pos;
	Vector2 oldVelTEMP = vel;

	/// Move the ball to a spot infinitesimally before its point of impact
	collideFraction = ClampFloat( collideFraction, 0.0f, 1.0f );
	collideFraction *= SCALE_BACK_COLLIDE_FRACTION_BALL_PLACEMENT;
	pos += (originalMovement * collideFraction);
	theConsole.Printf( "  Result of the collision:\n" );

	/// Calculate normal and displacement unit vectors for the virtual line of impact
	Vector2 unitLineNormal = pos - vertex->m_pos;
	unitLineNormal.Normalize();
	const Vector2 unitLineDisp( unitLineNormal.y, -unitLineNormal.x );
	theConsole.Printf( "    virtual line normal = (%g,%g), disp = (%g,%g)\n", unitLineNormal.x, unitLineNormal.y, unitLineDisp.x, unitLineDisp.y );
	theConsole.Printf( "    pos (%g,%g) -> (%g,%g)\n", oldPosTEMP.x, oldPosTEMP.y, pos.x, pos.y );

	/// Break the ball's velocity into parallel and perpendicular components
	float speedParallel = vel | unitLineDisp;
	float speedPerpendicular = vel | unitLineNormal;

	/// Reverse and dampen perpendicular motion (no sliding friction for vertex impacts)
	speedPerpendicular *= -BALL_DAMPENING_FRACTION_ON_IMPACT;

	/// Recombine components to discover resultant velocity after collision
	vel = (unitLineNormal * speedPerpendicular) + (unitLineDisp * speedParallel);
	theConsole.Printf( "    vel (%g,%g) -> (%g,%g)\n", oldVelTEMP.x, oldVelTEMP.y, vel.x, vel.y );

	/// Affect rotational velocity
	float rotationalVelocityDelta = (speedParallel * SPIN_FACTOR_FROM_COLLISION_WITH_VERTEX * ballEnergyFraction);
	bool areSignsEqual = (rotationalVelocityDelta * rotationalVelocity > 0.0f);
	if( !areSignsEqual )
	{
		rotationalVelocity *= 0.3f;
	}
	rotationalVelocity += rotationalVelocityDelta;

	g_TEMP_isPhysicsFrozen = true;
	g_TEMP_lineNormal = 100.0f * unitLineNormal;
	g_TEMP_lineDisp = 100.0f * unitLineDisp;
	g_TEMP_ballOldVel = oldVelTEMP;
	g_TEMP_ballNewVel = vel;
	g_TEMP_collisionVertex = vertex;
	g_TEMP_collisionLinedef = NULL;
}


//-------------------------------------------------------------------------------------------------
// ResolveCollisionWithLinedef
//-------------------------------------------------------------------------------------------------
void Ball::ResolveCollisionWithLinedef( MapLinedef* linedef, const Vector2& originalMovement, float collideFraction, float ballEnergyFraction )
{
	theConsole.Printf( "Collision on frame %d with LINEDEF %x (at %g,%g -> %g,%g):\n",
		theApplication.GetFrameNumber(), (unsigned int)linedef,
		linedef->m_start->m_pos.x, linedef->m_start->m_pos.y, linedef->m_end->m_pos.x, linedef->m_end->m_pos.y );
	theConsole.Printf( "  Ball movement was %g,%g ; collideFraction = %g\n", originalMovement.x, originalMovement.y, collideFraction );

	Vector2 oldPosTEMP = pos;
	Vector2 oldVelTEMP = vel;

	/// Move the ball to a spot infinitesimally before its point of impact
	collideFraction = ClampFloat( collideFraction, 0.0f, 1.0f );
	collideFraction *= SCALE_BACK_COLLIDE_FRACTION_BALL_PLACEMENT;
	pos += (originalMovement * collideFraction);

	theConsole.Printf( "  Result of the collision:\n" );
	theConsole.Printf( "    pos (%g,%g) -> (%g,%g)\n", oldPosTEMP.x, oldPosTEMP.y, pos.x, pos.y );

	/// Set up some convenience variables
	const Vector2& lineStart = linedef->m_start->m_pos;
	const Vector2& lineEnd = linedef->m_end->m_pos;
	Vector2 unitLineDisp = lineEnd - lineStart;
	unitLineDisp.Normalize();
	const Vector2 unitLineNormal( -unitLineDisp.y, unitLineDisp.x );

	/// Break the ball's velocity into parallel and perpendicular components
	float speedParallel = vel | unitLineDisp;
	float speedPerpendicular = vel | unitLineNormal;

	/// Reverse and dampen perpendicular motion; dampen parallel motion by friction factor
	speedPerpendicular *= -BALL_DAMPENING_FRACTION_ON_IMPACT;
	speedParallel *= BALL_DAMPENING_FRICTION_ON_IMPACT;

	/// Recombine components to discover resultant velocity after collision
	vel = (unitLineNormal * speedPerpendicular) + (unitLineDisp * speedParallel);
	theConsole.Printf( "    vel (%g,%g) -> (%g,%g)\n", oldVelTEMP.x, oldVelTEMP.y, vel.x, vel.y );

	/// Calculate target rotational velocity
	float targetRotationalVelocity = radius * speedParallel * 0.3f;

	/// Affect rotational velocity
	float rotationalVelocityDelta = (speedParallel * SPIN_FACTOR_FROM_COLLISION_WITH_LINEDEF);
	bool areSignsEqual = (rotationalVelocityDelta * rotationalVelocity >= 0.0f);
	if( !areSignsEqual )
	{
//		rotationalVelocity *= 0.3f;
		rotationalVelocity = (ROTATION_FRICTION_EFFICIENCY * targetRotationalVelocity) + ((1.0f - ROTATION_FRICTION_EFFICIENCY) * rotationalVelocity);
	}
	else
	{
//		rotationalVelocity = targetRotationalVelocity;
		rotationalVelocity = (ROTATION_FRICTION_EFFICIENCY * targetRotationalVelocity) + ((1.0f - ROTATION_FRICTION_EFFICIENCY) * rotationalVelocity);
	}
//	rotationalVelocity += rotationalVelocityDelta;

	g_TEMP_isPhysicsFrozen = true;
	g_TEMP_lineNormal = 100.0f * unitLineNormal;
	g_TEMP_lineDisp = 100.0f * unitLineDisp;
	g_TEMP_ballOldVel = oldVelTEMP;
	g_TEMP_ballNewVel = vel;
	g_TEMP_collisionVertex = NULL;
	g_TEMP_collisionLinedef = linedef;
}
*/





//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void Ball::Update()
{
	if( g_TEMP_physicsDebugMode && g_TEMP_enableFreezingPhysics && g_TEMP_isPhysicsFrozen )
	{
		return;
	}

	RunPhysics(); 
}


//-------------------------------------------------------------------------------------------------
// RunPhysics
//-------------------------------------------------------------------------------------------------
void Ball::RunPhysics() 
{ 
	float timeStep = (float)theApplication.GetFrameSeconds();
	timeStep = ClampFloat( timeStep, MIN_PHYSICS_TIME_DELTA, MAX_PHYSICS_TIME_DELTA );

	/// Apply constant gravity
	Vector2 gravityThisFrame = GRAVITY_ACCELERATION_VECTOR * timeStep;
	m_vel += gravityThisFrame;

	/// Update ball rotation angle based on angular velocity
	m_angleDegrees += (m_rotationalVelocity * timeStep);

	float physicsTimeLeft = timeStep;
	while( physicsTimeLeft > 0.0 ) 
	{ 
		RunPartialPhysics( physicsTimeLeft ); 
	}
} 


//-------------------------------------------------------------------------------------------------
// RunPartialPhysics
//-------------------------------------------------------------------------------------------------
void Ball::RunPartialPhysics( float& timeLeft ) 
{
	/// Leave a trail
	if( g_TEMP_pathSecondsLeft > 0.0 )
	{
		g_TEMP_pathSecondsLeft -= theApplication.GetFrameSeconds();
		g_TEMP_lastPath.push_back( m_pos );
	}

	Vector2 startingPos = m_pos;

	/// Determine if the ball is currently interpenetrating any linedefs or vertexes
	bool wasInterpenetrated = ResolveInterpenetrations();
	if( wasInterpenetrated )
	{
		theConsole.Printf( "...resolved interpenetration!\n" );
		g_TEMP_isCurrentlyInterpenetrating = true;
	}

	Vector2 move = m_vel * timeLeft; 
	LineSegment2 sweep( m_pos, move ); 
	float vertexCollisionFraction = FLT_MAX; 
	float linedefCollisionFraction = FLT_MAX; 

	MapVertex* collideVertex = theGame->GetCurrentMap().CheckForVertexCollision( sweep, m_radius, vertexCollisionFraction ); 
	MapLinedef* collideLinedef = theGame->GetCurrentMap().CheckForLinedefCollision( sweep, m_radius, linedefCollisionFraction ); 

	if( vertexCollisionFraction < linedefCollisionFraction ) 
	{ 
		// Collision with vertex 
		ResolveCollisionWithVertex( sweep, collideVertex, vertexCollisionFraction, timeLeft ); 
	} 
	else if( linedefCollisionFraction < vertexCollisionFraction ) 
	{ 
		// Collision with linedef 
		ResolveCollisionWithLinedef( sweep, collideLinedef, linedefCollisionFraction, timeLeft ); 
	} 
	else 
	{ 
		// No collision 
		ResolveNoCollision( sweep, timeLeft ); 
	} 

	/// Check if the ball collided (or would have collided) with any items
	std::list< Item* > collideItems;
	theGame->GetCurrentMap().CheckForCollisionAgainstItems( startingPos, m_radius, move, collideItems );
	if( !collideItems.empty() )
	{
		for( std::list< Item* >::iterator itemIter = collideItems.begin(); itemIter != collideItems.end(); ++itemIter )
		{
			if( (*itemIter)->IsCollectable() )
			{
				theGame->GetCurrentMap().RemoveItem( *itemIter );
			}
		}
	}
}       


//-------------------------------------------------------------------------------------------------
// ResolveCollisionWithLineSegment
//-------------------------------------------------------------------------------------------------
void Ball::ResolveCollisionWithLineSegment( const LineSegment2& discSweep, const LineSegment2& lineSeg, float timeFraction, float& timeLeft )
{
	/// Move the ball to a point just before the collision occurred 
	Vector2 partialMovement = discSweep.displacement * timeFraction * SCALE_BACK_COLLIDE_FRACTION_BALL_PLACEMENT; 
	m_pos += partialMovement; 

	/// Get a vector from linedef start to disc start 
	Vector2 vecToBall = discSweep.start - lineSeg.start; 

	/// Calculate unit parallel (displacement) and perpendicular (normal) vectors for the linedef 
	Vector2 unitParallel( lineSeg.displacement ); 
	unitParallel.Normalize(); 
	if( vecToBall % unitParallel < 0.0f ) // cross product
	{ 
		/// Linedef needs to be reversed 
		unitParallel *= -1.0f; 
	} 
	Vector2 unitPerpendicular( unitParallel.y, -unitParallel.x ); 

	/// Break the ball's motion into parallel and perpendicular components 
	float speedParallel             = m_vel | unitParallel; // dot product
	float speedPerpendicular        = m_vel | unitPerpendicular; // dot product

	/// Reflect and dampen perpendicular motion due to inelastic collision characteristics 
	speedPerpendicular *= -DAMPENING_MULTIPLIER_DUE_TO_INELASTIC_COLLISION; 

	/// Dampen parallel motion due to friction 
	speedParallel *= DAMPENING_MULTIPLIER_DUE_TO_FRICTION; 

	/// Recombine perpendicular and parallel components into new velocity 
	const Vector2 oldVelocity = m_vel;
	m_vel = (speedParallel * unitParallel) + (speedPerpendicular * unitPerpendicular); 

	/// Remove <timeFraction> time from <timeLeft> 
	timeLeft *= (1.0f - timeFraction); 

	/// Calculate impact energy and tell the game about it
	float ballEnergy = m_vel.CalcLengthSquared();
	float ballEnergyFraction = ballEnergy / (FULL_BALL_ENERGY * FULL_BALL_ENERGY);
	ballEnergyFraction = ClampFloat( ballEnergyFraction, 0.0f, 1.0f );
	theGame->BallBounce( ballEnergyFraction );

	/// Affect rotational velocity
	float targetRotationalVelocity = m_radius * speedParallel * 0.3f;
	m_rotationalVelocity = (ROTATION_FRICTION_EFFICIENCY * targetRotationalVelocity) + ((1.0f - ROTATION_FRICTION_EFFICIENCY) * m_rotationalVelocity);

	/// Physics debugging stuff
	g_TEMP_isPhysicsFrozen = true;
	g_TEMP_showCollisionPlanes = true;
	g_TEMP_lineNormal = 100.0f * unitPerpendicular;
	g_TEMP_lineDisp = 100.0f * unitParallel;
	g_TEMP_ballOldVel = oldVelocity;
	g_TEMP_ballNewVel = m_vel;
}
        

//-------------------------------------------------------------------------------------------------
// ResolveCollisionWithVertex
//-------------------------------------------------------------------------------------------------
void Ball::ResolveCollisionWithVertex( const LineSegment2& discSweep, MapVertex* vertex, float timeFraction, float& timeLeft )
{
	theConsole.Printf( "VERTEX Collision occurred; discSweep = (%g,%g),(%g,%g); vertex = (%g,%g); timeFraction = %g; timeLeft = %g seconds\n",
		discSweep.start.x, discSweep.start.y,
		discSweep.displacement.x, discSweep.displacement.y,
		vertex->m_pos.x, vertex->m_pos.y, timeFraction, timeLeft );

	g_TEMP_collisionVertex = vertex;
	g_TEMP_collisionLinedef = NULL;

	/// Create a virtual linedef off of which we are effectively being deflected
	const Vector2 vecToBall = discSweep.start - vertex->m_pos;
	const Vector2 parallelVirtualLineSeg( -vecToBall.y, vecToBall.x );
	LineSegment2 virtualLineSeg( vertex->m_pos - parallelVirtualLineSeg, parallelVirtualLineSeg + parallelVirtualLineSeg );

	/// Resolve this as a collision with that virtual line segment
	ResolveCollisionWithLineSegment( discSweep, virtualLineSeg, timeFraction, timeLeft );
} 


//-------------------------------------------------------------------------------------------------
// ResolveCollisionWithLinedef
//-------------------------------------------------------------------------------------------------
void Ball::ResolveCollisionWithLinedef( const LineSegment2& discSweep, MapLinedef* linedef, float timeFraction, float& timeLeft )
{ 
	theConsole.Printf( "LINEDEF Collision occurred; discSweep = (%g,%g),(%g,%g); linedef = (%g,%g)->(%g,%g); timeFraction = %g; timeLeft = %g seconds\n",
		discSweep.start.x, discSweep.start.y,
		discSweep.displacement.x, discSweep.displacement.y,
		linedef->m_start->m_pos.x, linedef->m_start->m_pos.y,
		linedef->m_end->m_pos.x, linedef->m_end->m_pos.y,
		timeFraction, timeLeft );

	g_TEMP_collisionVertex = NULL;
	g_TEMP_collisionLinedef = linedef;

	LineSegment2 lineSeg( linedef->m_start->m_pos, linedef->m_end->m_pos - linedef->m_start->m_pos ); 
	ResolveCollisionWithLineSegment( discSweep, lineSeg, timeFraction, timeLeft );
} 


//-------------------------------------------------------------------------------------------------
// ResolveNoCollision
//-------------------------------------------------------------------------------------------------        
void Ball::ResolveNoCollision( const LineSegment2& sweep, float& timeLeft ) 
{ 
	m_pos += sweep.displacement; 
	timeLeft = 0.0f; 
} 


