
#include "pch.hpp"
#include "gameCommon.hpp"
#include "Map.hpp"
#include "TheGame.hpp"


Map::Map( const std::string& mapName )
:
	_texBackdrop( "data/maps/" + mapName + ".tga" ),
	mapArea( mapName, true ),
	collectableItemsRemain( 0 )
{
	// create Items from MapObjects
	for( std::list<MapObject*>::iterator itemIter = mapArea.m_objectList.begin(); 
		itemIter != mapArea.m_objectList.end(); 
		++itemIter )
	{
		Item* newItem = new Item( *itemIter );
		items.push_back( newItem );
		if( newItem->IsCollectable() )
		{
			++collectableItemsRemain;
		}

		if ( newItem->IsBallSpawn() )
		{
			theGame->AddBall( newItem->pos );
		}
	}

	// create Walls from MapLinedefs
	for( std::list<MapLinedef*>::iterator wallIter = mapArea.m_linedefList.begin(); 
		wallIter != mapArea.m_linedefList.end(); 
		++wallIter )
	{
		walls.push_back( new Wall( *wallIter ) );
	}
}

Map::~Map()
{
}

void Map::Draw()
{
	glDisable( GL_BLEND );
	Vector2 mins( 0, 0 );
	Vector2 texMins( mins );
	Vector2 maxs( 1024, 1024 );
	Vector2 texMaxs( 1, 1 );
	theRenderer.DrawTexturedRect2d( _texBackdrop, mins, maxs, texMins, texMaxs );


	for( unsigned int i = 0; i < items.size(); i++ )
	{
		items[ i ]->Draw();
	}

	if( g_TEMP_physicsDebugMode )
	{
		glLineWidth( 3.0f );
		std::list<MapLinedef*>::iterator begin = mapArea.m_linedefList.begin();
		std::list<MapLinedef*>::iterator end = mapArea.m_linedefList.end();
		std::list<MapLinedef*>::iterator iter;

		for( iter = begin; iter != end; ++ iter )
		{
			MapLinedef* linedef = (*iter);
			LineSegment2 lineSeg( linedef->m_start->m_pos, linedef->m_end->m_pos - linedef->m_start->m_pos );
			theRenderer.DrawVector2d( lineSeg, Rgba::RED, Rgba::GREEN );
		}
		glLineWidth( 1.0f );
	}
}

/*
//-------------------------------------------------------------------------------------------------
// CheckForCollisionAgainstLinedefs
//-------------------------------------------------------------------------------------------------
MapLinedef* Map::CheckForCollisionAgainstLinedefs( const Vector2& pos, float radius, const Vector2& movement, float& timeFraction )
{
	const float LINEDEF_COLLIDE_RADIUS = 0.0f;
	MapLinedef* collideLinedef = NULL;
	float collideFraction = FLT_MAX;

	std::list<MapLinedef*>::iterator begin = mapArea.m_linedefList.begin();
	std::list<MapLinedef*>::iterator end = mapArea.m_linedefList.end();
	std::list<MapLinedef*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapLinedef* linedef = (*iter);
		LineSegment2 lineSeg( linedef->m_start->m_pos, linedef->m_end->m_pos - linedef->m_start->m_pos );
		if( DoMovingDiscAndLineSegmentCollide2d( pos, radius, movement, lineSeg ) )
		{
			/// We have a potential collision
			float fraction = TraceMovingDiscAndLineCollide2d( pos, movement, radius, lineSeg, Vector2::ZERO, LINEDEF_COLLIDE_RADIUS );
			if( fraction < collideFraction )
			{
				collideFraction = fraction;
				collideLinedef = linedef;
			}
		}
	}

	timeFraction = collideFraction;
	return collideLinedef;
}


//-------------------------------------------------------------------------------------------------
// CheckForCollisionAgainstVertexes
//-------------------------------------------------------------------------------------------------
MapVertex* Map::CheckForCollisionAgainstVertexes( const Vector2& pos, float radius, const Vector2& movement, float& timeFraction )
{
	const float VERTEX_COLLIDE_RADIUS = 0.0f;
	MapVertex* collideVertex = NULL;
	float collideFraction = FLT_MAX;

	std::list<MapVertex*>::iterator begin = mapArea.m_vertexList.begin();
	std::list<MapVertex*>::iterator end = mapArea.m_vertexList.end();
	std::list<MapVertex*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* vertex = (*iter);
		if( DoMovingDiscsCollide2d( pos, movement, radius, vertex->m_pos, VERTEX_COLLIDE_RADIUS ) )
		{
			/// We have a potential collision
			float fraction = TraceMovingDiscsCollide2d( pos, movement, radius, vertex->m_pos, Vector2::ZERO, VERTEX_COLLIDE_RADIUS );
			if( fraction < collideFraction )
			{
				collideFraction = fraction;
				collideVertex = vertex;
			}
		}
	}

	timeFraction = collideFraction;
	return collideVertex;
}

*/


//-------------------------------------------------------------------------------------------------
// CheckForCollisionAgainstItems
//-------------------------------------------------------------------------------------------------
Item* Map::CheckForCollisionAgainstItems(  const Vector2& pos, float radius, const Vector2& movement, std::list< Item* >& collideItems )
{
	collideItems.clear();

	for( std::vector< Item* >::iterator itemIter = items.begin(); itemIter != items.end(); ++itemIter )
	{
		Item* const item = *itemIter;
		bool collision = DoMovingDiscsCollide2d( pos, movement, radius, item->pos, item->radius );
		if( collision )
		{
			collideItems.push_back( item );
		}
	}
	return 0;
}


//-------------------------------------------------------------------------------------------------
// RemoveItem
//-------------------------------------------------------------------------------------------------
void Map::RemoveItem( Item* item )
{
	std::vector< Item* >::iterator itemIter = std::find( items.begin(), items.end(), item );
	ASSERT( itemIter != items.end() );
	items.erase( itemIter );
	
	if( item->IsCollectable() )
	{
		theGame->CollectItem();
		--collectableItemsRemain;
		ASSERT( collectableItemsRemain >= 0 );
	}
}



//-------------------------------------------------------------------------------------------------
// CheckForVertexCollision
//-------------------------------------------------------------------------------------------------
MapVertex* Map::CheckForVertexCollision( const LineSegment2& discSweep, float radius, float& timeFraction )
{
	const Vector2& discStart = discSweep.start;
	const Vector2& discMove = discSweep.displacement;

	timeFraction = FLT_MAX;
	MapVertex* collisionElement = NULL;

	std::list<MapVertex*>::iterator begin = mapArea.m_vertexList.begin();
	std::list<MapVertex*>::iterator end = mapArea.m_vertexList.end();
	std::list<MapVertex*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* element = (*iter);
		if( DoDiscAndLineSegmentOverlap2d( element->m_pos, radius, discSweep ) )
		{
			float fraction = TraceMovingDiscsCollide2d( discStart, discMove, radius, element->m_pos, Vector2::ZERO, 0.0f );
			if( fraction < timeFraction )
			{
				timeFraction = fraction;
				collisionElement = element;
			}
		}
	}

	return collisionElement;
}


//-------------------------------------------------------------------------------------------------
// CheckForLinedefCollision
//-------------------------------------------------------------------------------------------------
MapLinedef* Map::CheckForLinedefCollision( const LineSegment2& discSweep, float radius, float& timeFraction )
{
	timeFraction = FLT_MAX;
	MapLinedef* collisionElement = NULL;

	std::list<MapLinedef*>::iterator begin = mapArea.m_linedefList.begin();
	std::list<MapLinedef*>::iterator end = mapArea.m_linedefList.end();
	std::list<MapLinedef*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapLinedef* element = (*iter);
		LineSegment2 lineSeg( element->m_start->m_pos, element->m_end->m_pos - element->m_start->m_pos );
		float traceFraction;
		bool collision = HackTraceMovingDiscAndLineSegmentCollide2d( discSweep, radius, lineSeg, traceFraction );
		if( collision && (traceFraction < timeFraction) )
		{
			timeFraction = traceFraction;
			collisionElement = element;
		}
	}

	return collisionElement;
}



//-------------------------------------------------------------------------------------------------
// HackTraceMovingDiscAndLineSegmentCollide2d
//-------------------------------------------------------------------------------------------------
bool Map::HackTraceMovingDiscAndLineSegmentCollide2d( const LineSegment2& discSweep, float radius, const LineSegment2& lineSeg, float& traceFraction )
{
	const Vector2& lineStart = lineSeg.start;
	const Vector2 lineEnd = lineStart + lineSeg.displacement;
	Vector2 unitParallel( lineSeg.displacement );
	unitParallel.Normalize();
	Vector2 unitPerpendicular( unitParallel.y, -unitParallel.x );
	Vector2 radiusPerpendicular = radius * unitPerpendicular;

	Vector2 point1 = lineStart + radiusPerpendicular;
	Vector2 point2 = lineEnd + radiusPerpendicular;
	Vector2 point3 = lineStart - radiusPerpendicular;
	Vector2 point4 = lineEnd - radiusPerpendicular;

	LineSegment2 edge12( point1, point2 - point1 );
	LineSegment2 edge34( point3, point4 - point3 );

	float fraction12, fraction34;
	bool collision12 = TraceMovingPointAndLineSegmentCollide2d( discSweep, edge12, fraction12 );
	bool collision34 = TraceMovingPointAndLineSegmentCollide2d( discSweep, edge34, fraction34 );

	if( collision12 && collision34 )
	{
		traceFraction = MinFloat( fraction12, fraction34 );
		return true;
	}
	else if( collision12 )
	{
		traceFraction = fraction12;
		return true;
	}
	else if( collision34 )
	{
		traceFraction = fraction34;
		return true;
	}
	else
	{
		traceFraction = FLT_MAX;
		return false;
	}
}


