//-------------------------------------------------------------------------------------------------
// MapArea.cpp (game)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "Grid.hpp"
#include "MapArea.hpp"


//-------------------------------------------------------------------------------------------------
// Globals
//-------------------------------------------------------------------------------------------------
MapArea* g_currentArea = NULL;


//-------------------------------------------------------------------------------------------------
// Destructor()
//-------------------------------------------------------------------------------------------------
MapArea::~MapArea()
{
}


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
MapArea::MapArea( const std::string& mapName, bool isMapFlippedVertically )
:
m_isMapFlippedVertically( isMapFlippedVertically ),
m_mapName( mapName )
{
	std::string mapPath( "data/maps/" );
	mapPath += mapName;
	mapPath += ".map";
	PopulateFromFile( mapPath );
}


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
MapArea::MapArea()
:
m_isMapFlippedVertically( false )
{
}


//-------------------------------------------------------------------------------------------------
// ClearHighlight
//-------------------------------------------------------------------------------------------------
void MapArea::ClearHighlight()
{
	m_highlightList.clear();
}


//-------------------------------------------------------------------------------------------------
// ClearSelection
//-------------------------------------------------------------------------------------------------
void MapArea::ClearSelection()
{
	std::list<MapElement*>::iterator begin = m_selectionList.begin();
	std::list<MapElement*>::iterator end = m_selectionList.end();
	std::list<MapElement*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapElement* element = (*iter);
		element->m_isSelected = false;
	}

	m_selectionList.clear();
}


//-------------------------------------------------------------------------------------------------
// IntegrateVertexIntoWorld
//-------------------------------------------------------------------------------------------------
MapVertex* MapArea::IntegrateVertexIntoWorld( MapVertex* integrateVertex )
{
	/// See if this vertex is on top of an already existing vertex
	std::list<MapVertex*>::iterator begin = m_vertexList.begin();
	std::list<MapVertex*>::iterator end = m_vertexList.end();
	std::list<MapVertex*>::iterator iter;
	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* vertex = (*iter);
		if( vertex == integrateVertex )
		{
			continue;
		}

		if( IsDistanceBetweenPointsLess( integrateVertex->m_pos, vertex->m_pos, Grid::GetGridEpsilon() ) )
		{
			/// Transfer linedef ownership to existing vertex
			std::list<MapLinedef*>::iterator line_begin = integrateVertex->m_linedefs.begin();
			std::list<MapLinedef*>::iterator line_end = integrateVertex->m_linedefs.end();
			std::list<MapLinedef*>::iterator line_iter;
			for( line_iter = line_begin; line_iter != line_end; ++ line_iter )
			{
				MapLinedef* linedef = (*line_iter);
				if( linedef->m_start == integrateVertex )
				{
					linedef->m_start = vertex;
				}
				else if( linedef->m_end == integrateVertex )
				{
					linedef->m_end = vertex;
				}
				vertex->m_linedefs.push_back( linedef );
			}

			/// Strip this vertex of all linedef associations
			integrateVertex->m_linedefs.clear();
			DeleteMapVertex( integrateVertex );
			return vertex;
		}
	}

	return integrateVertex;
}


//-------------------------------------------------------------------------------------------------
// CreateNewVertex
//-------------------------------------------------------------------------------------------------
MapVertex* MapArea::CreateNewVertex( const Vector2& pos, bool allowDuplicatePosition )
{
	if( !allowDuplicatePosition )
	{
		/// Make sure this vertex isn't on top of an already existing vertex
		std::list<MapVertex*>::iterator begin = m_vertexList.begin();
		std::list<MapVertex*>::iterator end = m_vertexList.end();
		std::list<MapVertex*>::iterator iter;
		for( iter = begin; iter != end; ++ iter )
		{
			MapVertex* vertex = (*iter);
			if( IsDistanceBetweenPointsLess( pos, vertex->m_pos, Grid::GetGridEpsilon() ) )
			{
				/// Another vertex is already present; skip creating one and return it instead
				return vertex;
			}
		}
	}

	/// Create the new vertex
	MapVertex* newVertex = new MapVertex( pos );
	m_vertexList.push_back( newVertex );
	return newVertex;
}


//-------------------------------------------------------------------------------------------------
// CreateNewLinedef
//-------------------------------------------------------------------------------------------------
MapLinedef* MapArea::CreateNewLinedef( MapVertex* start, MapVertex* end )
{
	MapLinedef* newLinedef = new MapLinedef( start, end );
	m_linedefList.push_back( newLinedef );
	start->m_linedefs.push_back( newLinedef );
	end->m_linedefs.push_back( newLinedef );
	return newLinedef;
}


//-------------------------------------------------------------------------------------------------
// CreateNewObject
//-------------------------------------------------------------------------------------------------
MapObject* MapArea::CreateNewObject( const Vector2& pos )
{
	MapObject* newObject = new MapObject( pos );
	m_objectList.push_back( newObject );
	return newObject;
}


//-------------------------------------------------------------------------------------------------
// CreateNewObject
//-------------------------------------------------------------------------------------------------
MapObject* MapArea::CreateNewObject( const MapObject& objectToCopy, const Vector2& pos )
{
	MapObject* newObject = new MapObject( objectToCopy, pos );
	m_objectList.push_back( newObject );
	return newObject;
}


//-------------------------------------------------------------------------------------------------
// DeleteMapLinedef
//-------------------------------------------------------------------------------------------------
void MapArea::DeleteMapLinedef( MapLinedef* killLinedef, MapVertex* killerVertex )
{
	MapVertex* start = killLinedef->m_start;
	MapVertex* end = killLinedef->m_end;

	/// Find the linedef in the area linedef list
	std::list<MapLinedef*>::iterator line_begin = m_linedefList.begin();
	std::list<MapLinedef*>::iterator line_end = m_linedefList.end();
	std::list<MapLinedef*>::iterator line_iter;
	line_iter = std::find( line_begin, line_end, killLinedef );
	if( line_iter == line_end )
	{
		return;
	}

	/// Check if this was the only linedef connected to its start vertex
	if( start != killerVertex )
	{
		if( start->m_linedefs.size() == 1 )
		{
			DeleteMapVertex( start, killLinedef );
		}
		else
		{
			/// Remove this linedef from the surviving vertex's list
			std::list<MapLinedef*>::iterator list_begin = start->m_linedefs.begin();
			std::list<MapLinedef*>::iterator list_end = start->m_linedefs.end();
			std::list<MapLinedef*>::iterator list_iter;
			list_iter = std::find( list_begin, list_end, killLinedef );
			if( list_iter != list_end )
			{
				start->m_linedefs.erase( list_iter );
			}
		}
	}

	/// Check if this was the only linedef connected to its end vertex
	if( end != killerVertex )
	{
		if( end->m_linedefs.size() == 1 )
		{
			DeleteMapVertex( end, killLinedef );
		}
		else
		{
			/// Remove this linedef from the surviving vertex's list
			std::list<MapLinedef*>::iterator list_begin = end->m_linedefs.begin();
			std::list<MapLinedef*>::iterator list_end = end->m_linedefs.end();
			std::list<MapLinedef*>::iterator list_iter;
			list_iter = std::find( list_begin, list_end, killLinedef );
			if( list_iter != list_end )
			{
				end->m_linedefs.erase( list_iter );
			}
		}
	}

	/// Remove it from the list and destroy it
	m_linedefList.erase( line_iter );
	delete killLinedef;
}


//-------------------------------------------------------------------------------------------------
// DeleteMapVertex
//-------------------------------------------------------------------------------------------------
void MapArea::DeleteMapVertex( MapVertex* killVertex, MapLinedef* killerLinedef )
{
	/// Find the vertex in the area vertex list
	std::list<MapVertex*>::iterator begin = m_vertexList.begin();
	std::list<MapVertex*>::iterator end = m_vertexList.end();
	std::list<MapVertex*>::iterator iter;
	iter = std::find( begin, end, killVertex );
	if( iter == end )
	{
		return;
	}

	/// FIXME: also check for and destroy all linedefs attached to this vertex!
	std::list<MapLinedef*>::iterator line_begin = killVertex->m_linedefs.begin();
	std::list<MapLinedef*>::iterator line_end = killVertex->m_linedefs.end();
	std::list<MapLinedef*>::iterator line_iter;
	for( line_iter = line_begin; line_iter != line_end; ++ line_iter )
	{
		MapLinedef* linedef = (*line_iter);
		if( linedef != killerLinedef )
		{
			DeleteMapLinedef( linedef, killVertex );
		}
	}

	/// Remove it from the list and destroy it
	m_vertexList.erase( iter );
	delete killVertex;
}


//-------------------------------------------------------------------------------------------------
// DeleteMapObject
//-------------------------------------------------------------------------------------------------
void MapArea::DeleteMapObject( MapObject* killObject )
{
	/// Find the object in the area object list
	std::list<MapObject*>::iterator begin = m_objectList.begin();
	std::list<MapObject*>::iterator end = m_objectList.end();
	std::list<MapObject*>::iterator iter;
	iter = std::find( begin, end, killObject );
	if( iter == end )
	{
		return;
	}

	/// Remove it from the list and destroy it
	m_objectList.erase( iter );
	delete killObject;
}


//-------------------------------------------------------------------------------------------------
// DeleteMapElement
//-------------------------------------------------------------------------------------------------
void MapArea::DeleteMapElement( MapElement* element )
{
	if( element->m_type == MapElement::ELEMENT_TYPE_VERTEX )
	{
		DeleteMapVertex( reinterpret_cast<MapVertex*>(element) );
	}
	else if( element->m_type == MapElement::ELEMENT_TYPE_LINEDEF )
	{
		DeleteMapLinedef( reinterpret_cast<MapLinedef*>(element) );
	}
	else if( element->m_type == MapElement::ELEMENT_TYPE_OBJECT )
	{
		DeleteMapObject( reinterpret_cast<MapObject*>(element) );
	}
}


//-------------------------------------------------------------------------------------------------
// DeleteAllElements
//-------------------------------------------------------------------------------------------------
void MapArea::DeleteAllElements()
{
	m_highlightList.clear();
	m_selectionList.clear();

	while( m_vertexList.size() > 0 )
	{
		DeleteMapVertex( m_vertexList.front() );
	}

	while( m_linedefList.size() > 0 )
	{
		DeleteMapLinedef( m_linedefList.front() );
	}

	while( m_objectList.size() > 0 )
	{
		DeleteMapObject( m_objectList.front() );
	}
}


//-------------------------------------------------------------------------------------------------
// SelectElement
//-------------------------------------------------------------------------------------------------
void MapArea::SelectElement( MapElement* element )
{
	if( element->m_isSelected )
	{
		return;
	}

	m_selectionList.push_front( element );
	element->m_isSelected = true;
}


//-------------------------------------------------------------------------------------------------
// DeselectElement
//-------------------------------------------------------------------------------------------------
void MapArea::DeselectElement( MapElement* element )
{
	if( !element->m_isSelected )
	{
		return;
	}

	std::list<MapElement*>::iterator begin = m_selectionList.begin();
	std::list<MapElement*>::iterator end = m_selectionList.end();
	std::list<MapElement*>::iterator iter;
	iter = std::find( begin, end, element );
	VALIDATE( iter != end );
	m_selectionList.erase( iter );
	element->m_isSelected = false;
}


//-------------------------------------------------------------------------------------------------
// ToggleSelectionOfHighlightedElements
//-------------------------------------------------------------------------------------------------
void MapArea::ToggleSelectionOfHighlightedElements()
{
	std::list<MapElement*>::iterator begin = m_highlightList.begin();
	std::list<MapElement*>::iterator end = m_highlightList.end();
	std::list<MapElement*>::iterator iter;
	for( iter = begin; iter != end; ++ iter )
	{
		MapElement* element = (*iter);

		/// Disallow selection of vertexes (FIXME: may want to allow this after all)
		if( element->m_type == MapElement::ELEMENT_TYPE_VERTEX )
			continue;

		if( element->m_isSelected )
		{
			DeselectElement( element );
		}
		else
		{
			SelectElement( element );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// HighlightElement
//-------------------------------------------------------------------------------------------------
void MapArea::HighlightElement( MapElement* element )
{
	element->m_highlightFrameNumber = theApplication.GetFrameNumber();
	m_highlightList.push_back( element );
}


//-------------------------------------------------------------------------------------------------
// DeleteHighlightedAndSelected
//-------------------------------------------------------------------------------------------------
void MapArea::DeleteHighlightedAndSelected()
{
	if( !m_highlightList.empty() )
	{
		SelectElement( m_highlightList.front() );
	}
/*
	while( !m_highlightList.empty() )
	{
		MapElement* element = m_highlightList.front();
		DeleteMapElement( element );
		m_highlightList.pop_front();
	}
*/
	while( !m_selectionList.empty() )
	{
		MapElement* element = m_selectionList.front();
		DeleteMapElement( element );
		m_selectionList.pop_front();
	}
}


//-------------------------------------------------------------------------------------------------
// TranslateSelection
//-------------------------------------------------------------------------------------------------
void MapArea::TranslateSelection( const Vector2& translation )
{
	std::list<MapElement*>::iterator begin = m_selectionList.begin();
	std::list<MapElement*>::iterator end = m_selectionList.end();
	std::list<MapElement*>::iterator iter;
	std::list<MapElement*> translationList;

	/// Unmark all elements (none has moved yet)
	for( iter = begin; iter != end; ++ iter )
	{
		MapElement* element = (*iter);
		element->m_isMarked = false;
		if( element->m_type == MapElement::ELEMENT_TYPE_LINEDEF )
		{
			MapLinedef* linedef = reinterpret_cast<MapLinedef*>( element );
			linedef->m_start->m_isMarked = false;
			linedef->m_end->m_isMarked = false;
		}
	}

	/// Build a list of all vertexes that will be moving
	for( iter = begin; iter != end; ++ iter )
	{
		MapElement* element = (*iter);
		if( element->m_type == MapElement::ELEMENT_TYPE_VERTEX )
		{
			if( !element->m_isMarked )
			{
				translationList.push_back( element );
				element->m_isMarked = true;
			}
		}
		else if( element->m_type == MapElement::ELEMENT_TYPE_LINEDEF )
		{
			MapLinedef* linedef = reinterpret_cast<MapLinedef*>( element );
			if( !linedef->m_start->m_isMarked )
			{
				translationList.push_back( linedef->m_start );
				linedef->m_start->m_isMarked = true;
			}
			if( !linedef->m_end->m_isMarked )
			{
				translationList.push_back( linedef->m_end );
				linedef->m_end->m_isMarked = true;
			}
		}
	}

	/// Add selected objects to translation list
	begin = m_selectionList.begin();
	end = m_selectionList.end();
	for( iter = begin; iter != end; ++ iter )
	{
		MapElement* element = (*iter);
		if( element->m_type == MapElement::ELEMENT_TYPE_OBJECT )
		{
			translationList.push_back( *iter );
		}
	}

	/// Move all elements in the translation list
	std::list<MapElement*>::iterator transBegin = translationList.begin();
	std::list<MapElement*>::iterator transEnd = translationList.end();
	std::list<MapElement*>::iterator transIter;
	for( transIter = transBegin; transIter != transEnd; ++ transIter )
	{
		MapElement* element = (*transIter);
		element->Translate( translation );
	}
}


//-------------------------------------------------------------------------------------------------
// ParseVertexes
//-------------------------------------------------------------------------------------------------
void MapArea::ParseVertexes( ParseFile& mapFile )
{
	mapFile.ParseExpectedToken( "Vertexes" );
	mapFile.ParseExpectedToken( '{' );
	char nextChar = mapFile.AdvanceToNextToken();
	while( nextChar != '}' )
	{
		ParseNewVertex( mapFile );
		nextChar = mapFile.AdvanceToNextToken();
	}
	mapFile.ParseExpectedToken( '}' );
}


//-------------------------------------------------------------------------------------------------
// ParseLinedefs
//-------------------------------------------------------------------------------------------------
void MapArea::ParseLinedefs( ParseFile& mapFile, int vertIndexOffset )
{
	mapFile.ParseExpectedToken( "Linedefs" );
	mapFile.ParseExpectedToken( '{' );
	char nextChar = mapFile.AdvanceToNextToken();
	while( nextChar != '}' )
	{
		ParseNewLinedef( mapFile, vertIndexOffset );
		nextChar = mapFile.AdvanceToNextToken();
	}
	mapFile.ParseExpectedToken( '}' );
}


//-------------------------------------------------------------------------------------------------
// ParseObjects
//-------------------------------------------------------------------------------------------------
void MapArea::ParseObjects( ParseFile& mapFile )
{
	mapFile.ParseExpectedToken( "Objects" );
	mapFile.ParseExpectedToken( '{' );
	char nextChar = mapFile.AdvanceToNextToken();
	while( nextChar != '}' )
	{
		ParseNewObject( mapFile );
		nextChar = mapFile.AdvanceToNextToken();
	}
	mapFile.ParseExpectedToken( '}' );
}


//-------------------------------------------------------------------------------------------------
// PopulateFromFile
//-------------------------------------------------------------------------------------------------
bool MapArea::PopulateFromFile( const std::string& mapFilePath )
{
	int numInitialVerts = m_vertexList.size();

	ParseFile mapFile( mapFilePath );
	mapFile.StripComments();

	ParseVertexes( mapFile );
	ParseLinedefs( mapFile, numInitialVerts );
	ParseObjects( mapFile );

	return true;
}


//-------------------------------------------------------------------------------------------------
// ParseNewVertex
//-------------------------------------------------------------------------------------------------
void MapArea::ParseNewVertex( ParseFile& mapFile )
{
	Vector2 pos = mapFile.ParseVector2();
	if( m_isMapFlippedVertically )
	{
		pos.y = 1024.0f - pos.y;
	}
	CreateNewVertex( pos, true );
}


//-------------------------------------------------------------------------------------------------
// ParseNewLinedef
//-------------------------------------------------------------------------------------------------
void MapArea::ParseNewLinedef( ParseFile& mapFile, int vertIndexOffset )
{
	int startVertIndex = vertIndexOffset + mapFile.ParseInt();
	mapFile.ParseExpectedToken( ',' );
	int endVertIndex = vertIndexOffset + mapFile.ParseInt();

	MapVertex* start = GetNthVertex( startVertIndex );
	MapVertex* end = GetNthVertex( endVertIndex );

	MapLinedef* linedef = CreateNewLinedef( start, end );
	char nextChar = mapFile.AdvanceToNextToken();
	if( nextChar != '{' )
	{
		return;
	}

	/// Parse key-value properties
	mapFile.ParseExpectedToken( '{' );
	while( nextChar != '}' )
	{
		std::string key;
		std::string value;
		mapFile.ParseQuotedString( key );
		mapFile.ParseQuotedString( value );
		linedef->m_properties.SetKeyValue( key, value );
		nextChar = mapFile.AdvanceToNextToken();
	}
	mapFile.ParseExpectedToken( '}' );
}


//-------------------------------------------------------------------------------------------------
// ParseNewObject
//-------------------------------------------------------------------------------------------------
void MapArea::ParseNewObject( ParseFile& mapFile )
{
	Vector2 pos = mapFile.ParseVector2();
	if( m_isMapFlippedVertically )
	{
		pos.y = 1024.0f - pos.y;
	}

	MapObject* object = CreateNewObject( pos );
	char nextChar = mapFile.AdvanceToNextToken();
	if( nextChar != '{' )
	{
		return;
	}

	/// Parse key-value properties
	mapFile.ParseExpectedToken( '{' );
	while( nextChar != '}' )
	{
		std::string key;
		std::string value;
		mapFile.ParseQuotedString( key );
		mapFile.ParseQuotedString( value );
		object->m_properties.SetKeyValue( key, value );
		nextChar = mapFile.AdvanceToNextToken();
	}
	mapFile.ParseExpectedToken( '}' );
}


//-------------------------------------------------------------------------------------------------
// SaveVertexesToFile
//-------------------------------------------------------------------------------------------------
void MapArea::SaveVertexesToFile( FILE* file )
{
	fprintf( file, "Vertexes\n" );
	fprintf( file, "{\n" );

	std::list<MapVertex*>::iterator begin = m_vertexList.begin();
	std::list<MapVertex*>::iterator end = m_vertexList.end();
	std::list<MapVertex*>::iterator iter;
	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* vertex = (*iter);
		fprintf( file, "\t(%g, %g)\n", vertex->m_pos.x, vertex->m_pos.y );
	}

	fprintf( file, "}\n\n" );
}


//-------------------------------------------------------------------------------------------------
// SaveLinedefsToFile
//-------------------------------------------------------------------------------------------------
void MapArea::SaveLinedefsToFile( FILE* file )
{
	fprintf( file, "Linedefs\n" );
	fprintf( file, "{\n" );

	std::list<MapLinedef*>::iterator begin = m_linedefList.begin();
	std::list<MapLinedef*>::iterator end = m_linedefList.end();
	std::list<MapLinedef*>::iterator iter;
	for( iter = begin; iter != end; ++ iter )
	{
		MapLinedef* linedef = (*iter);
		int startVertIndex = GetVertIndex( linedef->m_start );
		int endVertIndex = GetVertIndex( linedef->m_end );
		fprintf( file, "\t%d, %d\n", startVertIndex, endVertIndex );
		SavePropertiesToFile( linedef->m_properties, file );
	}

	fprintf( file, "}\n\n" );
}


//-------------------------------------------------------------------------------------------------
// SaveObjectsToFile
//-------------------------------------------------------------------------------------------------
void MapArea::SaveObjectsToFile( FILE* file )
{
	fprintf( file, "Objects\n" );
	fprintf( file, "{\n" );

	std::list<MapObject*>::iterator begin = m_objectList.begin();
	std::list<MapObject*>::iterator end = m_objectList.end();
	std::list<MapObject*>::iterator iter;
	for( iter = begin; iter != end; ++ iter )
	{
		MapObject* object = (*iter);
		fprintf( file, "\t(%g, %g)\n", object->m_pos.x, object->m_pos.y );
		SavePropertiesToFile( object->m_properties, file );
	}

	fprintf( file, "}\n\n" );
}


//-------------------------------------------------------------------------------------------------
// SavePropertiesToFile
//-------------------------------------------------------------------------------------------------
void MapArea::SavePropertiesToFile( const MapProperties& properties, FILE* file )
{
	if( properties.GetNumKeyValues() > 0 )
	{
		fprintf( file, "\t{\n" );
		MapProperties::KeyPairMap::const_iterator begin = properties.m_keyValuePairs.begin();
		MapProperties::KeyPairMap::const_iterator end = properties.m_keyValuePairs.end();
		MapProperties::KeyPairMap::const_iterator iter;
		for( iter = begin; iter != end; ++ iter )
		{
			std::string key = iter->first;
			std::string value = iter->second;
			fprintf( file, "\t\t\"%s\" \"%s\"\n", key.c_str(), value.c_str() );
		}
		fprintf( file, "\t}\n" );
	}
}


//-------------------------------------------------------------------------------------------------
// SaveToFile
//-------------------------------------------------------------------------------------------------
bool MapArea::SaveToFile( const std::string& mapFilePath )
{
	FILE* file = fopen( mapFilePath.c_str(), "wt" );
	if( !file )
	{
		return false;
	}

	SaveVertexesToFile( file );
	SaveLinedefsToFile( file );
	SaveObjectsToFile( file );

	fclose( file );
	return true;
}


//-------------------------------------------------------------------------------------------------
// GetNthVertex
//-------------------------------------------------------------------------------------------------
MapVertex* MapArea::GetNthVertex( int vertIndex )
{
	std::list<MapVertex*>::iterator begin = m_vertexList.begin();
	std::list<MapVertex*>::iterator end = m_vertexList.end();
	std::list<MapVertex*>::iterator iter;

	int index = 0;
	for( iter = begin; iter != end; ++ iter )
	{
		if( index == vertIndex )
		{
			return (*iter);
		}

		index ++;
	}

	return NULL;
}


//-------------------------------------------------------------------------------------------------
// GetVertIndex
//-------------------------------------------------------------------------------------------------
int MapArea::GetVertIndex( const MapVertex* vertex )
{
	std::list<MapVertex*>::iterator begin = m_vertexList.begin();
	std::list<MapVertex*>::iterator end = m_vertexList.end();
	std::list<MapVertex*>::iterator iter;

	int index = 0;
	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* checkVertex = (*iter);
		if( checkVertex == vertex )
		{
			return index;
		}

		index ++;
	}

	return -1; // FIXME: this is dangerous and probably stupid
}


//-------------------------------------------------------------------------------------------------
// NumSelectedElements
//-------------------------------------------------------------------------------------------------
int MapArea::NumSelectedElements()
{
	return m_selectionList.size();
}


//-------------------------------------------------------------------------------------------------
// NumSelectedElementsOfType
//-------------------------------------------------------------------------------------------------
int MapArea::NumSelectedElementsOfType( const MapElement::ElementType& type )
{
	int count = 0;

	std::list<MapElement*>::iterator begin = m_selectionList.begin();
	std::list<MapElement*>::iterator end = m_selectionList.end();
	std::list<MapElement*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapElement* element = (*iter);
		if( element->m_type == type )
		{
			count ++;
		}
	}

	return count;
}



//-------------------------------------------------------------------------------------------------
// SetPropertyForHighlightedElements
//-------------------------------------------------------------------------------------------------
void MapArea::SetPropertyForHighlightedElements( const std::string& key, const std::string& value )
{
	if( m_highlightList.size() == 0 )
	{
		return;
	}

	MapElement* element = m_highlightList.front();
	if( element->m_type == MapElement::ELEMENT_TYPE_OBJECT )
	{
		MapObject* object = reinterpret_cast<MapObject*>( element );
		object->m_properties.SetKeyValue( key, value );
	}
	else if( element->m_type == MapElement::ELEMENT_TYPE_LINEDEF )
	{
		MapLinedef* linedef = reinterpret_cast<MapLinedef*>( element );
		linedef->m_properties.SetKeyValue( key, value );
	}
}


//-------------------------------------------------------------------------------------------------
// DeletePropertyForHighlightedElements
//-------------------------------------------------------------------------------------------------
void MapArea::DeletePropertyForHighlightedElements( const std::string& key )
{
	if( m_highlightList.size() == 0 )
	{
		return;
	}

	MapElement* element = m_highlightList.front();
	if( element->m_type == MapElement::ELEMENT_TYPE_OBJECT )
	{
		MapObject* object = reinterpret_cast<MapObject*>( element );
		object->m_properties.DeleteKey( key );
	}
	else if( element->m_type == MapElement::ELEMENT_TYPE_LINEDEF )
	{
		MapLinedef* linedef = reinterpret_cast<MapLinedef*>( element );
		linedef->m_properties.DeleteKey( key );
	}
}


//-------------------------------------------------------------------------------------------------
// FindClosestMapElementToWorldCoords
//-------------------------------------------------------------------------------------------------
MapElement* MapArea::FindClosestMapElementToWorldCoords( Vector2 worldCoords, float& distSquaredToElement )
{
	return NULL;
}


//-------------------------------------------------------------------------------------------------
// FindClosestMapVertexToWorldCoords
//-------------------------------------------------------------------------------------------------
MapVertex* MapArea::FindClosestMapVertexToWorldCoords( Vector2 worldCoords, float& distSquaredToElement )
{
	float closestDistSquared = FLT_MAX;
	MapVertex* closest = NULL;

	std::list<MapVertex*>::iterator begin = m_vertexList.begin();
	std::list<MapVertex*>::iterator end = m_vertexList.end();
	std::list<MapVertex*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapVertex* vertex = (*iter);
		float distSquared = DistanceSquaredBetweenPoints2d( vertex->m_pos, worldCoords );
		if( distSquared < closestDistSquared )
		{
			closestDistSquared = distSquared;
			closest = vertex;
		}
	}

	distSquaredToElement = closestDistSquared;
	return closest;
}


//-------------------------------------------------------------------------------------------------
// FindClosestMapLinedefToWorldCoords
//-------------------------------------------------------------------------------------------------
MapLinedef* MapArea::FindClosestMapLinedefToWorldCoords( Vector2 worldCoords, float& distSquaredToElement )
{
	float closestDistSquared = FLT_MAX;
	MapLinedef* closest = NULL;

	std::list<MapLinedef*>::iterator begin = m_linedefList.begin();
	std::list<MapLinedef*>::iterator end = m_linedefList.end();
	std::list<MapLinedef*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapLinedef* linedef = (*iter);
		LineSegment2 lineSegment( linedef->m_start->m_pos, linedef->m_end->m_pos - linedef->m_start->m_pos );
		float distSquared = DistanceSquaredFromLineSegmentToPoint2d( lineSegment, worldCoords, NULL );
		if( distSquared < closestDistSquared )
		{
			closestDistSquared = distSquared;
			closest = linedef;
		}
	}

	distSquaredToElement = closestDistSquared;
	return closest;
}


//-------------------------------------------------------------------------------------------------
// FindClosestMapObjectToWorldCoords
//-------------------------------------------------------------------------------------------------
MapObject* MapArea::FindClosestMapObjectToWorldCoords( Vector2 worldCoords, float& distSquaredToElement )
{
	float closestDistSquared = FLT_MAX;
	MapObject* closest = NULL;

	std::list<MapObject*>::iterator begin = m_objectList.begin();
	std::list<MapObject*>::iterator end = m_objectList.end();
	std::list<MapObject*>::iterator iter;

	for( iter = begin; iter != end; ++ iter )
	{
		MapObject* object = (*iter);
		float distSquared = DistanceSquaredBetweenPoints2d( object->m_pos, worldCoords );
		if( distSquared < closestDistSquared )
		{
			closestDistSquared = distSquared;
			closest = object;
		}
	}

	distSquaredToElement = closestDistSquared;
	return closest;
}





