//-------------------------------------------------------------------------------------------------
// Pane.cpp (game)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "Pane.hpp"


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC Pane* Pane::s_focusPane = NULL;


//-------------------------------------------------------------------------------------------------
// Destructor()
//-------------------------------------------------------------------------------------------------
Pane::~Pane()
{
}


//-------------------------------------------------------------------------------------------------
// Constructors
//-------------------------------------------------------------------------------------------------
Pane::Pane( ViewBase* view, const IntVector2& paneMins, const IntVector2& paneMaxs )
:
m_type( PANE_LEAF ),
m_frontPane( NULL ),
m_backPane( NULL ),
m_view( view ),
m_splitFraction( 0.5f ),
m_splitPoint( 0, 0 ),
m_mins( paneMins ),
m_maxs( paneMaxs ),
m_hasFocus( false )
{
	theConsole.Printf( "Creating pane #%u at %d,%d -> %d,%d\n", (unsigned int) this, m_mins.x, m_mins.y, m_maxs.x, m_maxs.y );

	if( m_view )
	{
		m_view->m_mins = m_mins + IntVector2::ONE;
		m_view->m_maxs = m_maxs - IntVector2::ONE;
		theConsole.Printf( "Setting view #%u at %d,%d -> %d,%d\n", (unsigned int) m_view, m_view->m_mins.x, m_view->m_mins.y, m_view->m_maxs.x, m_view->m_maxs.y );
	}
}


//-------------------------------------------------------------------------------------------------
// Draw
//-------------------------------------------------------------------------------------------------
void Pane::Draw()
{
	/// Check if this is a leaf pane
	if( m_type == PANE_LEAF )
	{
		/// Blank the pane area
		theRenderer.SetViewportAndScissor( m_mins - IntVector2::ONE, GetPaneSize() + IntVector2( 2, 2 ) );
		theRenderer.ClearScreen( Rgba::BLACK );

		/// Draw the pane border
		DrawPaneBorder();

		/// Draw the view itself
		if( m_view )
		{
			/// Set up the viewport and scissor region to be just within the pane (by 1 pixel)
			theRenderer.SetViewportAndScissor( m_view->m_mins, m_view->GetViewSize() );
			m_view->Draw();
			return;
		}
	}

	/// Recurse down both sides of the split
	if( m_frontPane )
	{
		m_frontPane->Draw();
	}

	if( m_backPane )
	{
		m_backPane->Draw();
	}
}


//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void Pane::Update()
{
	/// Check if this is a leaf pane
	if( m_type == PANE_LEAF )
	{
		/// Update the view itself
		if( m_view )
		{
			/// Set up the viewport and scissor region to be just within the pane (by 1 pixel)
			m_view->Update();
			return;
		}
	}

	/// Recurse down both sides of the split
	if( m_frontPane )
	{
		m_frontPane->Update();
	}

	if( m_backPane )
	{
		m_backPane->Update();
	}
}


//-------------------------------------------------------------------------------------------------
// DrawPaneBorder
//-------------------------------------------------------------------------------------------------
void Pane::DrawPaneBorder()
{
	glLineWidth( 1.0f );
	theRenderer.ResetViewportAndScissor();
	theRenderer.SetCoordinateSystemToScreenCoordinates2d();

	Rgba bright( 160, 160, 160 );
	Rgba medium( 128, 128, 128 );
	Rgba dim( 64, 64, 64 );

	glDisable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	if( m_hasFocus )
	{
		bright.SetRGBABytes( 255, 255, 0, 255 );
		medium.SetRGBABytes( 255, 255, 0, 255 );
		dim.SetRGBABytes( 255, 255, 0, 255 );
	}

	/// Draw outer pane buffer
	theRenderer.SetColor( Rgba::BLACK );
	glBegin( GL_LINE_LOOP );
	{
		glVertex2i( m_mins.x - 1, m_mins.y - 1 );
		glVertex2i( m_maxs.x + 1, m_mins.y - 1 );
		glVertex2i( m_maxs.x + 1, m_maxs.y + 1 );
		glVertex2i( m_mins.x - 1, m_maxs.y + 1 );
	}
	glEnd(); // GL_LINE_LOOP

	/// Draw pane boundary
	glLineWidth( 1.0f );
	glBegin( GL_LINE_LOOP );
	{
		theRenderer.SetColor( bright );
		glVertex2i( m_mins.x, m_mins.y );

		theRenderer.SetColor( medium );
		glVertex2i( m_maxs.x, m_mins.y );

		theRenderer.SetColor( dim );
		glVertex2i( m_maxs.x, m_maxs.y );

		theRenderer.SetColor( medium );
		glVertex2i( m_mins.x, m_maxs.y );
	}
	glEnd(); // GL_LINE_LOOP
}



//-------------------------------------------------------------------------------------------------
// Split
//-------------------------------------------------------------------------------------------------
void Pane::Split( ViewBase* newView, float splitFraction, bool isHorizontal, bool isNewPaneFront )
{
	m_splitFraction = splitFraction;
	int splitX = m_mins.x + (int)(splitFraction * (float)( m_maxs.x - m_mins.x ) );
	int splitY = m_mins.y + (int)(splitFraction * (float)( m_maxs.y - m_mins.y ) );
	m_splitPoint.SetXY( splitX, splitY );

	IntVector2 frontMins = m_mins;
	IntVector2 frontMaxs = m_maxs;
	IntVector2 backMins = frontMins;
	IntVector2 backMaxs = frontMaxs;

	if( isHorizontal )
	{
		m_type = PANE_SPLIT_VERTICAL_CUT;
		frontMaxs.x = m_splitPoint.x - 1;
		backMins.x = m_splitPoint.x + 1;
	}
	else
	{
		m_type = PANE_SPLIT_HORIZONTAL_CUT;
		frontMaxs.y = m_splitPoint.y - 1;
		backMins.y = m_splitPoint.y + 1;
	}

	ViewBase* frontView = newView;
	ViewBase* backView = m_view;
	if( !isNewPaneFront )
	{
		frontView = m_view;
		backView = newView;
	}

	m_frontPane = new Pane( frontView, frontMins, frontMaxs );
	m_backPane = new Pane( backView, backMins, backMaxs );
}


//-------------------------------------------------------------------------------------------------
// SplitRight
//-------------------------------------------------------------------------------------------------
void Pane::SplitRight( ViewBase* newView, float splitFraction )
{
	Split( newView, splitFraction, true, false );
}


//-------------------------------------------------------------------------------------------------
// SplitLeft
//-------------------------------------------------------------------------------------------------
void Pane::SplitLeft( ViewBase* newView, float splitFraction )
{
	Split( newView, splitFraction, true, true );
}


//-------------------------------------------------------------------------------------------------
// SplitAbove
//-------------------------------------------------------------------------------------------------
void Pane::SplitAbove( ViewBase* newView, float splitFraction )
{
	Split( newView, splitFraction, false, true );
}


//-------------------------------------------------------------------------------------------------
// SplitBelow
//-------------------------------------------------------------------------------------------------
void Pane::SplitBelow( ViewBase* newView, float splitFraction )
{
	Split( newView, splitFraction, false, false );
}



//-------------------------------------------------------------------------------------------------
// UpdateFocus
//-------------------------------------------------------------------------------------------------
void Pane::UpdateFocus( const IntVector2& cursorPos )
{
	m_hasFocus = false;
	if( m_view )
	{
		m_view->m_hasFocus = false;
	}

	if( m_type == Pane::PANE_SPLIT_VERTICAL_CUT || m_type == Pane::PANE_SPLIT_HORIZONTAL_CUT )
	{
		m_frontPane->UpdateFocus( cursorPos );
		m_backPane->UpdateFocus( cursorPos );
		return;
	}

	if( cursorPos.x >= m_mins.x &&
		cursorPos.x <= m_maxs.x &&
		cursorPos.y >= m_mins.y &&
		cursorPos.y <= m_maxs.y )
	{
		m_hasFocus = true;
		Pane::s_focusPane = this;
		if( m_view )
		{
			ViewBase::s_focusView = m_view;
			m_view->m_hasFocus = true;
		}
	}
}



//-------------------------------------------------------------------------------------------------
// GetPaneSize
//-------------------------------------------------------------------------------------------------
IntVector2 Pane::GetPaneSize() const
{
	return (m_maxs - m_mins) + IntVector2::ONE;
}



//-------------------------------------------------------------------------------------------------
// ProcessInputEvent
//-------------------------------------------------------------------------------------------------
bool Pane::ProcessInputEvent( TheInputSystem::inputEventType_e eventType, byte_t key )
{
	UNUSED_ARG eventType, key; // placate the compiler
	return false;
}


