//-------------------------------------------------------------------------------------------------
// TheGame.cpp (game)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheGame.hpp"
#include "TheEditor.hpp"

//-------------------------------------------------------------------------------------------------
// Globals
//-------------------------------------------------------------------------------------------------
const char* const	GAME_TITLE					= "Balls To The Wall v0.01";
const char* const	TITLE_SCREEN_PATH			= "data/ui/titlescreen.tga" ;
const char* const	GAME_UI_SCREEN_PATH			= "data/ui/ui_base01.tga" ;
const char* const	DIRECTIONAL_ARROW_PATH		= "data/sprites/arrow.tga" ;
const char* const	MAP_COMPLETE_SUCCESS_PATH	= "data/ui/map_complete_success.tga";
const char* const	SOUND_ITEM_COLLECTED_PATH	= "data/sounds/itemcollected.wav" ;
const char* const	SOUND_BALL_BOUNCE_PATH		= "data/sounds/ballbounce.wav" ;
const char* const	SOUND_BALL_SHOT_PATH		= "data/sounds/ballshot.wav" ;
const char* const	SOUND_MAP_COMPLETE_PATH		= "data/sounds/ballshot.wav" ;
const char* const	START_MAP_NAME				= "map01";
const int			POWER_DELTA			= 1 ;
const int			ROTATION_DELTA		= 1 ;
TheGame* theGame = NULL;

const float BALL_POWER_MULTIPLIER = 20.0f;
const double STARTUP_STATE_MINIMUM_SECONDS = 0.5;


std::list<Vector2> g_TEMP_lastPath;
double g_TEMP_pathSecondsLeft = 0.0;


//-------------------------------------------------------------------------------------------------
// Destructor()
//-------------------------------------------------------------------------------------------------
TheGame::~TheGame()
{
	if ( _titleScreen )
	{
		delete _titleScreen ;
		_titleScreen = NULL ;
	}

	if ( _gameUIScreen )
	{
		delete _gameUIScreen ;
		_gameUIScreen = NULL ;
	}

	if ( _directionalArrow )
	{
		delete _directionalArrow ;
		_directionalArrow = NULL ;
	}

	if( _mapCompleteSuccessScreen )
	{
		delete _mapCompleteSuccessScreen;
		_mapCompleteSuccessScreen = NULL;
	}

	while( !_balls.empty() )
	{
		delete _balls.back();
		_balls.pop_back();
	}

	_currentMap = INT_MAX;
	while( !_maps.empty() )
	{
		delete _maps.back();
		_maps.pop_back();
	}
}


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
TheGame::TheGame()
	:	_state( STATE_STARTUP ),
		_timeEnteredState( 0.0 ),
		_commandConsoleEnabled( false ),
		_commandBufferPosition(0),
		_titleScreen( NULL ),
		_gameUIScreen( NULL ),
		_directionalArrow( NULL ),
		_mapCompleteSuccessScreen( NULL ),
		_initialRotationAngle( 360 - 45 ),
		_initialPower( 50 ),
		_updatePowerDelta( 0 ),
		_updateRotationDelta( 0 ),
		_forcePower( false ),
		_forceRotation( false ),
		_isTurboPower( false ),
		_isTurboRotation( false ),
		_nextPowerUpdateTime( 0.0 ),
		_nextRotationUpdateTime( 0.0 ),
		_currentMap( INT_MAX ),
		_theBall( NULL ),
		_isSoundEnabled( true )
{
	strcpy( _commandBuffer, "" );
}


//-------------------------------------------------------------------------------------------------
// PreInit
//-------------------------------------------------------------------------------------------------
void TheGame::PreInit()
{
	g_TEMP_lastPath.clear();

	/// Assign application name
	theApplication.SetName( GAME_TITLE );

	theEditor = new TheEditor();
	theEditor->PreInit();
}



//-------------------------------------------------------------------------------------------------
// PostInit
//-------------------------------------------------------------------------------------------------
void TheGame::PostInit()
{
// 	theRenderer.ShowFpsToggle();
	theEditor->PostInit();

	_titleScreen				= new Texture( TITLE_SCREEN_PATH );
	_gameUIScreen				= new Texture( GAME_UI_SCREEN_PATH );
	_directionalArrow			= new Texture( DIRECTIONAL_ARROW_PATH );
	_mapCompleteSuccessScreen	= new Texture( MAP_COMPLETE_SUCCESS_PATH );

	theSoundSystem.RegisterSample( SOUND_ITEM_COLLECTED, SOUND_ITEM_COLLECTED_PATH );
	theSoundSystem.RegisterSample( SOUND_BALL_SHOT, SOUND_BALL_SHOT_PATH );
	theSoundSystem.RegisterSample( SOUND_BALL_BOUNCE, SOUND_BALL_BOUNCE_PATH );
	theSoundSystem.RegisterSample( SOUND_MAP_COMPLETE_SUCCESS, SOUND_MAP_COMPLETE_PATH );

	if( _state == STATE_STARTUP )
	{
		WarpTo( START_MAP_NAME );
	}
}

//-------------------------------------------------------------------------------------------------
// Calls appropriate mainloop for the current game state
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame()
{
	switch ( _state )
	{
		case STATE_STARTUP:					TheGame::RunFrameStartup();	break ;
		case STATE_STAGING:					TheGame::RunFrameStaging();	break ;
		case STATE_RUNNING:					TheGame::RunFrameRunning();	break ;
		case STATE_EDITING:					TheGame::RunFrameEditing();	break ;
		case STATE_MAP_COMPLETE_SUCCESS:	TheGame::RunFrameMapCompleteSuccess(); break;
	}

}



///--------------------------------------------------------------------------------------------------
/// Mainloop in Editing state
///--------------------------------------------------------------------------------------------------
void TheGame::RunFrameEditing( void )
{
	theEditor->RunFrame();
	DrawConsole();
}



///--------------------------------------------------------------------------------------------------
/// Mainloop in Staging state
///--------------------------------------------------------------------------------------------------
void TheGame::RunFrameStaging( void )
{
	SetUpCoordinateSystem();
	UpdateInitialStaging();
//	UpdateEverything();
	DrawEverything();
}


///--------------------------------------------------------------------------------------------------
/// Mainloop in Shooting state (ball is bouncing around)
///--------------------------------------------------------------------------------------------------
void TheGame::RunFrameRunning( void )
{
	SetUpCoordinateSystem();
	UpdateEverything();
	DrawEverything();
}


///--------------------------------------------------------------------------------------------------
/// Mainloop in Startup state
///--------------------------------------------------------------------------------------------------
void TheGame::RunFrameStartup( void )
{
	theRenderer.ClearScreen( Rgba::BLACK );
	SetUpCoordinateSystem();
	
	/// draw using the renderer and texture coordinates
	theRenderer.DrawTexturedRectCentered2d( *_titleScreen, Vector2(640.0f, 480.0f), Vector2((float)_titleScreen->GetWidth(), (float)_titleScreen->GetHeight() ) );

	/// After showing the screen for a minimum amount of time, display an instruction message
//	if( ElapsedTimeInCurrentState() > STARTUP_STATE_MINIMUM_SECONDS )
//	{
//		theRenderer.XYPrintf( 100.0f, 100.0f, "Press any key to begin!" );
//	}
}

//-------------------------------------------------------------------------------------------------
void TheGame::RunFrameMapCompleteSuccess( void )
{
	static double voob = 0;
	voob = 16 * fabs( sin( 2 * theApplication.GetTimeFrameBegan() ) ) - 8;

	theRenderer.ClearScreen( Rgba::BLACK );

	/*	
	if( _maps.size() > 0 && _maps.size() > _currentMap )
	{
		_maps[ _currentMap ]->Draw();
	}
	DrawUI();
	
	glEnable( GL_BLEND );
	glDisable( GL_CULL_FACE );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	glBegin( GL_QUADS );
	glColor4f( 0.0f, 0.0f, 0.0f, 0.6f );
	glVertex2f( 0, 0 );
	glVertex2f( 1280, 0 );
	glVertex2f( 1280, 1024 );
	glVertex2f( 0, 1024 );
	glEnd();
	*/
	
	glDisable( GL_BLEND );

	SetUpCoordinateSystem();
	theRenderer.DrawTexturedRectCentered2d( 
		*_mapCompleteSuccessScreen, 
		Vector2(640.0f, 480.0f + (float)voob), 
		Vector2((float)_mapCompleteSuccessScreen->GetWidth(), (float)_mapCompleteSuccessScreen->GetHeight() ) );

	DrawConsole();
}


//-------------------------------------------------------------------------------------------------
// ProcessCommandLine
//-------------------------------------------------------------------------------------------------
void TheGame::ProcessCommandLine( const std::string& cmdLine )
{
	cmdLine; // FIXME: placate the compiler for now
}


//-------------------------------------------------------------------------------------------------
// ProcessInputEvent
//-------------------------------------------------------------------------------------------------
bool TheGame::ProcessInputEvent( TheInputSystem::inputEventType_e eventType, byte_t key )
{
	bool wasHandled = false ;

	if( eventType == TheInputSystem::INPUT_EVENT_KEYDOWN && key == VK_F11 )
	{
		theRenderer.ShowFpsToggle();
		wasHandled = true;
	}
//	else if( eventType == TheInputSystem::INPUT_EVENT_KEYDOWN && key == VK_TILDE )
//	{
//		theRenderer.ShowConsoleToggle();
//		wasHandled = true;
//	}
	else if ( eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
	{
		wasHandled = OnKeyDown( key );
	}
	
	if ( !wasHandled )
	{
		switch ( GetGameState() )
		{
			case STATE_EDITING:
				wasHandled = theEditor->ProcessInputEvent( eventType, key );
				break ;
			case STATE_STAGING:
				wasHandled	= ProcessGameInput( eventType, key );
				break ;
			case STATE_RUNNING:
				wasHandled	= ProcessGameInput( eventType, key );
				break ;
			case STATE_STARTUP:
				if( ElapsedTimeInCurrentState() > STARTUP_STATE_MINIMUM_SECONDS )
				{
					wasHandled = true ;
					SetGameState( STATE_STAGING );
				}
				break ;
			case STATE_MAP_COMPLETE_SUCCESS:
				
				if( eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
				{
					wasHandled = true;
					SetGameState( STATE_STAGING );
					WarpToNextMap();
				}
				break;
		}
	}

	return wasHandled; // FIXME: this should return true if we handled the input event, false if not
}


///----------------------------------------------------------------------------------
/// Input routine for when the game is playing
///-----------------------------------------------------------------------------------
bool TheGame::ProcessGameInput( TheInputSystem::inputEventType_e eventType, byte_t key )
{
	bool wasHandled = false ;

	if( key == 'P' && eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
	{
		/// Toggle physics debug mode
		g_TEMP_physicsDebugMode = !g_TEMP_physicsDebugMode;
		wasHandled					= true ;
	}
	else if( key == 'F' && eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
	{
		/// Toggle physics debug mode
		g_TEMP_enableFreezingPhysics = !g_TEMP_enableFreezingPhysics;
		wasHandled					= true ;
	}
	else if( key == 'S' && eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
	{
		_isSoundEnabled = !_isSoundEnabled;

		/// FIXME: REMOVE THIS TEMP DEBUGGING CODE
		WarpTo( "map02" );
		_initialPower = 71;
		_initialRotationAngle = 102;
	}

	/// Catch key-ups in any state (so keys don't stick across state boundaries)
	if ( eventType == TheInputSystem::INPUT_EVENT_KEYUP )
	{
		if ( key == VK_LEFT || key == VK_RIGHT )
		{
			_isTurboRotation		= false;
			_updateRotationDelta	= 0 ;
			wasHandled				= true ;
		}
		else if ( key == VK_UP || key == VK_DOWN )
		{
			_updatePowerDelta		= 0 ;
			wasHandled				= true ;
		}
	}


	if ( GetGameState() == STATE_STAGING )
	{
		if ( eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
		{
			if ( key == VK_LEFT )
			{
				_forceRotation			= true;
				_isTurboRotation		= false;
				_updateRotationDelta	= -ROTATION_DELTA ;
				wasHandled				= true ;
			}
			else if ( key == VK_RIGHT )
			{
				_forceRotation			= true;
				_isTurboRotation		= false;
				_updateRotationDelta	= ROTATION_DELTA ;
				wasHandled				= true ;
			}
			else if ( key == VK_UP )
			{
				_forcePower				= true;
				_isTurboPower			= false;
				_updatePowerDelta		= POWER_DELTA ;
				wasHandled				= true ;
			}
			else if ( key == VK_DOWN )
			{
				_forcePower				= true;
				_isTurboPower			= false;
				_updatePowerDelta		= -POWER_DELTA ;
				wasHandled				= true ;
			}
			else if ( key == VK_SPACE )
			{
				//shoot the ball with _initialPower, _initialRotationAngle
				Map* map = _maps[ _currentMap ];
				if ( map )
				{
					ShootBall( _initialPower, _initialRotationAngle );
					SetGameState( STATE_RUNNING );
				}

				if( _isSoundEnabled )
				{
					theSoundSystem.PlaySample( SOUND_BALL_SHOT, 0 );
				}
				g_TEMP_isPhysicsFrozen	= false;
				wasHandled				= true ;
			}
			else if ( key == VK_ESCAPE )
			{
				theApplication.RequestShutdown();
				wasHandled				= true ;
			}
			else if( key == VK_SUBTRACT )
			{
				SetGameState( STATE_MAP_COMPLETE_SUCCESS );
			}
		}
		else if ( eventType == TheInputSystem::INPUT_EVENT_KEYREPEAT )
		{
			if ( key == VK_LEFT )
			{
				_forceRotation			= false;
				_isTurboRotation		= true;
				wasHandled				= true ;
			}
			else if ( key == VK_RIGHT )
			{
				_forceRotation			= false;
				_isTurboRotation		= true;
				wasHandled				= true ;
			}
			else if ( key == VK_UP )
			{
				_forcePower				= false;
				_isTurboPower			= true;
				wasHandled				= true ;
			}
			else if ( key == VK_DOWN )
			{
				_forcePower				= false;
				_isTurboPower			= true;
				wasHandled				= true ;
			}
		}
	}
	else if ( GetGameState() == STATE_RUNNING )
	{
		if ( key == VK_ESCAPE && eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
		{
			theApplication.RequestShutdown();
			wasHandled					= true ;
		}
		else if ( key == 'R' && eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
		{
			WarpTo( _currentMapName );
			SetGameState( STATE_STAGING );
			g_TEMP_isPhysicsFrozen		= false;
			wasHandled					= true ;
		}
		else if( key == 'G' && eventType == TheInputSystem::INPUT_EVENT_KEYDOWN )
		{
			/// Unfreeze physics simulation
			g_TEMP_isPhysicsFrozen		= false;
			wasHandled					= true ;
		}
	}

	return wasHandled ;
}



///----------------------------------------------------------------
/// Console stuff
///----------------------------------------------------------------
bool TheGame::OnKeyDown( const byte_t key )
{
	bool wasHandled = false ;

	if ( _commandConsoleEnabled == true )
	{
		if ( key == VK_RETURN )
		{
			ExecuteCommand();
//			theRenderer.ShowConsole( false );
//			_commandConsoleEnabled = false ;
			_commandBufferPosition	= 0 ;
			_commandBuffer[ _commandBufferPosition ] = 0 ;
		}
		else if ( key == VK_ESCAPE || key == VK_TILDE )
		{
			theRenderer.ShowConsole( false );
			_commandConsoleEnabled = false ;
		}
		else if ( key == VK_BACK )
		{
			if ( _commandBufferPosition )
			{
				_commandBuffer[ --_commandBufferPosition ]	= 0 ;
			}
		}
		else 
		{
			_commandBuffer[ _commandBufferPosition ]	= key ;
			_commandBufferPosition++ ;
			_commandBuffer[ _commandBufferPosition ]	= 0 ;
		}
		wasHandled = true ;
	}
	else if ( key == VK_TILDE )
	{
		theRenderer.ShowConsole( true );
		_commandConsoleEnabled = true ;
		wasHandled = true ;
	}

	return wasHandled ;
}


///-----------------------------------------------------
/// Console command processing
///-----------------------------------------------------
bool TheGame::ExecuteCommand( void )
{
	theConsole.Printf( "Executing Console Command: %s\n", _commandBuffer );

	if ( strnicmp( _commandBuffer, "warp", 4 ) == 0 )
	{
		char	levelName[ 255 ];
		int		argsMet		= sscanf( _commandBuffer, "WARP %s", &levelName );

		if( argsMet == 1 )
		{
			SetGameState( STATE_STAGING );
			WarpTo( levelName );
		}
	}
	else if ( strnicmp( _commandBuffer, "set", 3 ) == 0 )
	{
		char	propertyName[ 255 ];
		char	propertyValue[ 255 ];
		int		argsMet					= sscanf( _commandBuffer, "SET %s %s", &propertyName, &propertyValue );

		if ( argsMet == 2 )
		{
			theEditor->SetPropertyForHighlightedElements( propertyName, propertyValue );
		}
	}
	else if ( strnicmp( _commandBuffer, "delete", 6 ) == 0 )
	{
		char	propertyName[ 255 ];
		int		argsMet					= sscanf( _commandBuffer, "DELETE %s", &propertyName );

		if ( argsMet == 1 )
		{
			theEditor->DeletePropertyForHighlightedElements( propertyName );
		}
	}
	else if ( strnicmp( _commandBuffer, "save", 4 ) == 0 )
	{
		char	mapName[ 255 ] ;
		int		argsMet					= sscanf( _commandBuffer, "SAVE %s", &mapName );

		if ( argsMet == 1 )
		{
			theEditor->SaveMap( mapName );
		}
		else
		{
			theEditor->SaveMap( g_currentArea->m_mapName );
		}
	}
	else if ( strnicmp( _commandBuffer, "load", 4 ) == 0 )
	{
		char	mapName[ 255 ] ;
		int		argsMet					= sscanf( _commandBuffer, "LOAD %s", &mapName );

		if ( argsMet == 1 )
		{
			theEditor->LoadMap( mapName );
		}
	}
	else if ( strnicmp( _commandBuffer, "bg", 2 ) == 0 )
	{
		char	backgroundName[ 255 ] ;
		int		argsMet					= sscanf( _commandBuffer, "BG %s", &backgroundName );

		if ( argsMet == 1 )
		{
			std::string	backgroundPath( "data/maps/" );
			backgroundPath	+= backgroundName ;
			backgroundPath	+= ".tga" ;
			theEditor->SetBackgroundImage( backgroundPath );
		}
	}
	else if( strnicmp( _commandBuffer, "edit", 4 ) == 0 )
	{
		theEditor->EditCurrentMap();
	}
	else if( strnicmp( _commandBuffer, "play", 4 ) == 0 )
	{
		theEditor->PlayCurrentMap();
	}
	else if( strnicmp( _commandBuffer, "quit", 4 ) == 0 )
	{
		theApplication.RequestShutdown();
	}

	return false ;
}



//-------------------------------------------------------------------------------------------------
void TheGame::WarpTo( const std::string& mapName )
{
	while ( !_balls.empty() )
	{
		delete _balls.back();
		_balls.pop_back();
	}

	if( stricmp( mapName.c_str(), _currentMapName.c_str() ) )
	{
		g_TEMP_lastPath.clear();
	}

	_currentMap = LoadMap( mapName );
	_currentMapName	= mapName ;
}

//-------------------------------------------------------------------------------------------------
void TheGame::WarpToNextMap()
{

	Map& currentMap = GetCurrentMap();
	char newMapName[ 1024 ];
	strcpy( newMapName, currentMap.mapArea.m_mapName.c_str() );
	
	/// Find the number in the map name string
	char* numScan = newMapName;
	while( !isdigit( *numScan ) )
	{
		if( *numScan == NULL )
		{
			/// Map name does not contain a number; warp to self
			WarpTo( currentMap.mapArea.m_mapName );
			return;
		}
		numScan ++;
	}

	/// Extract and increment
	int number = atoi( numScan );
	number ++;
	char numberString[ 1024 ];
	sprintf( numberString, "%02d", number );

	/// Replace with new number
	*numScan = NULL;
	strcat( newMapName, numberString );

	/// Check if the new map exists
	if( DoesMapExist( newMapName ) )
	{
		WarpTo( newMapName );
	}
	else
	{
		WarpTo( START_MAP_NAME );
	}

/*
	// This is ass.  Love, Josh.

	// find the last forward or backward slash in the name
	const int fwslashIndex = currentMap.mapName.find_last_of( '/' );
	const int bkslashIndex = currentMap.mapName.find_last_of( '\\' );
	const int slashIndex = max( fwslashIndex, bkslashIndex ) + 1;

	const int dotIndex = currentMap.tgaFilename.find_last_of( '.' );

	ASSERT( slashIndex >= 0 );
	ASSERT( slashIndex < (int)currentMap.tgaFilename.length() );
	
	ASSERT( dotIndex > slashIndex ); // dot must be after slash or path doens't make sense
	ASSERT( dotIndex < (int)currentMap.tgaFilename.length() );

	// chop off extension and path
	const char* const baseName = currentMap.tgaFilename.c_str();
	char croppedName[ 128 ];
	memset( croppedName, 0, sizeof( croppedName) );
	strncpy( croppedName, &baseName[ slashIndex ], dotIndex - slashIndex );

	// find the number
	const char* namescan = croppedName;
	while( !isdigit( *namescan ) )
	{
		++namescan;
	}
	
	// convert number to int, increment
	const int number = atoi( namescan ) + 1;

	// chop number off of cropped name
	croppedName[ (int)namescan - (int)croppedName ] = 0;

	// warp to new level
	char newName[ 128 ];
	_snprintf( newName, sizeof( newName ), "%s%02d", croppedName, number ); 
*/
}

//-------------------------------------------------------------------------------------------------
int TheGame::FindLoadedMap( const std::string& mapName )
{
	int count = 0;
	for( std::vector< Map* >::iterator iter = _maps.begin(); iter != _maps.end(); ++iter, ++count )
	{
		if( !stricmp( (*iter)->mapArea.m_mapName.c_str(), mapName.c_str() ) )
		{
			// delete the fucker - hack to get map data to reload; should rewrite this whole deal
			Map* map = *iter;
			_maps.erase( iter );
			delete map;
			return -1;

			//return count;
		}
	}
	return -1;
}


//-------------------------------------------------------------------------------------------------
bool TheGame::DoesMapExist( const std::string& mapFilename )
{
	std::string mapPathName = "data/maps/" + mapFilename + ".map";
	FILE* file = fopen( mapPathName.c_str(), "rb" );
	if( file )
	{
		fclose( file );
		return true;
	}

	return false;
}


//-------------------------------------------------------------------------------------------------
int TheGame::LoadMap( const std::string& mapName )
{
	int mapIndex = FindLoadedMap( mapName );
	if( mapIndex == -1 )
	{
		mapIndex = _maps.size();
		_maps.push_back( new Map( mapName ) );
	}

	return mapIndex;
}

//-------------------------------------------------------------------------------------------------
bool TheGame::IsMapComplete() const
{
	return ( GetCurrentMap().collectableItemsRemain == 0 );
}


void TheGame::AddBall( Vector2 position )
{
	_theBall		= new Ball( "data/sprites/ball02.tga", "data/sprites/ballhighlight.tga" );
	_theBall->m_pos	= position ;
	_balls.push_back( _theBall );
}


void TheGame::ShootBall( int power, int rotationAngle )
{
	float speed = (float)power * BALL_POWER_MULTIPLIER ;
	_balls[0]->m_vel.SetFromRadiusCosSinDegrees( speed, rotationAngle );

	g_TEMP_lastPath.clear();
	g_TEMP_pathSecondsLeft = 45.0;
}



void TheGame::CollectItem( void )
{
	if( _isSoundEnabled )
	{
		theSoundSystem.PlaySample( SOUND_ITEM_COLLECTED, 0 );
	}
}


void TheGame::BallBounce( float impactHardness )
{
	const double minSecondsBetweenPlayingThisSound = 0.1f;
	static double lastTimeThisSoundWasPlayed = 0.0;
	const double timeNow = theApplication.GetTimeFrameBegan();
	const double timeElapsed = timeNow - lastTimeThisSoundWasPlayed;
	if( timeElapsed < minSecondsBetweenPlayingThisSound )
	{
		return;
	}

	if( _isSoundEnabled )
	{
		theSoundSystem.PlaySample( SOUND_BALL_BOUNCE, 0, TheSoundSystem::LOOP_DISABLE, impactHardness );
	}
}



//-------------------------------------------------------------------------------------------------
// SetGameState
//-------------------------------------------------------------------------------------------------
void TheGame::SetGameState( GameState state )
{
	_state = state;
	_timeEnteredState = theApplication.GetTimeFrameBegan();
}


//-------------------------------------------------------------------------------------------------
// ElapsedTimeInCurrentState
//-------------------------------------------------------------------------------------------------
double TheGame::ElapsedTimeInCurrentState()
{
	return theApplication.GetTimeFrameBegan() - _timeEnteredState;
}



//-------------------------------------------------------------------------------------------------
// ForceConsoleState
//-------------------------------------------------------------------------------------------------
void TheGame::ForceConsoleState( bool open )
{
	theRenderer.ShowConsole( open );
	_commandConsoleEnabled = open ;
}


